Type: Package
Package: geocodebr
Title: Geolocalização De Endereços Brasileiros (Geocoding Brazilian
        Addresses)
Version: 0.6.3
Authors@R: c(
    person("Rafael H. M.", "Pereira", , "rafa.pereira.br@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-2125-7465")),
    person("Daniel", "Herszenhut", , "dhersz@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-8066-1105")),
    person("Gabriel", "Garcia de Almeida", role = "aut",
           comment = c(ORCID = "0009-0003-3557-7328")),
    person("Arthur", "Bazolli", , "baz.arthur@gmail.com", role = "ctb"),
    person("Pedro", "Milreu Cunha", , "pedro.cunha@ipea.gov.br", role = "ctb"),
    person("ITpS - Instituto Todos pela Saúde", role = "fnd"),
    person("Ipea - Instituto de Pesquisa Econômica Aplicada", role = c("cph", "fnd"))
  )
Description: Método simples e eficiente de geolocalizar dados no Brasil. O 
    pacote é baseado em conjuntos de dados espaciais abertos de endereços 
    brasileiros, utilizando como fonte principal o Cadastro Nacional de Endereços 
    para Fins Estatísticos (CNEFE). O CNEFE é publicado pelo Instituto Brasileiro 
    de Geografia e Estatística (IBGE), órgão oficial de estatísticas e geografia 
    do Brasil. (A simple and efficient method for geolocating data in Brazil. The 
    package is based on open spatial datasets of Brazilian addresses, primarily 
    using the Cadastro Nacional de Endereços para Fins Estatísticos (CNEFE), 
    published by the Instituto Brasileiro de Geografia e Estatística (IBGE), 
    Brazil's official statistics and geography agency.)
License: MIT + file LICENSE
URL: https://github.com/ipea/geocodebr,
        https://ipea.github.io/geocodebr/
BugReports: https://github.com/ipea/geocodebr/issues
Depends: R (>= 4.1.0)
Imports: arrow (>= 15.0.1), checkmate, callr, cli, data.table, DBI,
        dplyr, duckdb, duckspatial (>= 1.0.0), enderecobr (>= 0.5.0),
        fs, glue, h3r, httr2 (>= 1.0.0), nanoarrow (>= 0.3.0.1),
        parallelly, purrr, rlang, sf, sfheaders, tools
Suggests: covr, dbplyr, geobr, ggplot2 (>= 3.3.1), knitr, rmarkdown,
        scales, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Language: pt
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-24 00:41:18 UTC; rafap
Author: Rafael H. M. Pereira [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2125-7465>),
  Daniel Herszenhut [aut] (ORCID:
    <https://orcid.org/0000-0001-8066-1105>),
  Gabriel Garcia de Almeida [aut] (ORCID:
    <https://orcid.org/0009-0003-3557-7328>),
  Arthur Bazolli [ctb],
  Pedro Milreu Cunha [ctb],
  ITpS - Instituto Todos pela Saúde [fnd],
  Ipea - Instituto de Pesquisa Econômica Aplicada [cph, fnd]
Maintainer: Rafael H. M. Pereira <rafa.pereira.br@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-24 14:40:02 UTC
Built: R 4.5.2; ; 2026-05-24 18:27:36 UTC; unix

Type: Package
Package: geocodebr
Title: Geolocalização De Endereços Brasileiros (Geocoding Brazilian
        Addresses)
Version: 0.6.3
Authors@R: c(
    person("Rafael H. M.", "Pereira", , "rafa.pereira.br@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-2125-7465")),
    person("Daniel", "Herszenhut", , "dhersz@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0001-8066-1105")),
    person("Gabriel", "Garcia de Almeida", role = "aut",
           comment = c(ORCID = "0009-0003-3557-7328")),
    person("Arthur", "Bazolli", , "baz.arthur@gmail.com", role = "ctb"),
    person("Pedro", "Milreu Cunha", , "pedro.cunha@ipea.gov.br", role = "ctb"),
    person("ITpS - Instituto Todos pela Saúde", role = "fnd"),
    person("Ipea - Instituto de Pesquisa Econômica Aplicada", role = c("cph", "fnd"))
  )
Description: Método simples e eficiente de geolocalizar dados no Brasil. O 
    pacote é baseado em conjuntos de dados espaciais abertos de endereços 
    brasileiros, utilizando como fonte principal o Cadastro Nacional de Endereços 
    para Fins Estatísticos (CNEFE). O CNEFE é publicado pelo Instituto Brasileiro 
    de Geografia e Estatística (IBGE), órgão oficial de estatísticas e geografia 
    do Brasil. (A simple and efficient method for geolocating data in Brazil. The 
    package is based on open spatial datasets of Brazilian addresses, primarily 
    using the Cadastro Nacional de Endereços para Fins Estatísticos (CNEFE), 
    published by the Instituto Brasileiro de Geografia e Estatística (IBGE), 
    Brazil's official statistics and geography agency.)
License: MIT + file LICENSE
URL: https://github.com/ipea/geocodebr,
        https://ipea.github.io/geocodebr/
BugReports: https://github.com/ipea/geocodebr/issues
Depends: R (>= 4.1.0)
Imports: arrow (>= 15.0.1), checkmate, callr, cli, data.table, DBI,
        dplyr, duckdb, duckspatial (>= 1.0.0), enderecobr (>= 0.5.0),
        fs, glue, h3r, httr2 (>= 1.0.0), nanoarrow (>= 0.3.0.1),
        parallelly, purrr, rlang, sf, sfheaders, tools
Suggests: covr, dbplyr, geobr, ggplot2 (>= 3.3.1), knitr, rmarkdown,
        scales, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Language: pt
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-24 00:41:18 UTC; rafap
Author: Rafael H. M. Pereira [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2125-7465>),
  Daniel Herszenhut [aut] (ORCID:
    <https://orcid.org/0000-0001-8066-1105>),
  Gabriel Garcia de Almeida [aut] (ORCID:
    <https://orcid.org/0009-0003-3557-7328>),
  Arthur Bazolli [ctb],
  Pedro Milreu Cunha [ctb],
  ITpS - Instituto Todos pela Saúde [fnd],
  Ipea - Instituto de Pesquisa Econômica Aplicada [cph, fnd]
Maintainer: Rafael H. M. Pereira <rafa.pereira.br@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-24 14:40:02 UTC
Built: R 4.5.2; ; 2026-05-25 02:11:40 UTC; unix

# Generated by roxygen2: do not edit by hand

export(busca_por_cep)
export(definir_campos)
export(definir_pasta_cache)
export(deletar_pasta_cache)
export(download_cnefe)
export(geocode)
export(geocode_reverso)
export(listar_dados_cache)
export(listar_pasta_cache)
importFrom(data.table,"%chin%")
importFrom(data.table,"%like%")
importFrom(data.table,":=")
importFrom(data.table,.GRP)
importFrom(data.table,.I)
importFrom(data.table,.N)
importFrom(data.table,.SD)
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,case_when)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(stats,na.omit)
importFrom(stats,weighted.mean)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# geocodebr: Geolocalização de Endereços Brasileiros <img align="right" src="man/figures/logo.svg" alt="" width="180">

[![CRAN
status](https://www.r-pkg.org/badges/version/geocodebr)](https://CRAN.R-project.org/package=geocodebr)
[![CRAN/METACRAN Total
downloads](https://cranlogs.r-pkg.org/badges/grand-total/geocodebr?color=blue)](https://CRAN.R-project.org/package=geocodebr)
[![check](https://github.com/ipea/geocodebr/workflows/check/badge.svg)](https://github.com/ipea/geocodebr/actions)
[![Codecov test
coverage](https://codecov.io/gh/ipea/geocodebr/branch/main/graph/badge.svg)](https://app.codecov.io/gh/ipea/geocodebr?branch=main)
[![Lifecycle:
experimental](https://lifecycle.r-lib.org/articles/figures/lifecycle-experimental.svg)](https://lifecycle.r-lib.org/articles/stages.html)

O **{geocodebr}** é um pacote computacional para geolicalização de
endereços Brasileiros. O pacote oferece uma maneira simples e eficiente
de geolocalizar dados sem limite de número de consultas. O pacote é
baseado em conjuntos de dados espaciais abertos de endereços
brasileiros, utilizando como fonte principal o Cadastro Nacional de
Endereços para Fins Estatísticos (CNEFE). O CNEFE é
[publicado](https://www.ibge.gov.br/estatisticas/sociais/populacao/38734-cadastro-nacional-de-enderecos-para-fins-estatisticos.html)
pelo Instituto Brasileiro de Geografia e Estatística (IBGE). Atualmente,
o pacote está disponível apenas em R.

## Instalação

A última versão estável pode ser baixada do CRAN com o comando:

``` r
# from CRAN
install.packages("geocodebr")
```

Caso prefira, a versão em desenvolvimento:

``` r
# install.packages("remotes")
remotes::install_github("ipea/geocodebr")
```

## Utilização

O {geocodebr} possui três funções principais para geolocalização de
dados:

1.  `geocode()`
2.  `geocode_reverso()`
3.  `busca_por_cep()`

### 1. Geolocalização: de endereços para coordenadas espaciais

Uma vez que você possui uma tabela de dados (`data.frame`) com endereços
no Brasil, a geolocalização desses dados pode ser feita em apenas dois
passos:

1.  O primeiro passo é usar a função `definir_campos()` para indicar os
    nomes das colunas no seu `data.frame` que correspondem a cada campo
    dos endereços.

2.  O segundo passo é usar a função `geocode()` para encontrar as
    coordenadas geográficas dos endereços de input.

``` r
library(geocodebr)
library(sf)

# carregando uma amostra de dados
input_df <- read.csv(system.file("extdata/small_sample.csv", package = "geocodebr"))

# Primeiro passo: indicar o nome das colunas com cada campo dos enderecos
campos <- geocodebr::definir_campos(
  logradouro = "nm_logradouro",
  numero = "Numero",
  cep = "Cep",
  localidade = "Bairro",
  municipio = "nm_municipio",
  estado = "nm_uf"
  )

# Segundo passo: geolocalizar
df <- geocodebr::geocode(
  enderecos = input_df,
  campos_endereco = campos,
  resultado_completo = FALSE,
  resolver_empates = TRUE,
  resultado_sf = TRUE,
  verboso = FALSE
  )
```

Os resultados do **{geocodebr}** são classificados em seis categorias
gerais de `precisao`, dependendo do nível de exatidão com que cada
endereço de input foi encontrado nos dados do CNEFE. Os resultados
trazem ainda uma estimativa da incerteza da localização encontrada como
um `desvio_metros`. Para mais informações, consulte a documentação da
função ou a [**vignette
“geocode”**](https://ipea.github.io/geocodebr/articles/geocode.html).

### 2. Geolocalização reversa: de coordenadas espaciais para endereços

A função `geocode_reverso()`, por sua vez, permite a geolocalização
reversa, ou seja, a busca de endereços próximos a um conjunto de
coordenadas geográficas. A função pode ser útil, por exemplo, para
identificar endereços próximos a pontos de interesse, como escolas,
hospitais, ou locais de acidentes. Aqui um exemplo de como usar a
função:

``` r
# amostra de pontos espaciais
pontos <- readRDS(
   system.file("extdata/pontos.rds", package = "geocodebr")
   )

pontos <- pontos[1:20,]

# geocode reverso
df_enderecos <- geocodebr::geocode_reverso(
 pontos = pontos,
 dist_max = 1000,
 verboso = FALSE
 )
```

Mais detalhes na [**vignette “geocode
reverso”**](https://ipea.github.io/geocodebr/articles/geocode_reverso.html).

### 3. Busca por CEPs

Por fim, a função `busca_por_cep()` permite fazer consultas de CEPs para
encontrar endereços associados a cada CEP e suas coordenadas espaciais.

``` r
# amostra de CEPs
ceps <- c("70390-025", "20071-001")

df_ceps <- geocodebr::busca_por_cep(
 cep = ceps,
 resultado_sf = TRUE,
 verboso = FALSE
 )
```

## Nota <a href="https://www.ipea.gov.br"><img src="man/figures/ipea_logo.png" alt="IPEA" align="right" width="300"/></a>

Os dados originais do CNEFE são coletados pelo Instituto Brasileiro de
Geografia e Estatística (IBGE). O **{geocodebr}** foi desenvolvido por
uma equipe do Instituto de Pesquisa Econômica Aplicada (Ipea), e conta
com apoio do Instituto Todos pela Saúde (ITpS).

## Instituições utilizando o {geocodebr}

Além de diversos pesquisadores e empresas que utilizam o {geocodebr}, o
pacote também tem sido utilizado por algumas instituições públicas no
planejamento e avaliação de políticas públicas. Entre elas:

- Instituto Brasileiro de Geografia e Estatistica (IBGE)
- Banco Central do Brasil (BCB)
- Ministério do Desenvolvimento Social e Combate à Fome (MDS)

## Projetos relacionados

Existem diversos pacotes de geolocalização disponíveis, muitos dos quais
podem ser utilizados em R (listados abaixo). A maioria dessas
alternativas depende de softwares e conjuntos de dados comerciais,
geralmente impondo limites de número de consultas gratuitas. Em
contraste, as principais vantagens do **{geocodebr}** são que o pacote:
(a) é completamente gratuito, permitindo consultas ilimitadas sem nenhum
custo; (b) opera com alta velocidade e escalabilidade eficiente,
permitindo geocodificar milhões de endereços em apenas alguns minutos,
sem a necessidade de infraestrutura computacional avançada ou de alto
desempenho.

- [{arcgisgeocode}](https://cran.r-project.org/package=arcgisgeocode)
  and [{arcgeocoder}](https://cran.r-project.org/package=arcgeocoder):
  utiliza serviço de geocode do ArcGIS
- [{nominatimlite}](https://cran.r-project.org/package=nominatimlite):
  baseado dados do OSM
- [{photon}](https://cran.r-project.org/package=photon): baseado dados
  do OSM
- [{tidygeocoder}](https://cran.r-project.org/package=tidygeocoder): API
  para diversos servicos de geolocalização
- [{googleway}](https://cran.r-project.org/package=googleway) and
  [{mapsapi}](https://cran.r-project.org/package=mapsapi): interface
  para API do Google Maps

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)

## -----------------------------------------------------------------------------
library(geocodebr)

# leitura de amostra de dados
ends <- read.csv(system.file("extdata/small_sample.csv", package = "geocodebr"))

# definição dos campos de endereço
campos <- definir_campos(
  estado = "nm_uf",
  municipio = "nm_municipio",
  logradouro = "nm_logradouro",
  numero = "Numero",
  cep = "Cep",
  localidade = "Bairro"
)

## -----------------------------------------------------------------------------
# geolicalização
ends_geo <- geocode(
  enderecos = ends, 
  campos_endereco = campos, 
  resultado_completo = FALSE,
  resolver_empates = TRUE,
  h3_res = 9,
  resultado_sf = FALSE,
  verboso = FALSE
  )

head(ends_geo)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)

## -----------------------------------------------------------------------------
library(geocodebr)
library(sf)

# amostra de pontos espaciais
pontos <- readRDS(
  system.file("extdata/pontos.rds", package = "geocodebr")
)

pontos <- pontos[1:20,]

# geocode reverso
df_enderecos <- geocodebr::geocode_reverso(
  pontos = pontos,
  dist_max = 1000,
  verboso = FALSE,
  n_cores = 1
)

head(df_enderecos)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)

## ----eval = FALSE-------------------------------------------------------------
# install.packages("geocodebr")

## ----eval = FALSE-------------------------------------------------------------
# # install.packages("remotes")
# remotes::install_github("ipea/geocodebr")

## -----------------------------------------------------------------------------
library(geocodebr)
library(sf)

# carregando uma amostra de dados
input_df <- read.csv(system.file("extdata/small_sample.csv", package = "geocodebr"))

# Primeiro passo: inidicar o nome das colunas com cada campo dos enderecos
campos <- geocodebr::definir_campos(
  logradouro = "nm_logradouro",
  numero = "Numero",
  cep = "Cep",
  localidade = "Bairro",
  municipio = "nm_municipio",
  estado = "nm_uf"
)

# Segundo passo: geolocalizar
df <- geocodebr::geocode(
  enderecos = input_df,
  campos_endereco = campos,
  resultado_completo = FALSE,
  resolver_empates = FALSE,
  h3_res = NULL,
  resultado_sf = FALSE,
  verboso = FALSE,
  cache = TRUE,
  n_cores = 1
)


## -----------------------------------------------------------------------------
# amostra de pontos espaciais
pontos <- readRDS(
  system.file("extdata/pontos.rds", package = "geocodebr")
)

# seleciona somente os primeiros 20 pontos
pontos <- pontos[1:20,]

# geocode reverso
df_enderecos <- geocodebr::geocode_reverso(
  pontos = pontos,
  dist_max = 1000,
  verboso = FALSE,
  n_cores = 1
)


## -----------------------------------------------------------------------------
# amostra de CEPs
ceps <- c("70390-025", "20071-001")

df_ceps <- geocodebr::busca_por_cep(
  cep = ceps,
  h3_res = 9,
  resultado_sf = FALSE,
  verboso = FALSE
)


## -----------------------------------------------------------------------------
geocodebr::listar_pasta_cache()

geocodebr::listar_dados_cache()

library(geocodebr)
library(ipeadatalake)
library(dplyr)
library(data.table)
library(bench)
library(janitor)





# cad unico --------------------------------------------------------------------

get_cad_year <- function(yyyymm){

  # yyyymm = 201912
  # data de atualizacao do cadastro
  dat <- lubridate::ym(yyyymm)
  dat_limite <- dat - (24*30) # 24 meses

  cad_con <- ipeadatalake::ler_cadunico(
    data = yyyymm,
    base = 'familia',
    as_data_frame = F,
    colunas = c("co_familiar_fam", "dt_atualizacao_fam", "co_uf", "cd_ibge_cadastro",
                "no_localidade_fam", "no_tip_logradouro_fam",
                "no_tit_logradouro_fam", "no_logradouro_fam",
                "nu_logradouro_fam", "ds_complemento_fam",
                "ds_complemento_adic_fam",
                "nu_cep_logradouro_fam", "co_unidade_territorial_fam",
                "no_unidade_territorial_fam", "co_local_domic_fam")
  ) |>
    filter(cd_ibge_cadastro %in% c(2927408, 2919207)) |>  # salvador e lauro de freitas |>
    filter(dt_atualizacao_fam > dat_limite)


  # a <- tail(cad, n = 100) |> collect()

  # compose address fields
  cad <- cad_con |>
    mutate(ano = yyyymm,
          no_tip_logradouro_fam = ifelse(is.na(no_tip_logradouro_fam), '', no_tip_logradouro_fam),
           no_tit_logradouro_fam = ifelse(is.na(no_tit_logradouro_fam), '', no_tit_logradouro_fam),
           no_logradouro_fam = ifelse(is.na(no_logradouro_fam), '', no_logradouro_fam)
    ) |>
    mutate(abbrev_state = co_uf,
           code_muni = cd_ibge_cadastro,
           logradouro = paste(no_tip_logradouro_fam, no_tit_logradouro_fam, no_logradouro_fam),
           numero = nu_logradouro_fam,
           cep = nu_cep_logradouro_fam,
           bairro = no_localidade_fam) |>
    select(co_familiar_fam,
           ano,
           abbrev_state,
           code_muni,
           logradouro,
           numero,
           cep,
           bairro) |>
    dplyr::compute() |>
    dplyr::collect()

  return(cad)
  }


todos_anos <- paste0(2012:2019, 12) |> as.numeric()



cad <- pbapply::pblapply(
  X = todos_anos,
  FUN = get_cad_year
  ) |>
  data.table::rbindlist()


cad$id <- 1:nrow(cad)

fields_cad <- geocodebr::definir_campos(
  logradouro = 'logradouro',
  numero = 'numero',
  cep = 'cep',
  localidade = 'bairro',
  municipio = 'code_muni',
  estado = 'abbrev_state'
)


bench::system_time(
  cadgeo <- geocode(
    enderecos  = cad,
    campos_endereco = fields_cad,
    n_cores = 7,
    verboso = T,
    resultado_completo = T,
    resolver_empates = T,
    resultado_sf = FALSE,
    h3_res = 9
  )
)

# 121.043 empates

nrow(cadgeo)
janitor::tabyl(cadgeo$precisao)
quantile(cadgeo$desvio_metros, probs = c(.5, .75, .8, .9, .92))

register_cnefe_table <- function(con, match_type) {
  # nocov start

  # message("register_cnefe_table")

  # match_type = "dn01"

  # get corresponding table and key cols
  cnefe_table_name <- get_reference_table(match_type)

  # build path to local file
  files <- geocodebr::listar_dados_cache()
  path_to_parquet <- files[grepl(paste0(cnefe_table_name, ".parquet"), files)]

  # make sure we get only the one file from current data release
  path_to_parquet <- path_to_parquet[ grepl(data_release, path_to_parquet) ]



  # # ----------------------------------------------------------------------------
  # # check if table already exists
  # recorded_tbls <- duckdb::duckdb_list_arrow(conn = con)
  # if (cnefe_table_name %in% recorded_tbls) {
  #   return(TRUE)
  # }
  #
  # # determine geographical scope of the search
  # input_states <- DBI::dbGetQuery(con, "SELECT DISTINCT estado FROM input_padrao_db;")$estado |>
  #   sort()
  #
  # input_municipio <- DBI::dbGetQuery(con, "SELECT DISTINCT municipio FROM input_padrao_db;")$municipio |>
  #   sort()
  #
  # # connect to parquet file
  # cnefe_tbl <- arrow_open_dataset( path_to_parquet )
  #
  # # filter cnefe to include only states and municipalities
  # if (length(input_municipio) < 5000 | length(input_states) < 27) {
  #
  #   cnefe_tbl <- cnefe_tbl |>
  #     dplyr::filter(estado %in% input_states) |>
  #     dplyr::filter(municipio %in% input_municipio) |>
  #     dplyr::compute()
  # }
  #
  # # register filtered_cnefe to db
  # duckdb::duckdb_register_arrow(
  #   conn = con,
  #   name =  cnefe_table_name,
  #   arrow_scannable = cnefe_tbl
  #   )
  #
  # # ----------------------------------------------------------------------------

  # if table already exists, stop
  if (duckdb::dbExistsTable(con, cnefe_table_name)) {
    return(TRUE)
  }

  query_filter_cnefe <- glue::glue(
    "CREATE TEMP TABLE IF NOT EXISTS {cnefe_table_name} AS
          WITH unique_munis AS (
              SELECT DISTINCT municipio
              FROM input_padrao_db
          ),
          unique_states AS (
              SELECT DISTINCT estado
              FROM input_padrao_db
          )
          SELECT *
          FROM read_parquet('{path_to_parquet}') m
          WHERE m.estado IN (SELECT estado FROM unique_states)
               AND m.municipio IN (SELECT municipio FROM unique_munis);"
  )

  DBI::dbExecute(con, query_filter_cnefe)

  # # create index
  # if (cnefe_table_name %in% c("municipio_logradouro_cep_localidade",
  #                             "municipio_logradouro_numero_cep_localidade")) {
  #
  #   # index name
  #   idx_name <- paste0("idx_", cnefe_table_name)
  #
  #   # columns to index
  #   cols_index <- strsplit(x = cnefe_table_name, "_") |> unlist()
  #   cols_index <- cols_index[!cols_index %in% "numero"]
  #
  #   cols_index <- c("municipio", "logradouro")
  #   cols_expr <- paste(c("estado", cols_index), collapse = ", ")
  #
  #     query_index <- glue::glue("
  #       CREATE INDEX IF NOT EXISTS {idx_name}
  #         ON {cnefe_table_name} ({cols_expr});")
  #
  #     DBI::dbExecute(con, query_index)
  # }

  return(TRUE)
} # nocov end


# create small table with unique logradouros
register_unique_logradouros_table <- function(con, match_type) {
  # nocov start

  # match_type = "pn03"
  # get_reference_table(match_type)

  # get corresponding key cols and unique cols to keep
  key_cols <- get_key_cols(match_type)

  # determine reference table
  cnefe_table_name <- ifelse(
    match_type %in% c("pn03", "pa03", "pl03"),
    "municipio_logradouro_localidade",
    "municipio_logradouro_cep_localidade"
  )

  # create name of table with unique logradouros
  unique_logr_tbl_name <- paste0("unique_logr_", cnefe_table_name)

  # cols to keep unique values
  select_cols <- key_cols[!key_cols %in% c("numero")]

  # path to parquet
  unique_logr_tbl_parquet <- paste0(cnefe_table_name, ".parquet")
  files <- geocodebr::listar_dados_cache()
  path_to_parquet <- files[grepl(unique_logr_tbl_parquet, files)]

  # make sure we get only the one file from current data release
  path_to_parquet <- path_to_parquet[ grepl(data_release, path_to_parquet) ]

  # should use DISTINCT rows
  DISTINCT <- "DISTINCT"
  if (
    cnefe_table_name == "municipio_logradouro_localidade" |
      all(c("localidade", "cep") %in% select_cols)
  ) {
    DISTINCT <- ""
  }

  # # ----------------------------------------------------------------------------
  # # check if table already exists
  # recorded_tbls <- duckdb::duckdb_list_arrow(conn = con)
  # if (unique_logr_tbl_name %in% recorded_tbls) {
  #   return(unique_logr_tbl_name)
  # }
  #
  # # determine geographical scope of the search
  # input_states <- DBI::dbGetQuery(con, "SELECT DISTINCT estado FROM input_padrao_db;")$estado |>
  #   sort()
  # input_municipio <- DBI::dbGetQuery(con, "SELECT DISTINCT municipio FROM input_padrao_db;")$municipio |>
  #   sort()
  #
  # # connect to parquet file
  # unique_logradouros_arrw <- arrow_open_dataset( path_to_parquet )
  #
  #
  # # filter unique logradouros to include only states and municipalities
  # if (length(input_municipio) < 5000 | length(input_states) < 27) {
  #
  #   unique_logradouros_arrw <- unique_logradouros_arrw |>
  #     dplyr::filter(estado %in% input_states) |>
  #     dplyr::filter(municipio %in% input_municipio) |>
  #     dplyr::compute()
  # }
  #
  # if (DISTINCT=="DISTINCT") {
  #   unique_logradouros_arrw <- unique_logradouros_arrw |>
  #     dplyr::select(dplyr::all_of(select_cols))
  #     dplyr::distinct() |>
  #     dplyr::compute()
  # }
  #
  #
  # duckdb::duckdb_register_arrow(
  #   conn = con,
  #   name = unique_logr_tbl_name,
  #   arrow_scannable = unique_logradouros_arrw
  # )
  # # ----------------------------------------------------------------------------

  # ainda dah pra fazer (1) manter apenas as unique_cols, fazer filtro apenas quando
  # forem poucos munis < 5000

  # if table already exists, stop
  if (duckdb::dbExistsTable(con, unique_logr_tbl_name)) {
    return(unique_logr_tbl_name)
  }

  select_cols <- paste(select_cols, collapse = ', ')

  # se tabela raiz ja existe, filtra dela
  if (duckdb::dbExistsTable(con, cnefe_table_name)) {
    query_unique_logradouros <- glue::glue(
      "CREATE TEMP TABLE IF NOT EXISTS {unique_logr_tbl_name} AS
          WITH unique_munis AS (
              SELECT DISTINCT municipio
              FROM input_padrao_db
          ),
          unique_states AS (
              SELECT DISTINCT estado
              FROM input_padrao_db
          )

          SELECT {DISTINCT} {select_cols}
            FROM {cnefe_table_name}
            WHERE {cnefe_table_name}.estado IN (SELECT estado FROM unique_states)
              AND {cnefe_table_name}.municipio IN (SELECT municipio FROM unique_munis);"
      )

    # caso contrario, leia o parquet
  } else {
    query_unique_logradouros <- glue::glue(
      "CREATE TEMP TABLE IF NOT EXISTS {unique_logr_tbl_name} AS
            WITH unique_munis AS (
                SELECT DISTINCT municipio
                FROM input_padrao_db
            )

          SELECT {DISTINCT} {select_cols}
              FROM read_parquet('{path_to_parquet}') m
              WHERE m.municipio IN (SELECT municipio FROM unique_munis);"
    )
  }
  DBI::dbExecute(con, query_unique_logradouros)

  # set index ?
  # key_cols <- get_key_cols(match_type)

  return(unique_logr_tbl_name)
} # nocov end


# write_all_cnefe_tables_to_db <- function(con){
#
#   all_tables <- c("municipio_logradouro_numero_cep_localidade",
#                   "municipio_logradouro_numero_localidade",
#                   "municipio_logradouro_cep_localidade",
#                   "municipio_logradouro_localidade",
#                   "municipio_cep_localidade",
#                   "municipio_cep",
#                   "municipio_localidade",
#                   "municipio"
#   )
#
#
#   # build path to local file
#   files <- geocodebr::listar_dados_cache()
#
#
#   for (i in all_tables){
#
#     path_to_parquet <- files[grepl( paste0(i, ".parquet"), files)]
#
#     query_filter_cnefe <- glue::glue(
#       "CREATE TEMP TABLE {i} AS
#           WITH unique_munis AS (
#               SELECT DISTINCT municipio
#               FROM input_padrao_db
#           ),
#           unique_states AS (
#               SELECT DISTINCT estado
#               FROM input_padrao_db
#           )
#           SELECT *
#           FROM read_parquet('{path_to_parquet}') m
#           WHERE m.municipio IN (SELECT municipio FROM unique_munis)
#                AND m.estado IN (SELECT estado FROM unique_states);"
#
#     )
#
#     DBI::dbExecute(con, query_filter_cnefe)
#   }
#
# }

# register all geocodebr-cnefe tables
register_geocodebr_tables <- function(con) {
  # nocov start

  all_tables <- geocodebr::listar_dados_cache()

  for (i in all_tables) {
    tb_name <- basename(i)
    tb_name <- fs::path_ext_remove(tb_name)

    temp_arrow <- arrow::open_dataset(i)

    duckdb::duckdb_register_arrow(con, tb_name, temp_arrow)
  }
} # nocov ends

#' Faz download dos dados do CNEFE
#'
#' Faz o download de uma versão pre-processada e enriquecida do CNEFE (Cadastro
#' Nacional de Endereços para Fins Estatísticos) que foi criada para o uso deste
#' pacote.
#'
#' @param tabela Nome da tabela para ser baixada. Por padrão, baixa `"todas"`.
#' @template verboso
#' @template cache
#'
#' @return Retorna o caminho para o diretório onde os dados foram salvos.
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' download_cnefe(verboso = FALSE)
#'
#' @export
download_cnefe <- function(tabela = "todas", verboso = TRUE, cache = TRUE) {
  all_files <- c(
    "municipio_logradouro_numero_localidade.parquet", # 4 largest files       ok 3
    "municipio_logradouro_numero_cep_localidade.parquet", # 4 largest files    ok 1
    "municipio.parquet",
    "municipio_cep.parquet",
    "municipio_cep_localidade.parquet",
    "municipio_localidade.parquet",
    # "municipio_logradouro.parquet",
    # "municipio_logradouro_numero_cep.parquet", # 4 largest files
    # "municipio_logradouro_cep.parquet",
    "municipio_logradouro_cep_localidade.parquet", #  ok 1
    # "municipio_logradouro_numero.parquet", # 4 largest files
    "municipio_logradouro_localidade.parquet" #  ok 3
  )
  all_files_basename <- fs::path_ext_remove(all_files)

  # check input
  checkmate::assert_logical(verboso, any.missing = FALSE, len = 1)
  checkmate::assert_logical(cache, any.missing = FALSE, len = 1)

  # seleciona tabela
  if (tabela != "todas") {
    if (!any(all_files %like% tabela)) {
      cli::cli_abort(
        "A 'tabela' deve ser uma das seguintes op\u00e7\u00f5es: {all_files_basename}"
      )
    }

    all_files <- all_files_basename[all_files_basename == tabela]
    all_files <- paste0(all_files, ".parquet")
  }

  data_urls <- glue::glue(
    "https://github.com/ipeaGIT/padronizacao_cnefe/releases/",
    "download/{data_release}/{all_files}"
  )

  # create dir with data release inside the cache dir / dir is versioned
  if (!cache) {
    cache_dir <- as.character(fs::path_norm(tempfile("geocodebr_temp")))
    data_dir <- glue::glue("{cache_dir}/geocodebr_data_release_{data_release}")
    if (!dir.exists(data_dir)) {
      fs::dir_create(data_dir, recurse = TRUE)
    }
  } else {
    # apaga release antigo se houver
    apaga_data_release_antigo()

    cache_dir <- geocodebr::listar_pasta_cache()
    data_dir <- glue::glue("{cache_dir}/geocodebr_data_release_{data_release}")
    if (!dir.exists(data_dir)) {
      fs::dir_create(data_dir, recurse = TRUE)
    }
  }

  # we only need to download data that hasn't been downloaded yet. note that if
  # cache=FALSE data_dir is always empty, so we download all required data

  existing_files <- list.files(data_dir)

  files_to_download <- setdiff(all_files, existing_files)
  files_to_download <- data_urls[all_files %in% files_to_download]

  if (length(files_to_download) == 0) {
    if (verboso) {
      message_usando_cnefe_local()
    }

    return(invisible(cache_dir))
  }

  downloaded_files <- download_files(data_dir, files_to_download, verboso)

  # the download_dir object below should be identical to data_dir, but we return
  # its value, instead of data_dir, just to make sure the that data is
  # downloaded to the correct dir and that nothing went wrong between setting
  # data_dir and downloading the data

  download_dir <- unique(fs::path_dir(downloaded_files))

  return(invisible(cache_dir))
}


download_files <- function(data_dir, files_to_download, verboso) {
  requests <- lapply(files_to_download, httr2::request)

  dest_files <- fs::path(data_dir, basename(files_to_download))

  responses <- perform_requests_in_parallel(requests, dest_files, verboso)

  response_errored <- purrr::map_lgl(
    responses,
    function(r) inherits(r, "error")
  )

  if (any(response_errored)) {
    error_cnefe_download_failed()
  }

  return(dest_files)
}

perform_requests_in_parallel <- function(requests, dest_files, verboso) {
  # we create this wrapper around httr2::req_perform_parallel just for testing
  # purposes. it's easier to mock this function when testing than to mock a
  # function from another package.
  #
  # related test: "errors if could not download the data for one or more states"
  # in test-download_cnefe
  #
  # related help page:
  # https://testthat.r-lib.org/reference/local_mocked_bindings.html

  if (verboso) {
    message_baixando_cnefe()
  }

  httr2::req_perform_parallel(
    requests,
    paths = dest_files,
    on_error = "continue",
    progress = ifelse(verboso == TRUE, ' ', FALSE)
  )
}

error_cnefe_download_failed <- function() {
  geocodebr_error(
    c(
      "Nao foi possivel baixar os dados do CNEFE.",
      "i" = "Por favor, tente novamente."
    ),
    call = rlang::caller_env(n = 2)
  )
}

# skip tests because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("arrow")


tester <- function(tabela = "municipio_localidade", verboso = TRUE, cache = TRUE) {
  download_cnefe(tabela, verboso, cache)
}

test_that("errors with incorrect input", {
  expect_error(tester(tabela = 'banana'))
  expect_error(tester(tabela = 1))
  expect_error(tester(tabela = NA))
  expect_error(tester(tabela = c(TRUE, TRUE)))

  expect_error(tester(verboso = 1))
  expect_error(tester(verboso = NA))
  expect_error(tester(verboso = c(TRUE, TRUE)))

  expect_error(tester(cache = 1))
  expect_error(tester(cache = NA))
  expect_error(tester(cache = c(TRUE, TRUE)))
})

test_that("returns the path to the directory where the files were saved", {
  result <- tester()
  expect_identical(result, file.path(listar_pasta_cache()))

  result <- tester(tabela = "municipio_cep")
  expect_identical(result, file.path(listar_pasta_cache()))

})

test_that("cache usage is controlled by the cache argument", {

  result <- tester(cache = TRUE)

  expect_identical(
    result,
    as.character( fs::path(tools::R_user_dir("geocodebr", which = "cache")))
    )

  # using a mocked binding for perform_requests_in_parallel here just to save us
  # some time. as long as none of its elements is a failed request, the function
  # will make download_files return the path to the files that would be
  # downloaded, which is basically what we want to test here
  local_mocked_bindings(
    perform_requests_in_parallel = function(...) TRUE
  )

  result <- tester(cache = FALSE)
  expect_true(
    grepl(file.path(fs::path_norm(tempdir()), "geocodebr_temp"), result)
  )
})

# test_that("errors if could not download one or more files", {
#   local_mocked_bindings(
#     perform_requests_in_parallel = function(...) {
#       httr2::req_perform_parallel(
#         list(httr2::request("FAILURE")),
#         on_error = "continue"
#       )
#     }
#   )
#
#   expect_error(
#     tester(cache = FALSE),
#     class = "geocodebr_error_cnefe_download_failed"
#   )
#
#   expect_snapshot(tester(cache = FALSE), error = TRUE, cnd_class = TRUE)
# })

geocodebr_error <- function(message, call, .envir = parent.frame()) {
  error_call <- sys.call(-1)
  error_function <- as.name(error_call[[1]])

  error_classes <- c(
    paste0("geocodebr_error_", sub("^error_", "", error_function)),
    "geocodebr_error"
  )

  cli::cli_abort(message, class = error_classes, call = call, .envir = .envir)
}


error_input_nao_padronizado <- function() {
  geocodebr_error(
    c(
      "Os dados de entrada nao estao padronizados.",
      "i" = "Use o argumento `padronizar_enderecos = TRUE`, ou
      padronize os dados de input com `enderecobr::padronizar_enderecos(..., formato_estados = 'sigla', formato_numeros = 'integer')`."
    ),
    call = rlang::caller_env(n = 2)
  )
}

parent_function <- function() error_test()

error_test <- function() {
  geocodebr_error(c("test", "*" = "info"), call = rlang::caller_env())
}

test_that("erro funciona corretamente", {
  expect_error(parent_function(), class = "geocodebr_error_test")
  expect_error(parent_function(), class = "geocodebr_error")

  expect_snapshot(parent_function(), error = TRUE, cnd_class = TRUE)
})

trata_empates_geocode_duckdb <- function(
  con = parent.frame()$con,
  resultado_completo = parent.frame()$resultado_completo,
  resolver_empates = parent.frame()$resolver_empates,
  verboso = parent.frame()$verboso
) {
  # 1) checa se tem empates --------------------------------------

  n_casos_empate <- DBI::dbGetQuery(
    conn = con,
    statement = "SELECT COUNT(*) AS n_casos_empate
                    FROM (
                      SELECT tempidgeocodebr
                      FROM output_db
                      GROUP BY tempidgeocodebr
                      HAVING COUNT(*) > 1
                    ) AS repeated;"
  )[[1]]

  # 2) se nao tiver mais empates, termina aqui --------------------------------------
  if (n_casos_empate == 0) {
    return(n_casos_empate)
  }

  # 3) se nao for para resolver empates: ------------------------------------------
  # - calcula / identifica casos de empate
  # - gera warning
  # - retorna resultado assim mesmo
  if (isFALSE(resolver_empates)) {
    # adicionar coluna de empate
    DBI::dbExecute(
      conn = con,
      statement = "CREATE OR REPLACE TEMP TABLE output_db2 AS
                    SELECT *,
                    (COUNT(*) OVER (PARTITION BY tempidgeocodebr) > 1) AS empate
                    FROM output_db;"
    )

    cli::cli_warn(
      "Foram encontrados {n_casos_empate} casos de empate. Estes casos foram
      marcados com valor `TRUE` na coluna 'empate', e podem ser inspecionados na
      coluna 'endereco_encontrado'. Alternativamente, use `resolver_empates = TRUE`
      para que o pacote lide com os empates automaticamente. Ver
      documenta\u00e7\u00e3o da fun\u00e7\u00e3o."
    )

    return(n_casos_empate)
  }

  # Haversine macro (kept for speed; consider spatial extension later)
  DBI::dbExecute(
    con,
    "
    CREATE MACRO IF NOT EXISTS haversine(lat1, lon1, lat2, lon2) AS (
      6378137 * 2 * ASIN(
        SQRT(
          POWER(SIN(RADIANS(lat2 - lat1) / 2), 2) +
          COS(RADIANS(lat1)) * COS(RADIANS(lat2)) *
          POWER(SIN(RADIANS(lon2 - lon1) / 2), 2)
        )
      )
    );
  "
  )

  # 3 se for para resolver empates, trata de 3 casos separados -----------------
  # D) nao empatados
  # E) empatados perdidos (dist > 1Km e lograoduros ambiguos)
  #    solucao: usa caso com maior contagem_cnefe
  # F) empatados mas que da pra salvar (dist < 1km e logradouros nao ambiguos)
  #    solucao: agrega casos provaveis de serem na mesma rua com media ponderada
  #    das coordenadas, mas retorna  endereco_encontrado do caso com maior
  #    contagem_cnefe
  # questao documentada no issue 37

  additional_cols_final <- ""
  cols_encontradas <- ""

  if (isTRUE(resultado_completo)) {
    additional_cols_final <- glue::glue(
      ", logradouro_encontrado, numero_encontrado, cep_encontrado,
        localidade_encontrada, municipio_encontrado, estado_encontrado,
        similaridade_logradouro, contagem_cnefe, empate, cod_setor"
    )

    cols_encontradas <- glue::glue(
      ", logradouro_encontrado, numero_encontrado, cep_encontrado,
        localidade_encontrada, municipio_encontrado, estado_encontrado,
        similaridade_logradouro, cod_setor"
    )
  }

  sql_resolve <- glue::glue(
    "CREATE OR REPLACE TEMP TABLE output_db2 AS

      -- A) tabela *base* calculate empates iniciais (inclui mesmo aqueles a menos de 300 metros)
      WITH
        base AS (
          SELECT
            *,
            (COUNT(*) OVER (PARTITION BY tempidgeocodebr) > 1) AS empate_inicial,
            ROW_NUMBER() OVER (PARTITION BY tempidgeocodebr ORDER BY contagem_cnefe DESC, desvio_metros, endereco_encontrado) AS id
          FROM output_db
        ),

      -- B) tabela *distd* calculate distancia entre os casos empatados
      distd AS (
          SELECT
            b.*,
            CASE WHEN empate_inicial THEN
              haversine(
                lat, lon,
                LEAD(lat) OVER (PARTITION BY tempidgeocodebr ORDER BY id),
                LEAD(lon) OVER (PARTITION BY tempidgeocodebr ORDER BY id)
              )
            END AS dist_geocodebr_metros
          FROM base b
        ),

      -- C) tabela *filtered* pra manter apenas casos de empate que estao a mais de 300 metros e atualiza coluna de empate
      filtered AS (
          SELECT
            d.*,
            (COUNT(*) OVER (PARTITION BY tempidgeocodebr) > 1) AS empate,
            MAX(dist_geocodebr_metros) OVER (PARTITION BY tempidgeocodebr) AS max_dist
          FROM distd d
          WHERE (empate_inicial IS FALSE)
             OR (empate_inicial AND dist_geocodebr_metros IS NULL)
             OR (empate_inicial AND dist_geocodebr_metros > 300)
        ),

      -- D) tabela *df_sem_empate* com os casos sem empate
      df_sem_empate AS (
          SELECT
            tempidgeocodebr,
            lat,
            lon,
            endereco_encontrado,
            tipo_resultado,
            contagem_cnefe,
            desvio_metros,
            empate {cols_encontradas}
          FROM filtered
          WHERE empate = FALSE
        ),

        -- E) empatados perdidos (exemplo: max_dist > 1000; acrescente demais regras aqui)
        df_empates_perdidos AS (
          SELECT
            tempidgeocodebr,
            lat,
            lon,
            endereco_encontrado,
            tipo_resultado,
            contagem_cnefe,
            desvio_metros,
            TRUE AS empate {cols_encontradas}
          FROM filtered
          WHERE empate = TRUE
            AND (
              max_dist > 1000
              OR log_causa_confusao
              OR REGEXP_MATCHES(endereco_encontrado,
                  '(RUA (QUATRO|QUATORZE|QUINZE|DEZESSEIS|DEZESSETE|DEZOITO|DEZENOVE|VINTE|TRINTA|QUARENTA|CINQUENTA|SESSENTA|SETENTA|OITENTA|NOVENTA))'
              )
            )
            AND NOT REGEXP_MATCHES(logradouro_encontrado, '\\\\bDE (JANEIRO|FEVEREIRO|MARCO|ABRIL|MAIO|JUNHO|JULHO|AGOSTO|SETEMBRO|OUTUBRO|NOVEMBRO|DEZEMBRO)\\\\b')
          QUALIFY ROW_NUMBER()
            OVER (PARTITION BY tempidgeocodebr ORDER BY contagem_cnefe DESC) = 1
        ),

        -- F) empatados salvaveis = restantes (nao em a nem b)
        empates_restantes AS (
          SELECT f.*
          FROM filtered f
          WHERE f.empate = TRUE
            AND NOT EXISTS (SELECT 1 FROM df_sem_empate s WHERE s.tempidgeocodebr = f.tempidgeocodebr)
            AND NOT EXISTS (SELECT 1 FROM df_empates_perdidos p WHERE p.tempidgeocodebr = f.tempidgeocodebr)
        ),
        empates_wavg AS (
          SELECT
            e.*,
            (SUM(lat * contagem_cnefe) OVER (PARTITION BY tempidgeocodebr)
              / NULLIF(SUM(contagem_cnefe) OVER (PARTITION BY tempidgeocodebr), 0)) AS lat_wavg,
            (SUM(lon * contagem_cnefe) OVER (PARTITION BY tempidgeocodebr)
              / NULLIF(SUM(contagem_cnefe) OVER (PARTITION BY tempidgeocodebr), 0)) AS lon_wavg
          FROM empates_restantes e
        ),
        df_empates_salve AS (
          SELECT
            tempidgeocodebr,
            lat_wavg AS lat,
            lon_wavg AS lon,
            endereco_encontrado,
            tipo_resultado,
            contagem_cnefe,
            desvio_metros,
            TRUE AS empate {cols_encontradas}
          FROM empates_wavg
          QUALIFY ROW_NUMBER()
            OVER (PARTITION BY tempidgeocodebr ORDER BY contagem_cnefe DESC) = 1
        )

      -- junta as 3 tabelas numa soh (df_sem_empate, df_empates_perdidos, df_empates_salve)
      SELECT
        tempidgeocodebr, lat, lon, tipo_resultado, desvio_metros,
        endereco_encontrado {additional_cols_final}
      FROM df_sem_empate
      UNION ALL
      SELECT
        tempidgeocodebr, lat, lon, tipo_resultado, desvio_metros,
        endereco_encontrado {additional_cols_final}
      FROM df_empates_perdidos
      UNION ALL
      SELECT
        tempidgeocodebr, lat, lon, tipo_resultado, desvio_metros,
        endereco_encontrado {additional_cols_final}
      FROM df_empates_salve;"
  )

  DBI::dbExecute(con, sql_resolve)

  if (verboso) {
    plural <- ifelse(n_casos_empate == 1, 'caso', 'casos')
    message(glue::glue(
      "Foram encontrados e resolvidos {n_casos_empate} {plural} de empate."
    ))
  }

  return(n_casos_empate)
}

#' proximos passos
#' 0. padronizacao cnefe usado aprox linear para lidar com ruas curvas
#' 1. add census tracts
#' 2. check ground truth data for distance
#' 3. more efficient string dist with cache
#' 4. make solving ties more efficient
#' 5. helper function to parallelize chuncks of the same size
#' 6. targets to track the evolution of the package performance
#'
#'
#'
# esperado
5 - deterministico certinho
7312 - OTIMO antes rua errada no bairro errada, agora pega rua certa no certo

# action
4398 - ! "RUA  DO BECO 12" continua com match probabilistico
1690 - ! v2 parece melhor erro na padronizacao???? enderecobr::padronizar_logradouros('RODOVIA  BR 364') 666666 checar como esta no cnfe e nas tabelas
9577 - ! v2 parece melhor erro na padronizacao???? enderecobr::padronizar_logradouros('RODOVIA  AC 475')
9470 - padronizacao???? enderecobr::padronizar_logradouros


15294 - ! antes pega rua certa no bairro certo (v2 melhor) ????

**** caso 1 - eh melhor o novo

# to ponder
463 -  1 antes pega rua errada no bairro certo, agora pega rua certa no bairro errado lol (muitos 7)
5000 - 1 antes pega rua errada no bairro certo, agora pega rua certa no bairro errado lol
7290 - 1 antes pega rua errada no bairro certo, agora pega rua certa no bairro errado lol (muitos empates)
8490 - 1 antes pega rua errada no bairro certo, agora pega rua certa no bairro errado lol
9808 - 1 antes pega rua errada no bairro certo, agora pega rua certa no bairro errado lol
423 -  1 antes pega rua errada no bairro certo, agora NAO pega rua mas acerta o bairro (desvio grande)
13814 - 1 antes pega rua certa no cep errado , agora NAO pega rua mas acerta o bairro (desvio 700)


5385 - 2 antes pega rua certa no bairro errado, agora NAO pega rua mas acerta o bairro (desvio grande)
7230 - 2 antes pega rua certa no bairro errado, agora NAO pega rua mas acerta o bairro (desvio 700)

13135 - v2 antes pega rua certa no bairro certo, agora NAO pega rua mas acerta o bairro (desvio 900)
14063 - nao tem salvacao, v3 melhor pq eh honesto com desvio grande
15734 - 3 antes pega rua certa no bairro errado, agora NAO pega rua mas acerta o bairro (desvio pequeno anyway)
13292 - 4 antes pega rua certa no bairro certo, agora NAO pega rua mas acerta o bairro (desvio pequeno anyway)
10714 - 4 antes pega rua certa no bairro certo, agora NAO pega rua mas acerta o bairro (desvio pequeno anyway)
9842 - 5 antes rua errada no bairro certo e cep errado, agora pega rua certa no bairro certo e cep errado (dificil)


333 - igual "RUA DOM PEDROII" - probabilistico
4998 - igual "RUA DOM PEDRO II" - deterministico

#' instalar extensoes do duckdb
#' - spatial - acho q nao vale a pena por agora
#'
#' #' (non-deterministic search)
#' - fts - Adds support for Full-Text Search Indexes / "https://medium.com/@havus.it/enhancing-database-search-full-text-search-fts-in-mysql-1bb548f4b9ba"
#'
#'
#'
#' adicionar dados de POI da Meta /overture
#' adicionar dados de enderecos Meta /overture
#'   # NEXT STEPS
#'   - (ok) interpolar numeros na mesma rua
#'   - (next) join probabilistico com fts_main_documents.match_bm25
#'   - (next) calcular nivel de erro na agregacao do cnefe
#'   - optimize disk and parallel operations in duckdb
#'   - casos de rodovias
#'   - interpolar numeros separando impares e pares
#'   - CASES NOT FOUND ? AND THEIR EFFECT ON THE PROGRESS BAR



#' take-away
#' 1) incluir LIKE no campo d elogradouro melhor MUITO performance, encontrando
#' muito mais casos em cases de match 1, 2, 3 e 4
#' o melhor mesmo seria usar fts
#' 2) isso tem pequeno efeito de diminuir performance do dani, e 0 efeito no rafa
#'
#' 3) no rafa aida tem um residudo de que alguns casos em que as coordenadas nao
#' foram agregadas, entao tem alguns 'id's que se repetem no output
#'  - a razao eh pq a agregacao sai diferente para logradouros diferentes mas
#'  com o mesmo padrao LIKE. Ex. "RUA AVELINO CAVALCANTE" e "TRAVESSA AVELINO CAVALCANTE"
#'
#'  exemplos
#' id == 1637 caso de diferentes ruas no mesmo condominio
#'            "RUA DOIS VILA RICA" e "RUA XXVI QUADRA E VILA RICA"
#' id == 1339 (esse se resolve pq sao bairros diferentes)


devtools::load_all('.')

# open input data
data_path <- system.file("extdata/large_sample.parquet", package = "geocodebr")
input_df <- arrow::read_parquet(data_path)


# enderecos = input_df
# n_cores = 7
# ncores <- 7
# verboso = T
# cache = TRUE
# resultado_completo = T
# resultado_sf = F
# campos_endereco <- geocodebr::definir_campos(
#   logradouro = 'logradouro',
#   numero = 'numero',
#   cep = 'cep',
#   localidade = 'bairro',
#   municipio = 'municipio',
#   estado = 'uf')
# resolver_empates = T
# padronizar_enderecos = T
# h3_res = NULL

# benchmark different approaches ------------------------------------------------------------------
ncores <- 7

# input_df <- input_df[c(1,84, 284), ]
# input_df <- input_df[id %in%  dfgeo$id[dfgeo$tipo_resultado=='ei03'] ]


campos <- geocodebr::definir_campos(
  logradouro = 'logradouro',
  numero = 'numero',
  cep = 'cep',
  localidade = 'bairro',
  municipio = 'municipio',
  estado = 'uf'
)

# test probabilistic
# input_df <- input_df[c(7, 32, 34, 71, 173, 1348)]  # pn02 pi02 pn03 pi03 pr01 pr02
# temp_df <- filter(input_df,id %in% c(1371)  )


bench::mark(
  v3 <- geocode(
    enderecos = input_df,
    campos_endereco = campos,
     n_cores = 7,
    resultado_completo = F,
    verboso = T,
    # resultado_sf = T,
    resolver_empates = T,
    # h3_res = 9,
    cache= T, padronizar_enderecos = T
  )
)


gc(T,T,T)
bench::mark(

  iterations = 1, check = F,

  # callr = geocode_callr(
  #   enderecos = df,
  #   campos_endereco = campos,
  #   n_cores = 7,
  #   resultado_completo = F,
  #   resolver_empates = T
  #   ),

  original = geocode(
    enderecos = input_df,
    campos_endereco = campos,
    n_cores = 7,
    resultado_completo = F,
    resolver_empates = T
    )
)

#   expression     min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory
#   <bch:expr> <bch:t> <bch:>     <dbl> <bch:byt>    <dbl> <int> <dbl>   <bch:tm> <list> <list>
# 1 original     24.3m  24.3m  0.000685    8.24GB  0.00479     1     7      24.3m <dt>   <Rprofmem>

#   expression     min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory
# v0.4.0 CRAN     33.5m  33.5m  0.000497    8.06GB  0.00746     1    15      33.5m <NULL> <Rprofmem>
# v0.5.0 dev      22.2m  22.2m  0.000749    8.05GB  0.00674     1     9      22.2m <dt>   <Rprofmem>
# v0.5.0 devcallr 5.94m  5.94m  0.00280   1016.2MB  0           1     0      5.94m <NULL> <Rprofmem>


# args: n_cores = 7, resultado_completo = F, verboso = T, resultado_sf = T, resolver_empates = T, h3_res = 9
# expression      min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory     time       gc
# v0.4.0        28.6s  28.6s    0.0350      83MB    0.245     1     7      28.6s <sf>   <Rprofmem> <bench_tm> <tibble>
# v0.5.0_dev    7.63s  7.63s     0.131    39.4MB    0.918     1     7      7.63s <sf>   <Rprofmem> <bench_tm> <tibble>





empates <- subset(dfgeo, empate==T)
a <- table(empates$id) |> as.data.frame()
table(empates$tipo_resultado)

case <- subset(empates, id ==9477)
case <- subset(empates, endereco_encontrado %like% 'BLOCO')
mapview::mapview(case)


round(table(dfgeo$precisao) / nrow(dfgeo) *100, 1)
round(table(dfgeo2$precisao) / nrow(dfgeo2) *100, 1)

# tempo:
# com 32.75s    855 empates
# sem 13.7s     655 empates

# precisao:
#       cep        localidade        logradouro         municipio            numero numero_aproximado
# com   2.5               0.3              14.8               0.6              45.8              36.0
# sem  31.4               2.3               7.2               0.6              33.3              26.2

temp <- left_join( select(dfgeo, id, tipo_resultado), select(dfgeo2, id, tipo_resultado), by = 'id'  )
t <- table(temp$tipo_resultado.x, temp$tipo_resultado.y) |> as.data.frame()
t <- filter(t, Freq>0)
View(t)





# 20K FRESH
# Unit: seconds
#         expr      min       lq     mean   median       uq      max neval
#    rafa_drop 10.10379 10.25397 10.35726 10.39609 10.41682 10.61564     5
# rafa_drop_db 10.33163 10.41716 10.53599 10.57100 10.67138 10.68880     5
#    rafa_keep 11.12560 11.17642 11.40483 11.30231 11.56706 11.85277     5
# rafa_keep_db 10.99093 11.18329 11.26758 11.18585 11.40816 11.56966     5
#      dani_df 10.99375 11.04987 11.47305 11.16092 11.43530 12.72541     5







# 20 K
#     expression      min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory     time
#     <bch:expr>   <bch:> <bch:>     <dbl> <bch:byt>    <dbl> <int> <dbl>   <bch:tm> <list> <list>     <list>
#   1 rafa_drop     10.4s  10.5s    0.0952      53MB    0.171     5     9      52.5s <NULL> <Rprofmem> <bench_tm>
#   2 rafa_drop_db  10.6s  10.7s    0.0935    27.5MB    0.168     5     9     53.46s <NULL> <Rprofmem> <bench_tm>
#   3 rafa_keep     11.5s  11.7s    0.0859    27.7MB    0.155     5     9     58.23s <NULL> <Rprofmem> <bench_tm>
#   4 rafa_keep_db  11.3s  11.6s    0.0867    27.6MB    0.139     5     8     57.69s <NULL> <Rprofmem> <bench_tm>
#   5 dani_df       11.5s  11.9s    0.0823    24.9MB    0.148     5     9      1.01m <NULL> <Rprofmem> <bench_tm>





devtools::load_all('.')
library(tictoc)
library(dplyr)
library(data.table)
library(ipeadatalake)
library(mapview)
library(sfheaders)
library(sf)
options(scipen = 999)
mapview::mapviewOptions(platform = 'leafgl')
set.seed(42)





geocodebr::get_cache_dir() |>
  geocodebr:::arrow_open_dataset()  |>
  filter(estado=="PR") |>
  filter(municipio == "CURITIBA") |>
  dplyr::compute() |>
  filter(logradouro_sem_numero %like% "DESEMBARGADOR HUGO SIMAS") |>
  dplyr::collect()


cnf <- ipeadatalake::read_cnefe(year = 2022) |>
  #  filter(code_state=="41") |>
  dplyr::filter(code_muni == 4106902) |>
  dplyr::collect()


d <- cnf |>
  filter(nom_seglogr %like% "HUGO SIMAS")

'DESEMBARGADOR'






# small sample data ------------------------------------------------------------------
devtools::load_all('.')
library(dplyr)

# open input data
data_path <- system.file("extdata/small_sample.csv", package = "geocodebr")
input_df <- read.csv(data_path)

campos <- geocodebr::definir_campos(
  logradouro = "nm_logradouro",
  numero = "Numero",
  cep = "Cep",
  localidade = "Bairro",
  municipio = "nm_municipio",
  estado = "nm_uf"
)

# enderecos = input_df
# campos_endereco = campos
# n_cores = 7
# verboso = T
# cache=T
# resultado_completo = F
# resolver_empates = T
# resultado_sf = FALSE
# h3_res =9

dfgeo <- geocodebr::geocode(
    enderecos = input_df,
    campos_endereco = campos,
    n_cores = 7,
    resultado_completo = T,
    resolver_empates = F,
    verboso = T
  )


identical(df_rafaF$id, input_df$id)


table(df_rafa_loop$precision) / nrow(df_rafa_loop)*100


unique(df_rafa$match_type) |> length()

table(df_rafa$match_type)


# parallel callr --------------------------------------

library(future.callr)
library(future)
library(furrr)

future::plan(future::multisession(workers = 3))

input_df$estado <- enderecobr::padronizar_estados(input_df$uf)




bench::bench_time(
a <-   split(input_df, f = input_df$estado) |>
    furrr::future_map(
      .f = function(x){
        geocode(
          enderecos = x,
          campos_endereco = campos,
          resultado_completo = F,
          resolver_empates = T
        )
      }
        )

)


future::plan(future.callr::callr)

bench::bench_time(
  split(input_df, f = "uf") |>
    furrr::future_map(
      .f = function(x){
        geocode_callr(
          enderecos = x,
          campos_endereco = campos,
          n_cores = 7,
          resultado_completo = F,
          resolver_empates = T
        )
      }
    )

)

bench::bench_time(
        geocode_callr(
          enderecos = df,
          campos_endereco = campos,
          n_cores = 10,
          resultado_completo = F,
          resolver_empates = T
        )
)

#' Geolocaliza endereços no Brasil
#'
#' Geocodifica endereços brasileiros com base nos dados do CNEFE. Os endereços
#' de input devem ser passados como um `data.frame`, no qual cada coluna
#' descreve um campo do endereço (logradouro, número, cep, etc). Os resuldos dos
#' endereços geolocalizados podem seguir diferentes níveis de precisão. Consulte
#' abaixo a seção "Detalhes" para mais informações. As coordenadas de output
#' utilizam o sistema de coordenadas geográficas SIRGAS 2000, EPSG 4674.
#'
#' @param enderecos Um `data.frame`. Os endereços a serem geolocalizados. Cada
#'    coluna deve representar um campo do endereço.
#' @param campos_endereco Um vetor de caracteres. A correspondência entre cada
#'    campo de endereço e o nome da coluna que o descreve na tabela `enderecos`.
#'    A função [definir_campos()] auxilia na criação deste vetor e realiza
#'    algumas verificações nos dados de entrada. Campos de endereço passados
#'    como `NULL` serão ignorados, e a função deve receber pelo menos um campo
#'    não nulo, além  dos campos `"estado"` e `"municipio"`, que são
#'    obrigatórios. Note que o campo  `"localidade"` é equivalente a 'bairro'.
#' @param resultado_completo Lógico. Indica se o output deve incluir colunas
#'    adicionais, como o endereço encontrado de referência. Por padrão, é `FALSE`.
#' @param resolver_empates Lógico. Alguns resultados da geolocalização podem
#'    indicar diferentes coordenadas possíveis (e.g. duas ruas diferentes com o
#'    mesmo nome em uma mesma cidade). Esses casos são trados como 'empate' e o
#'    parâmetro `resolver_empates` indica se a função deve resolver esses empates
#'    automaticamente. Por padrão, é `TRUE`, e a função retorna apenas o caso
#'    mais provável. Para mais detalhes sobre como é feito o processo de
#'    desempate, consulte abaixo a seção "Detalhes".
#' @template resultado_sf
#' @template h3_res
#' @param padronizar_enderecos Lógico. Indica se os dados de endereço de entrada
#'    devem ser padronizados. Por padrão, é `TRUE`. Essa padronização é essencial
#'    para uma geolocalizaçao correta. Alerta! Apenas utilize
#'    `padronizar_enderecos = FALSE` caso os dados de input já tenham sido
#'    padronizados anteriormente com `enderecobr::padronizar_enderecos(..., formato_estados = 'sigla', formato_numeros = 'integer')`.
#' @template verboso
#' @template cache
#' @template n_cores
#'
#' @return Retorna o `data.frame` de input `enderecos` adicionado das colunas de
#'   latitude (`lat`) e longitude (`lon`), bem como as colunas (`precisao` e
#'   `tipo_resultado`) que indicam o nível de precisão e o tipo de resultado.
#'   Alternativamente, o resultado pode ser um objeto `sf`.
#'
#' @template precision_section
#' @template empates_section
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(geocodebr)
#'
#' # ler amostra de dados
#' data_path <- system.file("extdata/small_sample.csv", package = "geocodebr")
#' input_df <- read.csv(data_path)[1:2,]
#'
#' fields <- geocodebr::definir_campos(
#'   logradouro = "nm_logradouro",
#'   numero = "Numero",
#'   cep = "Cep",
#'   localidade = "Bairro",
#'   municipio = "nm_municipio",
#'   estado = "nm_uf"
#' )
#'
#' df <- geocodebr::geocode(
#'   enderecos = input_df,
#'   campos_endereco = fields,
#'   resolver_empates = TRUE
#'   )
#'
#' head(df)
#'
#' @export
geocode <- function(
  enderecos,
  campos_endereco = definir_campos(),
  resultado_completo = FALSE,
  resolver_empates = TRUE,
  resultado_sf = FALSE,
  h3_res = NULL,
  padronizar_enderecos = TRUE,
  verboso = TRUE,
  cache = TRUE,
  n_cores = NULL
) {
  callr::r(
    func = function(
      enderecos,
      campos_endereco,
      resultado_completo,
      resolver_empates,
      resultado_sf,
      h3_res,
      padronizar_enderecos,
      verboso,
      cache,
      n_cores
    ) {
      # Run internal engine
      geocode_core(
        enderecos = enderecos,
        campos_endereco = campos_endereco,
        resultado_completo = resultado_completo,
        resolver_empates = resolver_empates,
        resultado_sf = resultado_sf,
        h3_res = h3_res,
        padronizar_enderecos = padronizar_enderecos,
        verboso = verboso,
        cache = cache,
        n_cores = n_cores
      )
    },
    args = list(
      enderecos = enderecos,
      campos_endereco = campos_endereco,
      resultado_completo = resultado_completo,
      resolver_empates = resolver_empates,
      resultado_sf = resultado_sf,
      h3_res = h3_res,
      padronizar_enderecos = padronizar_enderecos,
      verboso = verboso,
      cache = cache,
      n_cores = n_cores
    ),
    show = TRUE,
    package = TRUE
  )
}


#' @keywords internal
geocode_core <- function(
  enderecos = parent.frame()$enderecos,
  campos_endereco = parent.frame()$campos_endereco,
  resultado_completo = parent.frame()$resultado_completo,
  resolver_empates = parent.frame()$resolver_empates,
  resultado_sf = parent.frame()$resultado_sf,
  h3_res = parent.frame()$h3_res,
  padronizar_enderecos = parent.frame()$padronizar_enderecos,
  verboso = parent.frame()$verboso,
  cache = parent.frame()$cache,
  n_cores = parent.frame()$n_cores
) {
  # ## ---- tiny timing toolkit (self-contained) ------------------------------
  # .make_timer <- function(verbose = TRUE) {
  #   .marks <- list()
  #   .t0_rt  <- proc.time()[["elapsed"]]     # monotonic wall clock
  #   .t_prev <- .t0_rt
  #
  #   fmt <- function(secs) sprintf("%.3f s", secs)
  #
  #   mark <- function(label) {
  #     now <- proc.time()[["elapsed"]]
  #     step  <- now - .t_prev
  #     total <- now - .t0_rt
  #     .marks <<- append(.marks, list(list(label = label, step = step, total = total)))
  #     .t_prev <<- now
  #     if (verbose) message(sprintf("[%s] +%s (total %s)", label, fmt(step), fmt(total)))
  #     invisible(now)
  #   }
  #
  #   summary <- function(print_summary = verbose) {
  #     if (length(.marks) == 0) return(invisible(data.frame()))
  #     df <- data.frame(
  #       step = vapply(.marks, `[[`, "", "label"),
  #       step_sec = vapply(.marks, `[[`, 0.0, "step"),
  #       total_sec = vapply(.marks, `[[`, 0.0, "total"),
  #       stringsAsFactors = FALSE
  #     ) |>
  #       dplyr::mutate(step_relative = round(step_sec / max(total_sec)*100, 1))
  #
  #     if (print_summary) {
  #       message("— Timing summary —")
  #       print(df, row.names = FALSE)
  #     }
  #     df
  #   }
  #
  #   time_it <- function(label, expr) {
  #     force(label)
  #     res <- eval.parent(substitute(expr))
  #     mark(label)
  #     invisible(res)
  #   }
  #
  #   list(mark = mark, summary = summary, time_it = time_it)
  # }
  # timer <- .make_timer(verbose = isTRUE(verboso))
  # on.exit(timer$summary(), add = TRUE)
  # ## -----------------------------------------------------------------------

  # check input
  checkmate::assert_data_frame(enderecos)
  checkmate::assert_logical(resultado_completo, any.missing = FALSE, len = 1)
  checkmate::assert_logical(resolver_empates, any.missing = FALSE, len = 1)
  checkmate::assert_logical(resultado_sf, any.missing = FALSE, len = 1)
  checkmate::assert_logical(padronizar_enderecos, any.missing = FALSE, len = 1)
  checkmate::assert_logical(verboso, any.missing = FALSE, len = 1)
  checkmate::assert_logical(cache, any.missing = FALSE, len = 1)
  checkmate::assert_numeric(
    h3_res,
    null.ok = TRUE,
    lower = 0,
    upper = 15,
    max.len = 16
  )

  # allow only letters, numbers, and underscore in colnames
  check_clean_colnames(enderecos)


  # systime start 66666 ----------------
  # timer$mark("Start")

  # normalize input data -------------------------------------------------------
  # standardizing the addresses table to increase the chances of finding a match
  # in the CNEFE data

  if (isTRUE(padronizar_enderecos)) {
    if (verboso) {
      message_standardizing_addresses()
    }

    campos_endereco <- assert_and_assign_address_fields(
      campos_endereco,
      enderecos
    )

    input_padrao <- enderecobr::padronizar_enderecos(
      enderecos = enderecos,
      campos_do_endereco = enderecobr::correspondencia_campos(
        logradouro = campos_endereco[["logradouro"]],
        numero = campos_endereco[["numero"]],
        cep = campos_endereco[["cep"]],
        bairro = campos_endereco[["localidade"]],
        municipio = campos_endereco[["municipio"]],
        estado = campos_endereco[["estado"]]
      ),
      formato_estados = "sigla",
      formato_numeros = 'integer'
    )
  }

  if (isFALSE(padronizar_enderecos)) {
    input_padrao <- data.table::copy(enderecos)
  }

  # checa se input foi mesmo padronizado
  all_cols_padr <- c(
    "estado_padr",
    "municipio_padr",
    "logradouro_padr",
    "numero_padr",
    "cep_padr",
    "bairro_padr"
  )
  check_padr <- all(all_cols_padr %in% names(input_padrao))

  if (isFALSE(check_padr)) {
    error_input_nao_padronizado()
  }

  # keep and rename colunms of input_padrao to use the
  # same column names used in cnefe data set
  data.table::setDT(input_padrao)
  cols_to_keep <- names(input_padrao)[names(input_padrao) %like% '_padr']
  input_padrao <- input_padrao[, .SD, .SDcols = c(cols_to_keep)]
  names(input_padrao) <- c(gsub("_padr", "", names(input_padrao)))

  if ('bairro' %in% names(input_padrao)) {
    data.table::setnames(
      x = input_padrao,
      old = 'bairro',
      new = 'localidade'
    )
  }

  # systime padronizacao 66666 ----------------
  # timer$mark("Padronizacao")

  # create temp id
  data.table::setDT(enderecos)[, tempidgeocodebr := 1:nrow(input_padrao)]
  input_padrao[, tempidgeocodebr := 1:nrow(input_padrao)]

  # temp coluna de logradouro q sera usada no match probabilistico
  input_padrao[, temp_lograd_determ := NA_character_]
  input_padrao[, similaridade_logradouro := NA_real_]

  # # sort input data
  # input_padrao <- input_padrao[order(estado, municipio, logradouro, numero, cep, localidade)]

  # downloading cnefe
  cnefe_dir <- download_cnefe(
    tabela = 'todas',
    verboso = verboso,
    cache = cache
  )

  # creating a temporary db and register the input table data
  con <- create_geocodebr_db(n_cores = n_cores)

  # register standardized input data
  input_padrao_arrw <- arrow::as_arrow_table(input_padrao)
  DBI::dbWriteTableArrow(
    con,
    name = "input_padrao_db",
    input_padrao_arrw,
    overwrite = TRUE,
    temporary = TRUE
  )

  # systime register standardized 66666 ----------------
  # timer$mark("Register standardized input")

  # cria coluna "log_causa_confusao" identificando logradouros que geram confusao
  # issue https://github.com/ipeaGIT/geocodebr/issues/67
  cria_col_logradouro_confusao(con)

  # create an empty output table that will be populated -----------------------------------------------

  # Define schema
  if (isFALSE(resultado_completo)) {

    schema_output_db <- arrow::schema(
      tempidgeocodebr = arrow::int32(),
      lat = arrow::float16(),
      # Equivalent to NUMERIC(8,6)
      lon = arrow::float16(),
      endereco_encontrado = arrow::string(),
      logradouro_encontrado = arrow::string(),
      tipo_resultado = arrow::string(),
      contagem_cnefe = arrow::int32(),
      desvio_metros = arrow::int32(),
      log_causa_confusao = arrow::boolean(),
      similaridade_logradouro = arrow::float16()
    )

  } else {
    schema_output_db <- arrow::schema(
      tempidgeocodebr = arrow::int32(),
      lat = arrow::float16(),
      # Equivalent to NUMERIC(8,6)
      lon = arrow::float16(),
      endereco_encontrado = arrow::string(),
      logradouro_encontrado = arrow::string(),
      tipo_resultado = arrow::string(),
      contagem_cnefe = arrow::int32(),
      desvio_metros = arrow::int32(),
      log_causa_confusao = arrow::boolean(),
      #
      numero_encontrado = arrow::int32(),
      localidade_encontrada = arrow::string(),
      cep_encontrado = arrow::string(),
      municipio_encontrado = arrow::string(),
      estado_encontrado = arrow::string(),
      similaridade_logradouro = arrow::float16(),
      cod_setor = arrow::string()
    )
  }

  output_db_arrow <- arrow::arrow_table(schema = schema_output_db)
  DBI::dbWriteTableArrow(
    con,
    name = "output_db",
    output_db_arrow,
    overwrite = TRUE,
    temporary = TRUE
  )

  # START MATCHING -----------------------------------------------

  # start progress bar
  if (verboso) {
    prog <- create_progress_bar(input_padrao)
    message_looking_for_matches()
  }

  n_rows <- nrow(input_padrao)
  matched_rows <- 0

  # start matching
  for (match_type in all_possible_match_types) {
    # get key cols
    key_cols <- get_key_cols(match_type)

    if (verboso) {
      update_progress_bar(matched_rows, match_type)
    }

    # somente busca essa categoria match_type se todas colunas estiverem na base
    # caso contrario, passa para proxima categoria
    if (all(key_cols %in% names(input_padrao))) {
      # select match function
      match_fun <-
        if (match_type %in% c(number_exact_types, exact_types_no_number)) {
          match_cases
        } else if (match_type %in% number_interpolation_types) {
          match_weighted_cases
        } else if (
          match_type %in%
            c(probabilistic_exact_types, probabilistic_types_no_number)
        ) {
          match_cases_probabilistic
        } else if (match_type %in% probabilistic_interpolation_types) {
          match_weighted_cases_probabilistic
        }

      n_rows_affected <- match_fun(
        con,
        match_type = match_type,
        key_cols = key_cols,
        resultado_completo = resultado_completo
      )

      matched_rows <- matched_rows + n_rows_affected

      # leave the loop early if we find all addresses before covering all cases
      if (matched_rows == n_rows) break
    }
  }

  if (verboso) {
    finish_progress_bar(matched_rows)
  }

  # systime matching 66666 ----------------
  # timer$mark("Matching")

  if (verboso) {
    message_preparando_output()
  }

  # casos de empate -----------------------------------------------

  empates_resolvidos <- trata_empates_geocode_duckdb(
    con,
    resultado_completo,
    resolver_empates,
    verboso
  )

  # systime resolve empates 66666 ----------------
  # timer$mark("Resolve empates")

  # bring original input back -----------------------------------------------

  # output with all original columns
  duckdb::dbWriteTable(
    con,
    "input_db",
    enderecos,
    temporary = TRUE,
    overwrite = TRUE
  )
  # enderecos_arrw <- arrow::as_arrow_table(enderecos)
  # DBI::dbWriteTableArrow(con, name = "input_db", enderecos_arrw,
  #                        overwrite = TRUE, temporary = TRUE)

  # systime write original input back 66666 ----------------
  # timer$mark("Write original input back")

  # add precision column ----------------
  output_table_to_use <- ifelse(
    empates_resolvidos == 0,
    'output_db',
    'output_db2'
  )
  add_precision_col(con, update_tb = output_table_to_use)

  # systime add precision 66666 ----------------
  # timer$mark("Add precision")

  x_columns <- names(enderecos)

  output_df <- merge_results_to_input(
    con,
    x = 'input_db',
    y = output_table_to_use,
    key_column = 'tempidgeocodebr',
    select_columns = x_columns,
    resultado_completo = resultado_completo
  )

  # systime merge results 66666 ----------------
  # timer$mark("Merge results")

  data.table::setDT(output_df)

  # drop geocodebr temp id column
  output_df[, tempidgeocodebr := NULL]

  # Disconnect from DuckDB when done
  duckdb::dbDisconnect(con)

  # add H3
  if (!is.null(h3_res)) {
    for (i in h3_res) {
      colname <- paste0(
        'h3_',
        formatC(h3_res, width = 2, flag = "0")
      )

      output_df[
        !is.na(lat),
        {{ colname }} := h3r::latLngToCell(lat = lat, lng = lon, resolution = i)
      ]
    }

    # systime add h3 66666 ----------------
    # timer$mark("Add H3")
  }

  # remove data.table class
  data.table::setindex(output_df, NULL)
  data.table::setDF(output_df)

  # convert df to simple feature
  if (isTRUE(resultado_sf)) {
    output_sf <- sfheaders::sf_point(
      obj = output_df,
      x = 'lon',
      y = 'lat',
      keep = TRUE
    )

    sf::st_crs(output_sf) <- 4674

    # systime convert to sf 66666 ----------------
    # timer$mark("Convert to sf")

    return(output_sf)
  }

  return(output_df[])
}

# skip tests because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("arrow")


data_path <- system.file("extdata/small_sample.csv", package = "geocodebr")
input_df <- read.csv(data_path)

campos <- geocodebr::definir_campos(
  logradouro = "nm_logradouro",
  numero = "Numero",
  cep = "Cep",
  localidade = "Bairro",
  municipio = "nm_municipio",
  estado = "nm_uf"
)

tester <- function(enderecos = input_df,
                   campos_endereco = campos,
                   resultado_completo = FALSE,
                   resolver_empates = FALSE,
                   resultado_sf = FALSE,
                   h3_res = NULL,
                   padronizar_enderecos = TRUE,
                   verboso = FALSE,
                   cache = TRUE,
                   n_cores = 1) {
  geocode(
    enderecos,
    campos_endereco,
    resultado_completo,
    resolver_empates,
    resultado_sf,
    h3_res,
    padronizar_enderecos,
    verboso,
    cache,
    n_cores
  )
}

test_that("expected output", {

  std_output <- tester()

  # find expected match cases
  match_types_found <- unique(std_output$tipo_resultado)
  testthat::expect_true(length(match_types_found) == 17)

  # full results
  full_output <- tester(resultado_completo = TRUE)


  all_cols_completas <- c("lat", "lon", "precisao", "tipo_resultado",
    "desvio_metros", "endereco_encontrado", "logradouro_encontrado",
    "numero_encontrado", "cep_encontrado", "localidade_encontrada",
    "municipio_encontrado", "estado_encontrado", "similaridade_logradouro",
    "contagem_cnefe", "empate", "cod_setor")

  testthat::expect_true(all(all_cols_completas %in% names(full_output)))

  # add h3
  std_output_h3 <- tester(h3_res = c(3,4))
  testthat::expect_true(all(c('h3_03', 'h3_04') %in% names(std_output_h3)))

  # output in sf format
  sf_output <- tester(resultado_sf = TRUE)
  testthat::expect_true(is(sf_output , 'sf'))
})


test_that("argumento padronizar endereco", {

  # erro se input nao estiver padronizado
  testthat::expect_error( tester(padronizar_enderecos = FALSE) )

  # sucesso se tiver sido padronizado
  campos2 <- enderecobr::correspondencia_campos(
    logradouro = "nm_logradouro",
    numero = "Numero",
    cep = "Cep",
    bairro = "Bairro",
    municipio = "nm_municipio",
    estado = "nm_uf"
  )

  input_padr <- enderecobr::padronizar_enderecos(
    enderecos = input_df,
    campos_do_endereco = campos2,
    formato_estados = "sigla",
    formato_numeros = 'integer'
    )

  testthat::succeed(
    tester(enderecos = input_padr, padronizar_enderecos = FALSE)
  )

})


test_that("test empates", {

  # com empates
  std_output <- tester(resolver_empates = FALSE)
  testthat::expect_true(nrow(std_output) > nrow(input_df))

  # resolvendo empates
  std_output <- tester(verboso = TRUE, resolver_empates = TRUE)
  testthat::expect_true(nrow(std_output) == nrow(input_df))

  # output ordenado como input
  testthat::expect_true( all(std_output$id == 1:nrow(input_df)) )

  # add h3
  std_output_h3 <- tester(h3_res = 3)
  testthat::expect_true('h3_03' %in% names(std_output_h3))
})

# test_that("test no messages", {
#
#   testthat::expect_no_message(
#     testthat::expect_warning(
#       std_output <- tester(verboso = FALSE,
#                            resolver_empates = FALSE)
#                                )
#     )
#
#    testthat::expect_no_message(
#     std_output <- tester(verboso = FALSE, resolver_empates = TRUE)
#                             )
# })


test_that("errors with incorrect input", {


  expect_error(tester(unclass(input_df)))

  # df com caracter especial nas colunas
  input_df2 <- input_df |> mutate("col!test" = 1)
  expect_error(tester(enderecos = input_df2))

  expect_error(tester(campos_endereco = 1))
  expect_error(tester(campos_endereco = c(hehe = "nm_logradouro")))
  expect_error(tester(campos_endereco = c(logradouro = "hehe")))

  expect_error(tester(resultado_completo = 1))
  expect_error(tester(resultado_completo = NA))
  expect_error(tester(resultado_completo = c(TRUE, TRUE)))

  expect_error(tester(resolver_empates = 1))
  expect_error(tester(resolver_empates = NA))
  expect_error(tester(resolver_empates = c(TRUE, TRUE)))

  expect_error(tester(resultado_sf = 1))
  expect_error(tester(resultado_sf = NA))
  expect_error(tester(resultado_sf = c(TRUE, TRUE)))

  expect_error(tester(padronizar_enderecos = "a"))
  expect_error(tester(padronizar_enderecos = 50))
  expect_error(tester(padronizar_enderecos = Inf))

  expect_error(tester(h3_res = "a"))
  expect_error(tester(h3_res = 50))
  expect_error(tester(h3_res = Inf))

  expect_error(tester(n_cores = "a"))
  expect_error(tester(n_cores = 0))
  expect_error(tester(n_cores = Inf))

  expect_error(tester(verboso = 1))
  expect_error(tester(verboso = NA))
  expect_error(tester(verboso = c(TRUE, TRUE)))

  expect_error(tester(cache = 1))
  expect_error(tester(cache = NA))
  expect_error(tester(cache = c(TRUE, TRUE)))
})

# possibilidade de usar dados da meta e overtura para compelemtar enderecos
# mas pelo visto os dados deles vem do cnefe ?
library(arrow)
# install_arrow()
library(sf)
library(dplyr)
library(geobr)
library(duckdb)



# install extension

con <- duckdb::dbConnect(duckdb::duckdb(), dbdir=":memory:")  # run on memory
DBI::dbExecute(con, "FORCE INSTALL httpfs")
DBI::dbExecute(con, "LOAD httpfs")


ext <- DBI::dbGetQuery(con, "FROM duckdb_extensions()")




muni <- geobr::lookup_muni(name_muni = 'Guarulhos')
muni <- geobr::read_municipality(code_muni = muni$code_muni)

bb <- sf::st_bbox(muni)

# https://cran.r-project.org/web/packages/overturemapsr/index.html

# https://walker-data.com/posts/overture-buildings/
# https://data.humdata.org/organization/meta

buildings <- open_dataset('s3://overturemaps-us-west-2/release/2024-05-16-beta.0/theme=buildings?region=us-west-2')

nrow(buildings)


df <- arrow::open_dataset('s3://overturemaps-us-west-2/release/2024-09-18.0')
names(df)


library(overturemapsr)

t <- overturemapsr::get_all_overture_types()

d <- overturemapsr::dataset_path(overture_type = 'locality')

df <- overturemapsr::record_batch_reader(overture_type = 'place', bbox = bb)

df$bbox <- NULL
df$sources <- NULL
df$names <- NULL
sapply(df, class)

library(ggalluvial)

temp <- temp_dfgeo |> select(id, tipo_resultado, precisao) |> unique()
# 37.9s
### viable sequence of match types for really large datasets
all_possible_match_types <- c(
  "dn01", "da01",
  "dn02", "da02",
  "dn03", "da03",
  "dn04", "da04",
  "pn01", "pa01", "pn02", "pa02", "pn03", "pa03", #"pn04", "pa04", # too costly
  "dl01",
  "dl02",
  "dl03",
  "dl04",         # pl04",  # too costly
  "pl01", "pl02", "pl03",
  "dc01", "dc02", "db01", "dm01"
)



temp2 <- temp_dfgeo2 |> select(id, tipo_resultado2 =tipo_resultado, precisao2 = precisao) |> unique()
# 30s
all_possible_match_types <- c(
  "dn01", "da01",
  "dn02", "da02",
  "dn03", "da03",
  "dn04", "da04",
  "pn01", "pa01", "pn02", "pa02", "pn03", "pa03", #"pn04", "pa04", # too costly
  "dl01",         "pl01",
  "dl02",         "pl02",
  "dl03",         "pl03",
  "dl04",         # pl04",  # too costly
  "dc01", "dc02", "db01", "dm01"
)


df <- left_join(temp, temp2)


table(df$precisao,df$precisao2 )

table(df$precisao )

table(df$precisao2 )

data.table::setDT(df)[, Freq := 1]

df_agreg <- df[, .(Freq = .N), by = .(tipo_resultado, tipo_resultado2, precisao, precisao2)]
df_agreg <- df_agreg[tipo_resultado != tipo_resultado2]
df_agreg <- df_agreg[precisao != precisao2]

library(ggplot2)
library(ggalluvial)
library(ggplotly)


gg <- ggplot(df_agreg,
             aes(y = Freq, axis1 = tipo_resultado, axis2 = tipo_resultado2)) +
  geom_alluvium(aes(fill = tipo_resultado2), width = 1/12) +
  geom_stratum(width = 1/12, fill = "white", color = "grey") +
  geom_label(stat = "stratum", aes(label = after_stat(stratum)), size=2) +
  scale_x_discrete(limits = c("antes", "depois"), expand = c(.05, .05)) +
  # scale_fill_brewer(type = "qual", palette = "Set1") +
  scale_fill_viridis_d()

gg
# plotly::ggplotly(gg)

ggsave(gg, filename = 'ggaluvia.png', width = 18, height = 12, units='cm')



# pn01            da04
#
# dfall <- left_join(dfgeo, dfgeo2, by='id') |>
#   filter(tipo_resultado.x=='pn01' &
#            tipo_resultado.y=='da04')
#
# dfall$id

filter(dfgeo,id %in% c(1371)  )
filter(dfgeo2,id %in% c(1371)  )
#' por que id '463' nao foi encontrado na cat da01 ? e sim no da02
#' em tese, nao deveria ter nenhuma transicao de p para d
#' outro exemplo: de pa01 para da03
#'  - ora, se foi pa01, encontrou bairro e cep
#'  - se depois achou com da03, entao achou com logradouro determ, e soh o bairro mas nao o cep
#'
#' outro caso pn01 e depois vira da04
#' pn03 e depois vira da04





############## sankey plot ----------------------------------------
library(ggplot2)
library(ggsankey)
library(dplyr)
library(data.table)

system.time(danid <- dani()) # 60.10
system.time(rafad <- rafa()) # 39.54


df <- left_join(select(danid, c('id', 'match_type')),
                select(rafad, c('id', 'match_type')), by='id')

data.table::setDT(df)
data.table::setnames(
  df,
  old = c('match_type.x', 'match_type.y'),
  new = c('case_dani', 'case_rafa') )


transition_mtrx <- round(table(df$case_dani, df$case_rafa) / nrow(df)*100, 1)


df[, case_dani := as.integer(gsub('case_', '', case_dani)) ]
df_count <- df[, .(count = round(.N / nrow(df)*100, 2)), by= .(case_dani, case_rafa)]

# df_long <- data.table::melt(data = df, id.vars='id')
# head(df_long)

# df_sankey <- ggsankey::make_long(df, case_dani, case_rafa)
df_sankey <- ggsankey::make_long(df_count, case_dani, case_rafa, value = count)
df_sankey$value <- round(df_sankey$value / nrow(df)*100, 2)
head(df_sankey)

#cats <- c(1:4, 44, 5:12)
cats <- c(12:5, 44, 4:1)
df_sankey$node <- factor(x = df_sankey$node,
                         levels = cats,
                         ordered = T)

df_sankey$next_node <- factor(x = df_sankey$next_node,
                              levels = cats,
                              ordered = T)

fig <- ggplot(data = df_sankey,
              aes(x = x,
                  next_x = next_x,
                  node = node,
                  next_node = next_node,
                  fill = factor(node)
                  , label = value
              )) +
  geom_sankey() +
  theme_sankey(base_size = 16)  + geom_sankey_label()

fig
plotly::ggplotly(fig)


library(plotly)

fig <- plot_ly(
  type = "sankey",
  orientation = "h",

  node = list(
    label = c("A1", "A2", "B1", "B2", "C1", "C2"),
    color = c("blue", "blue", "blue", "blue", "blue", "blue"),
    pad = 15,
    thickness = 20,
    line = list(
      color = "black",
      width = 0.5
    )
  ),

  link = list(
    source = c(0,1,0,2,3,3),
    target = c(2,3,3,4,4,5),
    value =  c(8,4,2,8,4,2)
  )
)
fig <- fig %>% layout(
  title = "Basic Sankey Diagram",
  font = list(
    size = 10
  )
)

fig





df_count <- df[, .(count = round(.N / nrow(df)*100, 2)), by= .(case_dani, case_rafa)]

df_sankey <- ggsankey::make_long(df, case_dani, case_rafa)

df_sankey <- left_join(df_sankey, df_count,
          by=c('node'='case_dani', 'next_node' = 'case_rafa'))


head(df_sankey)

#cats <- c(1:4, 44, 5:12)
cats <- c(12:5, 44, 4:1)
df_sankey$node <- factor(x = df_sankey$node,
                         levels = cats,
                         ordered = T)

df_sankey$next_node <- factor(x = df_sankey$next_node,
                              levels = cats,
                              ordered = T)
data.table::setDT(df_sankey)[, count2 := ifelse(is.na(count), lead(count), count)]
head(df_sankey)

fig <- ggplot(data = df_sankey,
              aes(x = x,
                  next_x = next_x,
                  node = node,
                  next_node = next_node,
                  fill = factor(node)
                  , label = count2
              )) +
  geom_sankey() +
  theme_sankey(base_size = 16)  + geom_sankey_label()

fig
plotly::ggplotly(fig)

# ===================================================================
# SCRIPT 2: BENCHMARK DE PERFORMANCE
#
# Objetivo: Comparar a velocidade de execução das funções de
# tratamento de empates em amostras de dados de diferentes tamanhos.
# ===================================================================

# --- 1. SETUP E CONFIGURAÇÃO ---
library(devtools)
library(dplyr)
library(data.table)
library(cli)
library(bench)
library(readr)

devtools::load_all()
source("R/trata_empates_geocode_duckdb.R")
# --- 2. DEFINIÇÃO DE PARÂMETROS E DIRETÓRIOS ---

output_dir <- "tests/tests_pedro"

# --- 3. PREPARAÇÃO DOS DADOS DE BENCHMARK ---
cli::cli_h1("Preparação dos dados")

cli_alert_info("Carregando e preparando amostras de dados...")

output_df <- readRDS(paste0(output_dir, "/output_df_large_sample.rds"))
output_df_pequeno <- readRDS(paste0(output_dir, "/output_df_small_sample.rds"))
set.seed(123)
output_df_grande <- output_df |> slice_sample(n = 500000, replace = TRUE)
con <- geocodebr:::create_geocodebr_db(n_cores = 4)

cli_alert_success("Amostras de dados prontas.")

# --- 4. EXECUÇÃO DO BENCHMARK ---
cli::cli_h1("Execução do benchmark de performance")

suppressWarnings(benchmark_grid_trat <- bench::press(
  nome = c("Amostra pequena (27 obs)", "Amostra média (22k obs)", "Amostra grande (500k obs)"),
  resolver = TRUE,
  {
    data_atual <- if (nome == "Amostra pequena (27 obs)") {
      output_df_pequeno
    } else if (nome == "Amostra média (22k obs)") {
      output_df
    } else output_df_grande
    status_txt <- if(resolver) "COM resolução" else "SEM resolução"
    cli::cli_progress_step("Benchmarking: {nome} / {status_txt}")

    bench::mark(
      Original = geocodebr:::trata_empates_geocode(data_atual, resolver_empates = resolver, verboso = FALSE),
      DuckDB = trata_empates_geocode_duckdb(data_atual, resolver_empates = resolver, verboso = FALSE, con = con),
      min_iterations = 10,
      check = FALSE
    )
  }
)
)

print(benchmark_grid_trat)

suppressWarnings(benchmark_grid_sem_trat <- bench::press(
  nome = c("Amostra pequena (27 obs)", "Amostra média (22k obs)", "Amostra grande (500k obs)"),
  resolver = FALSE,
  {
    data_atual <- if (nome == "Amostra pequena (27 obs)") {
      output_df_pequeno
    } else if (nome == "Amostra média (22k obs)") {
      output_df
    } else output_df_grande
    status_txt <- if(resolver) "COM resolução" else "SEM resolução"
    cli::cli_progress_step("Benchmarking: {nome} / {status_txt}")

    bench::mark(
      Original = geocodebr:::trata_empates_geocode(data_atual, resolver_empates = resolver, verboso = FALSE),
      DuckDB = trata_empates_geocode_duckdb(data_atual, resolver_empates = resolver, verboso = FALSE, con = con),
      min_iterations = 10,
      check = FALSE
    )
  }
)
)

print(benchmark_grid_sem_trat)

cli_alert_success("Benchmark concluído!")

# --- 5. ANÁLISE E EXPORTAÇÃO DOS RESULTADOS ---
cli::cli_h1("Exportando resultados dos benchmarks")

# Exporta a tabela de resultados para um arquivo CSV
output_csv_path <- file.path(output_dir, "benchmark_resultados.csv")
readr::write_csv(benchmark_grid, output_csv_path)
cli_alert_success("Resultados numéricos do benchmark salvos em: '{.path {output_csv_path}}'")

#' @param h3_res Um número que indica a resolução espacial da célula hexagonal
#'    H3 da localização dos pontos retornados. Também aceita um vetor de números,
#'    e.g. `c(8, 9)` Por padrão, é `NULL`. Detalhes sobre as resoluções disponíveis
#'    em \url{https://h3geo.org/docs/core-library/restable/}

#' Geolocaliza endereços no Brasil
#'
#' Geocodifica endereços brasileiros com base nos dados do CNEFE. Os endereços
#' de input devem ser passados como um `data.frame`, no qual cada coluna
#' descreve um campo do endereço (logradouro, número, cep, etc). Os resuldos dos
#' endereços geolocalizados podem seguir diferentes níveis de precisão. Consulte
#' abaixo a seção "Detalhes" para mais informações. As coordenadas de output
#' utilizam o sistema de coordenadas geográficas SIRGAS 2000, EPSG 4674.
#'
#' @param enderecos Um `data.frame`. Os endereços a serem geolocalizados. Cada
#'    coluna deve representar um campo do endereço.
#' @param campos_endereco Um vetor de caracteres. A correspondência entre cada
#'    campo de endereço e o nome da coluna que o descreve na tabela `enderecos`.
#'    A função [definir_campos()] auxilia na criação deste vetor e realiza
#'    algumas verificações nos dados de entrada. Campos de endereço passados
#'    como `NULL` serão ignorados, e a função deve receber pelo menos um campo
#'    não nulo, além  dos campos `"estado"` e `"municipio"`, que são
#'    obrigatórios. Note que o campo  `"localidade"` é equivalente a 'bairro'.
#' @param resultado_completo Lógico. Indica se o output deve incluir colunas
#'    adicionais, como o endereço encontrado de referência. Por padrão, é `FALSE`.
#' @param resolver_empates Lógico. Alguns resultados da geolocalização podem
#'    indicar diferentes coordenadas possíveis (e.g. duas ruas diferentes com o
#'    mesmo nome em uma mesma cidade). Esses casos são trados como 'empate' e o
#'    parâmetro `resolver_empates` indica se a função deve resolver esses empates
#'    automaticamente. Por padrão, é `FALSE`, e a função retorna apenas o caso
#'    mais provável. Para mais detalhes sobre como é feito o processo de
#'    desempate, consulte abaixo a seção "Detalhes".
#' @template h3_res
#' @template resultado_sf
#' @template verboso
#' @template cache
#' @template n_cores
#'
#' @return Retorna o `data.frame` de input `enderecos` adicionado das colunas de
#'   latitude (`lat`) e longitude (`lon`), bem como as colunas (`precisao` e
#'   `tipo_resultado`) que indicam o nível de precisão e o tipo de resultado.
#'   Alternativamente, o resultado pode ser um objeto `sf`.
#'
#' @template precision_section
#' @template empates_section
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(geocodebr)
#'
#' # ler amostra de dados
#' data_path <- system.file("extdata/small_sample.csv", package = "geocodebr")
#' input_df <- read.csv(data_path)[1:2,]
#'
#' fields <- geocodebr::definir_campos(
#'   logradouro = "nm_logradouro",
#'   numero = "Numero",
#'   cep = "Cep",
#'   localidade = "Bairro",
#'   municipio = "nm_municipio",
#'   estado = "nm_uf"
#' )
#'
#' df <- geocodebr::geocode(
#'   enderecos = input_df,
#'   campos_endereco = fields,
#'   resolver_empates = TRUE,
#'   verboso = FALSE
#'   )
#'
#' head(df)
#'
#' @export
geocode <- function(enderecos,
                    campos_endereco = definir_campos(),
                    resultado_completo = FALSE,
                    resolver_empates = FALSE,
                    h3_res = NULL,
                    resultado_sf = FALSE,
                    verboso = TRUE,
                    cache = TRUE,
                    n_cores = 1 ){


  ## ---- tiny timing toolkit (self-contained) ------------------------------
  .make_timer <- function(verbose = TRUE) {
    .marks <- list()
    .t0_rt  <- proc.time()[["elapsed"]]     # monotonic wall clock
    .t_prev <- .t0_rt

    fmt <- function(secs) sprintf("%.3f s", secs)

    mark <- function(label) {
      now <- proc.time()[["elapsed"]]
      step  <- now - .t_prev
      total <- now - .t0_rt
      .marks <<- append(.marks, list(list(label = label, step = step, total = total)))
      .t_prev <<- now
      if (verbose) message(sprintf("[%s] +%s (total %s)", label, fmt(step), fmt(total)))
      invisible(now)
    }

    summary <- function(print_summary = verbose) {
      if (length(.marks) == 0) return(invisible(data.frame()))
      df <- data.frame(
        step = vapply(.marks, `[[`, "", "label"),
        step_sec = vapply(.marks, `[[`, 0.0, "step"),
        total_sec = vapply(.marks, `[[`, 0.0, "total"),
        stringsAsFactors = FALSE
      ) |>
        dplyr::mutate(step_relative = round(step_sec / max(total_sec)*100, 1))

      if (print_summary) {
        message("— Timing summary —")
        print(df, row.names = FALSE)
      }
      df
    }

    time_it <- function(label, expr) {
      force(label)
      res <- eval.parent(substitute(expr))
      mark(label)
      invisible(res)
    }

    list(mark = mark, summary = summary, time_it = time_it)
  }
  timer <- .make_timer(verbose = isTRUE(verboso))
  on.exit(timer$summary(), add = TRUE)
  ## -----------------------------------------------------------------------


  # check input
  checkmate::assert_data_frame(enderecos)
  checkmate::assert_logical(resultado_completo, any.missing = FALSE, len = 1)
  checkmate::assert_logical(resolver_empates, any.missing = FALSE, len = 1)
  checkmate::assert_logical(resultado_sf, any.missing = FALSE, len = 1)
  checkmate::assert_logical(verboso, any.missing = FALSE, len = 1)
  checkmate::assert_logical(cache, any.missing = FALSE, len = 1)
  checkmate::assert_number(n_cores, lower = 1, finite = TRUE)
  checkmate::assert_number(h3_res, null.ok = TRUE, lower = 0, upper = 15)
  campos_endereco <- assert_and_assign_address_fields(
    campos_endereco,
    enderecos
  )

  # normalize input data -------------------------------------------------------

  # standardizing the addresses table to increase the chances of finding a match
  # in the CNEFE data

  if (verboso) message_standardizing_addresses()

  # systime start 66666 ----------------
  timer$mark("Start")

  input_padrao <- enderecobr::padronizar_enderecos(
    enderecos,
    campos_do_endereco = enderecobr::correspondencia_campos(
      logradouro = campos_endereco[["logradouro"]],
      numero = campos_endereco[["numero"]],
      cep = campos_endereco[["cep"]],
      bairro = campos_endereco[["localidade"]],
      municipio = campos_endereco[["municipio"]],
      estado = campos_endereco[["estado"]]
    ),
    formato_estados = "sigla",
    formato_numeros = 'integer'
  )


  # keep and rename colunms of input_padrao to use the
  # same column names used in cnefe data set
  data.table::setDT(input_padrao)
  cols_to_keep <- names(input_padrao)[ names(input_padrao) %like% '_padr']
  input_padrao <- input_padrao[, .SD, .SDcols = c(cols_to_keep)]
  names(input_padrao) <- c(gsub("_padr", "", names(input_padrao)))

  if ('bairro' %in% names(input_padrao)) {
    data.table::setnames(
      x = input_padrao, old = 'bairro', new = 'localidade'
    )
  }

  # systime padronizacao 66666 ----------------
  timer$mark("Padronizacao")


  # create temp id
  data.table::setDT(enderecos)[, tempidgeocodebr := 1:nrow(input_padrao) ]
  input_padrao[, tempidgeocodebr := 1:nrow(input_padrao) ]

  # temp coluna de logradouro q sera usada no match probabilistico
  input_padrao[, temp_lograd_determ := NA_character_ ]
  input_padrao[, similaridade_logradouro := NA_real_ ]

  # # sort input data
  # input_padrao <- input_padrao[order(estado, municipio, logradouro, numero, cep, localidade)]

  # downloading cnefe
  cnefe_dir <- download_cnefe(
    tabela = 'todas',
    verboso = verboso,
    cache = cache
  )

  # creating a temporary db and register the input table data
  con <- create_geocodebr_db(n_cores = n_cores)



  # register standardized input data
  input_padrao_arrw <- arrow::as_arrow_table(input_padrao)
  DBI::dbWriteTableArrow(con, name = "input_padrao_db", input_padrao_arrw,
                         overwrite = TRUE, temporary = TRUE)


  # systime register standardized 66666 ----------------
  timer$mark("Register standardized input")


  # cria coluna "log_causa_confusao" identificando logradouros que geram confusao
  # issue https://github.com/ipeaGIT/geocodebr/issues/67
  cria_col_logradouro_confusao(con)


  # systime cria coluna "log_causa_confusao 66666 ----------------
  timer$mark("Cria coluna log_causa_confusao")



  # create an empty output table that will be populated -----------------------------------------------

  # Define schema
  schema_output_db <- arrow::schema(
    tempidgeocodebr = arrow::int32(),
    lat = arrow::float16(),  # Equivalent to NUMERIC(8,6)
    lon = arrow::float16(),
    endereco_encontrado = arrow::string(),
    logradouro_encontrado = arrow::string(),
    tipo_resultado = arrow::string(),
    contagem_cnefe = arrow::int32(),
    desvio_metros = arrow::int32(),
    log_causa_confusao = arrow::boolean()
  )

  if (isTRUE(resultado_completo)) {

    schema_output_db <- arrow::schema(
      tempidgeocodebr = arrow::int32(),
      lat = arrow::float16(),  # Equivalent to NUMERIC(8,6)
      lon = arrow::float16(),
      endereco_encontrado = arrow::string(),
      logradouro_encontrado = arrow::string(),
      tipo_resultado = arrow::string(),
      contagem_cnefe = arrow::int32(),
      desvio_metros = arrow::int32(),
      log_causa_confusao = arrow::boolean(),
      #
      numero_encontrado = arrow::int32(),
      localidade_encontrada = arrow::string(),
      cep_encontrado = arrow::string(),
      municipio_encontrado = arrow::string(),
      estado_encontrado = arrow::string(),
      similaridade_logradouro = arrow::float16()
    )
  }

  output_db_arrow <- arrow::arrow_table(schema = schema_output_db)
  DBI::dbWriteTableArrow(con, name = "output_db", output_db_arrow,
                         overwrite = TRUE, temporary = TRUE)


  # START MATCHING -----------------------------------------------

  # start progress bar
  if (verboso) {
    prog <- create_progress_bar(input_padrao)
    message_looking_for_matches()
  }

  n_rows <- nrow(input_padrao)
  matched_rows <- 0

  # start matching
  for (match_type in all_possible_match_types ) {

    # get key cols
    key_cols <- get_key_cols(match_type)

    if (verboso) update_progress_bar(matched_rows, match_type)

    # somente busca essa categoria match_type se todas colunas estiverem na base
    # caso contrario, passa para proxima categoria
    if (all(key_cols %in% names(input_padrao))) {

      # select match function
      match_fun <-
        if (match_type %in% c(number_exact_types, exact_types_no_number)) { match_cases
        } else if (match_type %in% number_interpolation_types ) { match_weighted_cases
        } else if (match_type %in% c(probabilistic_exact_types, probabilistic_types_no_number)) { match_cases_probabilistic
        } else if (match_type %in% probabilistic_interpolation_types) { match_weighted_cases_probabilistic
        }

      n_rows_affected <- match_fun(
        con,
        match_type = match_type,
        key_cols = key_cols,
        resultado_completo = resultado_completo
      )

      matched_rows <- matched_rows + n_rows_affected

      # leave the loop early if we find all addresses before covering all cases
      if (matched_rows == n_rows) break
    }

  }

  if (verboso) finish_progress_bar(matched_rows)

  # systime matching 66666 ----------------
  timer$mark("Matching")


  if (verboso) message_preparando_output()

  # add precision column
  add_precision_col(con, update_tb = 'output_db')


  # systime add precision 66666 ----------------
  timer$mark("Add precision")


  # casos de empate -----------------------------------------------

  empates_resolvidos <- trata_empates_geocode_duckdb3(con, resolver_empates, verboso)



  # systime resolve empates 66666 ----------------
  timer$mark("Resolve empates")


  # bring original input back -----------------------------------------------

  # output with all original columns
  duckdb::dbWriteTable(con, "input_db", enderecos,
                       temporary = TRUE, overwrite=TRUE)
  # enderecos_arrw <- arrow::as_arrow_table(enderecos)
  # DBI::dbWriteTableArrow(con, name = "input_db", enderecos_arrw,
  #                        overwrite = TRUE, temporary = TRUE)


  # systime write original input back 66666 ----------------
  timer$mark("Write original input back")



  x_columns <- names(enderecos)

  output_df <- merge_results_to_input(
    con,
    x='input_db',
    y='output_db2',
    key_column='tempidgeocodebr',
    select_columns = x_columns,
    resultado_completo = resultado_completo
  )

  # systime merge results 66666 ----------------
  timer$mark("Merge results")


  data.table::setDT(output_df)

  # drop geocodebr temp id column
  output_df[, tempidgeocodebr := NULL]


  # Disconnect from DuckDB when done
  duckdb::dbDisconnect(con)



  if(isFALSE(resultado_completo)){ output_df[, logradouro_encontrado := NULL]}

  # add H3
  if( !is.null(h3_res) ) {

    colname <- paste0(
      'h3_',
      formatC(h3_res, width = 2, flag = "0")
    )

    output_df[!is.na(lat),
              {{colname}} := h3r::latLngToCell(lat = lat,
                                               lng = lon,
                                               resolution = h3_res)
    ]

    # systime add h3 66666 ----------------
    timer$mark("Add H3")

  }


  # convert df to simple feature
  if (isTRUE(resultado_sf)) {
    output_sf <- sfheaders::sf_point(
      obj = output_df,
      x = 'lon',
      y = 'lat',
      keep = TRUE
    )

    sf::st_crs(output_sf) <- 4674

    # systime convert to sf 66666 ----------------
    timer$mark("Convert to sf")

    return(output_sf)
  }

  return(output_df[])
}





# cad unico 1 milhao de casos

# resolve empate com duckdb2
#                           step step_sec total_sec
#                          Start     0.01      0.01
#                   Padronizacao    17.11     17.12
#    Register standardized input     2.19     19.31
# Cria coluna log_causa_confusao     0.05     19.36
#                       Matching    90.73    110.09
#                  Add precision     0.19    110.28
#                Resolve empates   338.43    448.71
#      Write original input back     0.40    449.11
#                  Merge results     2.57    451.68
#                         Add H3     1.91    453.59
#                  Convert to sf     2.28    455.87

# resolve empate com duckdb 3
#                           step step_sec total_sec
#                          Start     0.02      0.02
#                   Padronizacao    17.28     17.30
#    Register standardized input     2.13     19.43
# Cria coluna log_causa_confusao     0.04     19.47
#                       Matching    95.38    114.85
#                  Add precision     0.20    115.05
#                Resolve empates    10.33    125.38
#      Write original input back     0.39    125.77
#                  Merge results     3.11    128.88
#                         Add H3     1.64    130.52
#                  Convert to sf     2.30    132.82


# resolve empate com data.table
#                            step step_sec total_sec
#                           Start     0.04      0.04
#                    Padronizacao    19.32     19.36
#     Register standardized input     2.82     22.18
#  Cria coluna log_causa_confusao     0.08     22.26
#                        Matching   112.00    134.26
#                   Add precision     0.17    134.43
#       Write original input back     0.41    134.84
#                   Merge results     5.77    140.61
#                 Resolve empates    23.28    163.89
#                          Add H3     1.47    165.36
#                   Convert to sf     3.09    168.45

# ----------------------------------------------------------------------------

# NAO resolve empate com duckdb2
#                            step step_sec total_sec
#                           Start     0.01      0.01
#                    Padronizacao    18.21     18.22
#     Register standardized input     2.17     20.39
#  Cria coluna log_causa_confusao     0.05     20.44
#                        Matching   100.59    121.03
#                   Add precision     0.17    121.20
#                 Resolve empates     6.13    127.33
#       Write original input back     0.34    127.67
#                   Merge results     3.81    131.48
#                          Add H3     1.86    133.34
#                   Convert to sf     2.99    136.33


# NAO resolve empate com data.table
#                           step step_sec total_sec
#                          Start     0.02      0.02
#                   Padronizacao    18.03     18.05
#    Register standardized input     2.61     20.66
# Cria coluna log_causa_confusao     0.06     20.72
#                       Matching    98.80    119.52
#                  Add precision     0.17    119.69
#      Write original input back     0.36    120.05
#                  Merge results     4.73    124.78
#                Resolve empates     6.72    131.50
#                         Add H3     1.45    132.95
#                  Convert to sf     2.19    135.14




# resolve empate com data.table 3 (43 milhoes)
#                           step step_sec total_sec
#                          Start     0.02      0.02
#                   Padronizacao   433.21    433.23
#    Register standardized input    79.50    512.73
# Cria coluna log_causa_confusao     2.86    515.59
#                       Matching  2163.66   2679.25
#                  Add precision    87.53   2766.78
#      Write original input back    47.19   2813.97
#                  Merge results  1913.81   4727.78
#                Resolve empates  1302.81   6030.59
#                         Add H3    51.11   6081.70
#                  Convert to sf   422.74   6504.44


# resolve empate com duckdb 3 OLD (43 milhoes)
#
#                           step step_sec total_sec
#                          Start     0.00      0.00
#                   Padronizacao   432.72    432.72
#    Register standardized input    81.80    514.52
# Cria coluna log_causa_confusao     2.86    517.38
#                       Matching  2293.57   2810.95
#                  Add precision    88.21   2899.16
#                Resolve empates  3043.82   5942.98
#      Write original input back    67.85   6010.83
#                  Merge results   436.00   6446.83
#                         Add H3   479.51   6926.34
#                  Convert to sf   297.13   7223.47



# resolve empate com duckdb 3 NEW (43 milhoes)
#                           step step_sec total_sec step_relative
#                          Start     0.03      0.03           0.0
#                   Padronizacao   445.64    445.67           6.2
#    Register standardized input    80.77    526.44           1.1
# Cria coluna log_causa_confusao     2.56    529.00           0.0
#                       Matching  2402.55   2931.55          33.3
#                  Add precision   107.81   3039.36           1.5
#                Resolve empates  2677.19   5716.55          37.1
#      Write original input back    67.67   5784.22           0.9
#                  Merge results   526.64   6310.86           7.3
#                         Add H3   545.69   6856.55           7.6
#                  Convert to sf   362.96   7219.51           5.0



# resolve empate com duckdb 3 NEW (43 milhoes) precisao pra depois
#                           step step_sec total_sec step_relative
#                          Start     0.03      0.03           0.0
#                   Padronizacao   431.75    431.78           7.1
#    Register standardized input    79.56    511.34           1.3
# Cria coluna log_causa_confusao     2.68    514.02           0.0
#                       Matching  2077.87   2591.89          34.0
#                Resolve empates  2176.24   4768.13          35.6
#      Write original input back    57.73   4825.86           0.9
#                  Add precision    64.05   4889.91           1.0
#                  Merge results   413.50   5303.41           6.8
#                         Add H3   449.43   5752.84           7.4
#                  Convert to sf   352.99   6105.83           5.8

data_release <- "v0.4.1"

listar_pasta_cache_padrao <- function() {
  fs::path(
    tools::R_user_dir("geocodebr", which = "cache")
  )
}

listar_arquivo_config <- function() {
  fs::path(
    tools::R_user_dir("geocodebr", which = "config"),
    "cache_dir"
  )
}

#' Define um diretório de cache para o geocodebr
#'
#' Define um diretório de cache para os dados do geocodebr. Essa configuração
#' é persistente entre sessões do R.
#'
#' @param path Uma string. O caminho para o diretório usado para armazenar os
#'   dados em cache.  Se `NULL`, o pacote usará um diretório versionado salvo
#'   dentro do diretório retornado por [tools::R_user_dir()].
#' @template verboso
#'
#' @return Retorna de forma invisível o caminho do diretório de cache.
#'
#' @examples
#' definir_pasta_cache(tempdir())
#'
#' # retoma pasta padrão do pacote
#' definir_pasta_cache( path = NULL)
#'
#' @export
definir_pasta_cache <- function(path, verboso = TRUE) {
  checkmate::assert_string(path, null.ok = TRUE)
  checkmate::assert_logical(verboso, null.ok = FALSE)

  if (is.null(path)) {
    cache_dir <- listar_pasta_cache_padrao()
  } else {
    cache_dir <- fs::path_norm(path)
  }

  if (isTRUE(verboso)) {
    cli::cli_inform(
      c("i" = "Definido como pasta de cache {.file {cache_dir}}."),
      class = "geocodebr_cache_dir"
    )
  }

  arquivo_config <- listar_arquivo_config()

  if (!fs::file_exists(arquivo_config)) {
    fs::dir_create(fs::path_dir(arquivo_config))
    fs::file_create(arquivo_config)
  }

  cache_dir <- as.character(cache_dir)

  writeLines(cache_dir, con = arquivo_config)

  return(invisible(cache_dir))
}


#' Obtém a pasta de cache usado no geocodebr
#'
#' Obtém o caminho da pasta utilizada para armazenar em cache os dados do
#' geocodebr. Útil para inspecionar a pasta configurada com [definir_pasta_cache()]
#' em uma sessão anterior do R. Retorna a pasta de cache padrão caso nenhuma
#' pasta personalizado tenha sido configurada anteriormente.
#'
#' @return O caminho da pasta de cache.
#'
#' @examples
#' listar_pasta_cache()
#'
#' @export
listar_pasta_cache <- function() {
  arquivo_config <- listar_arquivo_config()

  if (fs::file_exists(arquivo_config)) {
    cache_dir <- readLines(arquivo_config)
    cache_dir <- fs::path_norm(cache_dir)
  } else {
    cache_dir <- listar_pasta_cache_padrao()
  }

  cache_dir <- as.character(cache_dir)

  return(cache_dir)
}

#' Listar dados em cache
#'
#' Lista os dados salvos localmente na pasta de cache
#'
#' @param print_tree Um valor lógico. Indica se o conteúdo da pasta de cache
#'   deve ser exibido em um formato de árvore. O padrão é `FALSE`.
#'
#' @return O caminho para os arquivos em cache
#'
#' @examples
#' listar_dados_cache()
#'
#' listar_dados_cache(print_tree = TRUE)
#'
#' @export
listar_dados_cache <- function(print_tree = FALSE) {
  checkmate::assert_logical(print_tree, any.missing = FALSE, len = 1)

  cache_dir <- listar_pasta_cache()

  if (!fs::dir_exists(cache_dir)) {
    message_cache()
    return(NULL)
  }

  cached_data <- list.files(cache_dir, recursive = TRUE, full.names = TRUE)

  if (print_tree) {
    fs::dir_tree(cache_dir)
    return(invisible(cached_data))
  }

  return(cached_data)
}


#' Deletar pasta de cache do geocodebr
#'
#' Deleta todos arquivos da pasta do cache.
#'
#' @return Retorna de forma invisível o caminho do diretório de cache.
#'
#' @examplesIf identical(TRUE, FALSE)
#' deletar_pasta_cache()
#'
#' @export
deletar_pasta_cache <- function() {
  cache_dir <- listar_pasta_cache()

  unlink(cache_dir, recursive = TRUE)

  message_removed_cache_dir(cache_dir)

  return(invisible(cache_dir))
}

message_removed_cache_dir <- function(cache_dir) {
  geocodebr_message(
    c(
      "v" = "Deletada a pasta de cache que se encontrava em {.path {cache_dir}}."
    )
  )
}


#' Atualiza dados no release local
#'
#' Detecta automaticamente se o release local esta desatualizado. Se sim, apaga
#' a pasta de cache do release local e atualiza a versao do release no caminho
#' da pasta
#'
#' @return Retorna de forma invisível o caminho do diretório de cache.
#'
#' @keywords internal
apaga_data_release_antigo <- function() {

  # list cache local
  cache_dir <- geocodebr::listar_pasta_cache()

  # detect all release paths
  local_release_path <- list.dirs(cache_dir, recursive = T)[-1]
  local_release_path <- local_release_path[grep(
    'geocodebr_data_release_',
    local_release_path
  )]

  new_data_release_dir <- fs::path(
    cache_dir,
    glue::glue("geocodebr_data_release_{data_release}")
  )

  if (identical(as.character(new_data_release_dir), local_release_path)) {
    return(cache_dir)
  }

  # versao numerica dos releases local e do pacote
  local_release <- gsub("[^0-9]", "", basename(local_release_path)) |>
    as.numeric()
  pkg_release <- gsub("[^0-9]", "", data_release) |> as.numeric()

  if (length(local_release) == 0) {
    return(cache_dir)
  }

  check1 <- is.na(local_release)
  check2 <- local_release == pkg_release
  if (all(check1, check2)) {
    return(cache_dir)
  }

  if (any(local_release != pkg_release)) {
    # deleta os dados do release local
    suppressMessages(
      geocodebr::deletar_pasta_cache()
    )

    return(cache_dir)
  }
}

# skip tests because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("arrow")

cache_config_file <- geocodebr:::listar_arquivo_config()
default_cache_dir <- geocodebr:::listar_pasta_cache_padrao()

# definir_pasta_cache -----------------------------------------------------------

tester <- function(path = NULL) definir_pasta_cache(path)

test_that("errors with incorrect input", {
  expect_error(tester(1))
  expect_error(tester(c("aaa", "bbb")))
  expect_error(tester(path))

})

test_that("behaves correctly", {
  # if the cache config file exists, we save its current content just to make
  # sure our tests don't disturb any workflows we have. if it doesn't, we delete
  # the file we created during the test

  if (fs::file_exists(cache_config_file)) {
    config_file_content <- readLines(cache_config_file)
    on.exit(writeLines(config_file_content, cache_config_file), add = TRUE)
  } else {
    on.exit(fs::file_delete(cache_config_file), add = TRUE)
  }

  # by default uses a versioned dir inside the default R cache dir

  fn_result <- suppressMessages(definir_pasta_cache(path = NULL))
  expect_type(fn_result, "character")
  expect_identical(fn_result, as.character(default_cache_dir))
  expect_identical(readLines(cache_config_file), unclass(default_cache_dir))

  fn_result <- suppressMessages(definir_pasta_cache("aaa"))
  expect_type(fn_result, "character")
  expect_identical(fn_result, "aaa")
  expect_identical(readLines(cache_config_file), "aaa")
})

test_that("messages are formatted correctly", {
  if (fs::file_exists(cache_config_file)) {
    config_file_content <- readLines(cache_config_file)
    on.exit(writeLines(config_file_content, cache_config_file), add = TRUE)
  } else {
    on.exit(fs::file_delete(cache_config_file), add = TRUE)
  }

  expect_snapshot(
    definir_pasta_cache(path = NULL),
    transform = function(x) sub(default_cache_dir, "<path_to_default_dir>", x),
    cnd_class = TRUE
  )

  expect_snapshot(definir_pasta_cache("aaa"), cnd_class = TRUE)
})

# listar_pasta_cache -----------------------------------------------------------

test_that("behaves correctly", {
  # if the cache config file exists, we save its current content just to make
  # sure our tests don't disturb any workflows we have. if it doesn't, we delete
  # the file we created during the test

  if (fs::file_exists(cache_config_file)) {
    config_file_content <- readLines(cache_config_file)
    on.exit(writeLines(config_file_content, cache_config_file), add = TRUE)
  } else {
    on.exit(fs::file_delete(cache_config_file), add = TRUE)
  }

  # if the cache config file exists, return its content. otherwise, returns the
  # default cache dir

  if (fs::file_exists(cache_config_file)) fs::file_delete(cache_config_file)
  expect_identical(listar_pasta_cache(), as.character(default_cache_dir))

  writeLines("aaa", cache_config_file)
  expect_identical(listar_pasta_cache(), "aaa")
})

# listar_dados_cache --------------------------------------------------------

test_that("errors with incorrect input", {
  expect_error(listar_dados_cache(1))
  expect_error(listar_dados_cache(NA))
  expect_error(listar_dados_cache(c(TRUE, TRUE)))
})

test_that("behaves correctly", {
  # if the cache config file exists, we save its current content just to make
  # sure our tests don't disturb any workflows we have. if it doesn't, we delete
  # the file we created during the test

  if (fs::file_exists(cache_config_file)) {
    config_file_content <- readLines(cache_config_file)
    on.exit(writeLines(config_file_content, cache_config_file), add = TRUE)
  } else {
    on.exit(fs::file_delete(cache_config_file), add = TRUE)
  }

  # we set the cache dir to a temporary directory to not mess with any cached
  # data we may already have

  tmpdir <- tempfile()
  fs::dir_create(tmpdir)

  suppressMessages(definir_pasta_cache(tmpdir))

  expect_identical(listar_dados_cache(), character(0))

  # previously, we used download_cnefe(progress = FALSE) here to download cnefe
  # data before listing the content. however, this takes a long time, and
  # afterall we only need to make sure that the function lists whatever files we
  # have in the directory. so we create empty temp files to test if the function
  # is working

  file.create(fs::path(tmpdir, c("oie.parquet", "hello.parquet")))

  cnefe_files <- listar_dados_cache()
  expect_identical(basename(cnefe_files), c("hello.parquet", "oie.parquet"))

  # expect a tree-like message and invisible value when print_tree=TRUE

  expect_snapshot(
    listar_dados_cache(print_tree = TRUE),
    transform = function(x) sub(listar_pasta_cache(), "<path_to_cache_dir>", x)
  )
})

# deletar_pasta_cache ---------------------------------------------------------

test_that("deletar_pasta_cache behaves correctly", {
  # if the cache config file exists, we save its current content just to make
  # sure our tests don't disturb any workflows we have. if it doesn't, we delete
  # the file we created during the test

  if (fs::file_exists(cache_config_file)) {
    config_file_content <- readLines(cache_config_file)
    on.exit(writeLines(config_file_content, cache_config_file), add = TRUE)
  } else {
    on.exit(fs::file_delete(cache_config_file), add = TRUE)
  }

  # we set the cache dir to a temporary directory to not mess with any cached
  # data we may already have

  tmpdir <- tempfile()
  fs::dir_create(tmpdir)

  suppressMessages(definir_pasta_cache(tmpdir))

  file.create(fs::path(tmpdir, "oie.parquet"))
  expect_identical(basename(listar_dados_cache()), "oie.parquet")

  expect_snapshot(
    res <- deletar_pasta_cache(),
    cnd_class = TRUE,
    transform = function(x) sub(listar_pasta_cache(), "<path_to_cache_dir>", x)
  )

  # mensagem se pasta de cache estiver vazia
  expect_message( geocodebr::listar_dados_cache() )

  expect_identical(res, as.character(fs::path_norm(tmpdir)))
  expect_false(dir.exists(res))


})

#' Safely use arrow to open a Parquet file
#'
#' This function handles some failure modes, including if the Parquet file is
#' corrupted.
#'
#' @param filename A local Parquet file
#' @return An `arrow::Dataset`
#'
#' @keywords internal
arrow_open_dataset <- function(filename) {
  # nocov start

  tryCatch(
    arrow::open_dataset(filename, format = 'parquet'),
    error = function(e) {
      msg <- paste(
        "Arquivo local possivelmente corrompido. ",
        "Apague os arquivos do cache com 'geocodebr::deletar_pasta_cache()' e tente novamente.",
        sep = "\n"
      )
      stop(msg)
    }
  )
} # nocov end

#' Message when caching file
#'
#' @param local_file The address of a file passed from the download_file function.
#' @param cache Logical. Whether the cached data should be used.

#' @return A message
#'
#' @keywords internal
cache_message <- function(
  local_file = parent.frame()$local_file,
  cache = parent.frame()$cache
) {
  # nocov start

  # name of local file
  file_name <- basename(local_file[1])
  dir_name <- dirname(local_file[1])

  ## if file already exists
  # YES cache
  if (file.exists(local_file) & isTRUE(cache)) {
    message('Reading data cached locally.')
  }

  # NO cache
  if (file.exists(local_file) & isFALSE(cache)) {
    message('Overwriting data cached locally.')
  }

  ## if file does not exist yet
  # YES cache
  if (!file.exists(local_file) & isTRUE(cache)) {
    message(paste("Downloading data and storing it locally for future use."))
  }

  # NO cache
  if (!file.exists(local_file) & isFALSE(cache)) {
    message(paste(
      "Downloading data. Setting 'cache = TRUE' is strongly recommended to speed up future use. File will be stored locally at:",
      dir_name
    ))
  }
} # nocov end


#' Update input_padrao_db to remove observations previously matched
#'
#' @param con A db connection
#' @param update_tb String. Name of a table to be updated in con
#' @param reference_tb A table written in con used as reference
#'
#' @return Drops observations from input_padrao_db
#'
#' @keywords internal
update_input_db <- function(con, update_tb = 'input_padrao_db', reference_tb) {
  # nocov start

  # update_tb = 'input_padrao_db'
  # reference_tb = 'output_caso_1'

  query_remove_matched <- glue::glue(
    "DELETE FROM {update_tb}
     WHERE tempidgeocodebr IN (
      SELECT tempidgeocodebr
      FROM {reference_tb}
    );"
  )

  DBI::dbExecute(con, query_remove_matched)
} # nocov end


#' Add a column with info of geocode match_type
#'
#' @param con A db connection
#' @param update_tb String. Name of a table to be updated in con
#'
#' @return Adds a new column to a table in con
#'
#' @keywords internal
add_precision_col <- function(con, update_tb = NULL) {
  # nocov start

  # update_tb = "output_db"

  # add empty column
  query_add_col <- glue::glue(
    "ALTER TABLE {update_tb} ADD COLUMN precisao TEXT;"
  )
  DBI::dbExecute(con, query_add_col)

  # populate column
  query_precision_cats <- glue::glue(
    "
  UPDATE {update_tb}
  SET precisao = CASE
  WHEN tipo_resultado IN ('dn01', 'dn02', 'dn03', 'dn04',
                          'pn01', 'pn02', 'pn03', 'pn04') THEN 'numero'
  WHEN tipo_resultado IN ('da01', 'da02', 'da03', 'da04',
                          'pa01', 'pa02', 'pa03', 'pa04') THEN 'numero_aproximado'
  WHEN tipo_resultado IN ('dl01', 'dl02', 'dl03', 'dl04',
                          'pl01', 'pl02', 'pl03', 'pl04') THEN 'logradouro'
  WHEN tipo_resultado IN ('dc01', 'dc02') THEN 'cep'
  WHEN tipo_resultado = 'db01' THEN 'localidade'
  WHEN tipo_resultado = 'dm01' THEN 'municipio'
  ELSE NULL
  END;"
  )

  # DBI::dbExecute(con, query_precision_cats )
  DBI::dbExecute(con, query_precision_cats)
} # nocov end


merge_results_to_input <- function(
  con,
  x,
  y,
  key_column,
  select_columns,
  resultado_completo
) {
  # nocov start

  select_columns_y <- c(
    'lat',
    'lon',
    'precisao',
    'tipo_resultado',
    'desvio_metros',
    'endereco_encontrado'
  )

  if (isTRUE(resultado_completo)) {
    # select additional columns to output
    select_columns_y <- c(
      select_columns_y,
      'logradouro_encontrado',
      'numero_encontrado',
      'cep_encontrado',
      'localidade_encontrada',
      'municipio_encontrado',
      'estado_encontrado',
      'similaridade_logradouro',
      'contagem_cnefe',
      'empate',
      'cod_setor'
    )

    # relace NULL similaridade_logradouro as 1 because they were found deterministically
    DBI::dbExecute(
      con,
      glue::glue(
        "UPDATE {y}
      SET similaridade_logradouro = COALESCE(similaridade_logradouro, 1);"
      )
    )
  }

  # Create the SELECT clause dynamically
  select_x <- paste0(x, '.', c(select_columns), collapse = ', ')

  select_clause <- paste0(
    select_x,
    ',',
    paste0(glue::glue('{y}'), ".", select_columns_y, collapse = ", ")
  )

  # Create the JOIN clause dynamically
  join_condition <- paste(
    glue::glue("{x}.{key_column} = {y}.{key_column}"),
    collapse = ' ON '
  )

  # Create SQL query
  query <- glue::glue(
    "SELECT {select_clause}
        FROM {x}
        LEFT JOIN {y}
        ON {join_condition}
      ORDER BY
        {x}.tempidgeocodebr;"
  )

  # Execute the query and fetch the merged data
  merged_data <- DBI::dbGetQuery(con, query)

  return(merged_data)
} # nocov end


#' create index
#'
#' @keywords internal
create_index <- function(con, tb, cols, operation, overwrite = TRUE) {
  # nocov start

  idx <- paste0('idx_', tb)
  cols_group <- paste(cols, collapse = ", ")

  # check if table already has index
  i <- DBI::dbGetQuery(
    con,
    sprintf("SELECT * FROM duckdb_indexes WHERE table_name = '%s';", tb)
  )

  if (nrow(i) > 0 & isFALSE(overwrite)) {
    return(NULL)
  }
  if (nrow(i) > 0 & isTRUE(overwrite)) {
    DBI::dbExecute(con, sprintf('DROP INDEX IF EXISTS %s', idx))
  }

  query_index <- sprintf(
    '%s INDEX %s ON %s(%s);',
    operation,
    idx,
    tb,
    cols_group
  )
  DBI::dbExecute(con, query_index)
} # nocov end


get_key_cols <- function(match_type) {
  # nocov start
  relevant_cols <- if (match_type %in% c('dn01', 'da01', 'pn01', 'pa01')) {
    c("estado", "municipio", "logradouro", "numero", "cep", "localidade")
  } else if (match_type %in% c('dn02', 'da02', 'pn02', 'pa02')) {
    c("estado", "municipio", "logradouro", "numero", "cep")
  } else if (match_type %in% c('dn03', 'da03', 'pn03', 'pa03')) {
    c("estado", "municipio", "logradouro", "numero", "localidade")
  } else if (match_type %in% c('dn04', 'da04', 'pn04', 'pa04')) {
    c("estado", "municipio", "logradouro", "numero")
  } else if (match_type %in% c('dl01', 'pl01')) {
    c("estado", "municipio", "logradouro", "cep", "localidade")
  } else if (match_type %in% c('dl02', 'pl02')) {
    c("estado", "municipio", "logradouro", "cep")
  } else if (match_type %in% c('dl03', 'pl03')) {
    c("estado", "municipio", "logradouro", "localidade")
  } else if (match_type %in% c('dl04', 'pl04')) {
    c("estado", "municipio", "logradouro")
  } else if (match_type == 'dc01') {
    c("estado", "municipio", "cep", "localidade")
  } else if (match_type == 'dc02') {
    c("estado", "municipio", "cep")
  } else if (match_type == 'db01') {
    c("estado", "municipio", "localidade")
  } else if (match_type == 'dm01') {
    c("estado", "municipio")
  }

  return(relevant_cols)
} # nocov end

### ideal sequence of match types
all_possible_match_types <- c(
  "dn01",
  "da01",
  "pn01",
  "pa01",
  "dn02",
  "da02",
  "pn02",
  "pa02",
  "dn03",
  "da03",
  "pn03",
  "pa03",
  "dn04",
  "da04", #"pn04", "pa04", # too costly
  "dl01",
  "pl01",
  "dl02",
  "pl02",
  "dl03",
  "pl03",
  "dl04", # pl04",  # too costly
  "dc01",
  "dc02",
  "db01",
  "dm01"
)

# ### 2nd best viable sequence of match types for really large datasets ? testando com cadunico
# all_possible_match_types <- c(
#   "dn01", "da01",
#   "dn02", "da02",
#   "dn03", "da03",
#   "dn04", "da04",
#   "pn01", "pa01", "pn02", "pa02", "pn03", "pa03", #"pn04", "pa04", # too costly
#   "dl01",         "pl01",
#   "dl02",         "pl02",
#   "dl03",         "pl03",
#   "dl04",         # pl04",  # too costly
#   "dc01", "dc02", "db01", "dm01"
# )

number_exact_types <- c(
  "dn01",
  "dn02",
  "dn03",
  "dn04"
)

number_interpolation_types <- c(
  "da01",
  "da02",
  "da03",
  "da04"
)

probabilistic_exact_types <- c(
  "pn01",
  "pn02",
  "pn03",
  "pn04"
)

probabilistic_interpolation_types <- c(
  "pa01",
  "pa02",
  "pa03",
  "pa04"
)

exact_types_no_number <- c(
  "dl01",
  "dl02",
  "dl03",
  "dl04",
  "dc01",
  "dc02",
  "db01",
  "dm01"
)

probabilistic_types_no_number <- c(
  "pl01",
  "pl02",
  "pl03",
  "pl04"
)

exact_types__no_logradouro <- c(
  "dc01",
  "dc02",
  "db01",
  "dm01"
)


assert_and_assign_address_fields <- function(address_fields, addresses_table) {
  # nocov start
  possible_fields <- c(
    "logradouro",
    "numero",
    "cep",
    "localidade",
    "municipio",
    "estado"
  )

  col <- checkmate::makeAssertCollection()
  checkmate::assert_names(
    names(address_fields),
    type = "unique",
    subset.of = possible_fields,
    add = col
  )
  checkmate::assert_names(
    address_fields,
    subset.of = names(addresses_table),
    add = col
  )
  checkmate::reportAssertions(col)

  missing_fields <- setdiff(possible_fields, names(address_fields))

  missing_fields_list <- vector(mode = "list", length = length(missing_fields))
  names(missing_fields_list) <- missing_fields

  complete_fields_list <- append(as.list(address_fields), missing_fields_list)

  return(complete_fields_list)
} # nocov end


get_reference_table <- function(match_type) {
  # nocov start

  # key_cols = get_key_cols('da03')

  key_cols <- get_key_cols(match_type)

  # read corresponding parquet file
  table_name <- paste(key_cols, collapse = "_")
  table_name <- gsub('estado_municipio', 'municipio', table_name)

  # reference table
  if (match_type %like% 'dn02|pn02|da02|pa02|dn03|pn03') {
    table_name <- "municipio_logradouro_numero_cep_localidade"
  }

  if (match_type %like% 'da03|pa03|dn04|da04') {
    table_name <- "municipio_logradouro_numero_localidade"
  }

  if (match_type %like% 'dl02|pl02|dl03|pl03') {
    table_name <- "municipio_logradouro_cep_localidade"
  }

  if (match_type %like% 'dl04') {
    table_name <- "municipio_logradouro_localidade"
  }

  return(table_name)
} # nocov end


# min cutoff for string match
# min cutoff for probabilistic string match of logradouros
get_prob_match_cutoff <- function(match_type) {
  # nocov start
  min_cutoff <- ifelse(match_type %in% c('pn01', 'pa01', 'pl01'), 0.85, 0.9)
  return(min_cutoff)
} # nocov end


# create a dummy function that uses nanoarrow with no effect
# nanoarrow is only used internally in DBI::dbWriteTableArrow()
# however, if we do not put this dummy function here, CRAN check flags an error
dummy <- function() {
  # nocov start
  nanoarrow::as_nanoarrow_schema
} # nocov end


# Cria coluna dummy no input padronizado identificando se logradouro é daqueles
# que gera confusao (e.g. uma letra (e.g. RUA A, RUA B, RUA C, ....) ou compostos
# só por dígitos (RUA 1, RUA 10, RUA 20, ...))
cria_col_logradouro_confusao <- function(con) {
  # nocov start

  # Add the column with default 0 (avoids updating all rows later)
  DBI::dbExecute(
    con,
    "ALTER TABLE input_padrao_db
      ADD COLUMN log_causa_confusao BOOLEAN DEFAULT false;"
  )

  # Ambiguos numero por extenso
  ruas_num_ext <- paste(
    paste(
      "RUA",
      c(
        'UM',
        'DOIS',
        'TRES',
        'CINCO',
        'SEIS',
        'SETE',
        'OITO',
        'NOVE',
        'DEZ',
        'ONZE',
        'DOZE',
        'TREZE'
      )
    ),
    collapse = "|"
  )
  ruas_num_ext <- paste0("(", ruas_num_ext, ")$")

  # 2) Flip to 1 for rows matching our regex
  DBI::dbExecute(
    con,
    glue::glue(
      r"{UPDATE input_padrao_db
    SET log_causa_confusao = true
    WHERE
      (REGEXP_MATCHES(logradouro, '^(RUA|TRAVESSA|RAMAL|BECO|BLOCO|AVENIDA|RODOVIA|ESTRADA)\s+([A-Z]{{1,2}}-?|[0-9]{{1,3}}|[A-Z]{{1,2}}-?[0-9]{{1,3}}|[A-Z]{{1,2}}\s+[0-9]{{1,3}}|[0-9]{{1,3}}-?[A-Z]{{1,2}})(\s+KM( \d+)?)?$')
       OR REGEXP_MATCHES(logradouro, '{ruas_num_ext}')
       )
        -- ainda dah pra salvar enderecos com datas (e.g. 'RUA 15 DE NOVEMBRO')
        AND NOT REGEXP_MATCHES(logradouro, '\bDE (JANEIRO|FEVEREIRO|MARCO|ABRIL|MAIO|JUNHO|JULHO|AGOSTO|SETEMBRO|OUTUBRO|NOVEMBRO|DEZEMBRO)\b');}"
    )
  )
} # nocov end



check_clean_colnames <- function(df) {  # nocov start

  # # basic input check
  # if (!is.data.frame(df)) {
  #   cli::cli_abort("{.arg df} must be a data.frame.")
  # }

  cols <- colnames(df)

  # allow only letters, numbers, and underscore
  bad_cols <- cols[!grepl("^[A-Za-z0-9_]+$", cols)]

  if (length(bad_cols) > 0) {
    cli::cli_abort(c(
      "Invalid column names detected.",
      "x" = "Column names must use only letters, numbers, and underscores ({.val _}).",
      "i" = "Please rename these columns: {.val {bad_cols}}"
    ),
    "call" = rlang::caller_env()
    )
  }

} # nocov end

#' Busca por CEP
#'
#' Busca endereços e suas coordenadas geográficas a partir de um CEP. As
#' coordenadas de output utilizam o sistema de coordenadas geográficas SIRGAS
#' 2000, EPSG 4674.
#'
#' @param cep Vetor. Um CEP ou um vetor de CEPs com 8 dígitos.
#' @template h3_res
#' @template resultado_sf
#' @template verboso
#' @template cache
#'
#' @return Retorna um `data.frame` com os CEPs de input e os endereços presentes
#'   naquele CEP com suas coordenadas geográficas de latitude (`lat`) e
#'   longitude (`lon`). Alternativamente, o resultado pode ser um objeto `sf`.
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(geocodebr)
#'
#' # amostra de CEPs
#' ceps <- c("70390-025", "20071-001", "99999-999")
#'
#' df <- geocodebr::busca_por_cep(
#'   cep = ceps,
#'   h3_res = 10,
#'   verboso = TRUE
#'   )
#'
#' head(df)
#'
#' @export
busca_por_cep <- function(
  cep,
  h3_res = NULL,
  resultado_sf = FALSE,
  verboso = TRUE,
  cache = TRUE
) {
  # check input
  checkmate::assert_vector(cep)
  checkmate::assert_logical(resultado_sf, any.missing = FALSE, len = 1)
  checkmate::assert_logical(verboso, any.missing = FALSE, len = 1)
  checkmate::assert_logical(cache, any.missing = FALSE, len = 1)
  checkmate::assert_numeric(
    h3_res,
    null.ok = TRUE,
    lower = 0,
    upper = 15,
    max.len = 16
  )

  # normalize input data -------------------------------------------------------

  cep_padrao <- enderecobr::padronizar_ceps(cep)
  cep_padrao <- unique(cep_padrao)
  cep_padrao <- na.omit(cep_padrao)
  cep_padrao <- cep_padrao[cep_padrao != ""]

  # download cnefe  -------------------------------------------------------

  # downloading cnefe
  cnefe_dir <- download_cnefe(
    tabela = 'municipio_logradouro_cep_localidade',
    verboso = verboso,
    cache = cache
  )
  # creating a temporary db and register the input table data
  con <- create_geocodebr_db()

  # build path to local file
  path_to_parquet <- fs::path(
    listar_pasta_cache(),
    glue::glue("geocodebr_data_release_{data_release}"),
    paste0("municipio_logradouro_cep_localidade.parquet")
  )

  # ceps string
  unique_ceps <- glue::glue_collapse(glue::single_quote(cep_padrao), sep = ",")

  # select columns
  cols_to_keep <- c(
    "cep",
    "estado",
    "municipio",
    "logradouro",
    "localidade",
    "lon",
    "lat"
  )
  cols_to_keep <- paste0(cols_to_keep, collapse = ", ")

  # filtrar por uf ?
  query_filter_cnefe <- glue::glue(
    "SELECT {cols_to_keep}
        FROM read_parquet('{path_to_parquet}') m
        WHERE m.cep IN ({unique_ceps});"
  )

  output_df <- DBI::dbGetQuery(con, query_filter_cnefe)

  # add any missing cep
  missing_cep <- cep_padrao[!cep_padrao %in% output_df$cep]

  if (length(missing_cep) == length(cep_padrao)) {
    cli::cli_abort("Nenhum CEP foi encontrado.")
  }

  if (length(missing_cep) > 0) {
    temp_dt <- data.table::data.table(cep = missing_cep)
    output_df <- data.table::rbindlist(list(output_df, temp_dt), fill = TRUE)
  }

  data.table::setDT(output_df)

  # add H3
  if (!is.null(h3_res)) {
    for (i in h3_res) {
      colname <- paste0(
        'h3_',
        formatC(h3_res, width = 2, flag = "0")
      )

      output_df[
        !is.na(lat),
        {{ colname }} := h3r::latLngToCell(lat = lat, lng = lon, resolution = i)
      ]
    }
  }

  # convert df to simple feature
  if (isTRUE(resultado_sf)) {
    output_sf <- sfheaders::sf_point(
      obj = output_df,
      x = 'lon',
      y = 'lat',
      keep = TRUE
    )

    sf::st_crs(output_sf) <- 4674
    return(output_sf)
  }

  return(output_df)
}

# skip tests because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("arrow")


ceps_valid <- c("70390-025", "20071-001", "99999-999")
ceps_not_valid <- c("99999-999")


tester <- function(cep,
                   h3_res = NULL,
                   resultado_sf = FALSE,
                   verboso = TRUE,
                   cache = TRUE) {
  busca_por_cep(
    cep,
    h3_res,
    resultado_sf,
    verboso,
    cache
  )
}

test_that("expected output", {

  # expected results
  output <- tester(cep = ceps_valid)
  testthat::expect_true(nrow(output) == 5)

  # expected class
  testthat::expect_s3_class(output, 'data.frame')

  # add h3
  h3_output <- tester(cep = ceps_valid, h3_res = c(3,4))
  testthat::expect_true(all(c('h3_03', 'h3_04') %in% names(h3_output)))


  # output in sf format
  sf_output <- tester(cep = ceps_valid, resultado_sf = TRUE)
  testthat::expect_true(is(sf_output , 'sf'))
})



test_that("errors with incorrect input", {
  expect_error(tester(unclass(ceps_not_valid)))

  expect_error(tester(cep = 1))
  expect_error(tester(cep = 'banana'))
  expect_error(tester(cep = ceps_not_valid))

  expect_error(tester(resultado_completo = 1))
  expect_error(tester(resultado_completo = NA))
  expect_error(tester(resultado_completo = c(TRUE, TRUE)))

  expect_error(tester(resultado_sf = 1))
  expect_error(tester(resultado_sf = NA))
  expect_error(tester(resultado_sf = c(TRUE, TRUE)))

  expect_error(tester(verboso = 1))
  expect_error(tester(verboso = NA))
  expect_error(tester(verboso = c(TRUE, TRUE)))

  expect_error(tester(cache = 1))
  expect_error(tester(cache = NA))
  expect_error(tester(cache = c(TRUE, TRUE)))
})

match_cases <- function(
  # nocov start
  con = con,
  x = "input_padrao_db",
  output_tb = "output_db",
  key_cols = key_cols,
  match_type = match_type,
  resultado_completo
) {
  # match_type = "dn01"

  # get corresponding parquet table
  cnefe_table_name <- get_reference_table(match_type)
  y <- cnefe_table_name
  key_cols <- get_key_cols(match_type)

  # write cnefe table to db
  register_cnefe_table(con, match_type)

  # Create the JOIN condition by concatenating the key columns
  join_condition <- paste(
    glue::glue("{y}.{key_cols} = {x}.{key_cols}"),
    collapse = ' AND '
  )

  # cols from x that cannot be null
  # isso serve como filtro pre-join, pra fazer o join soh em quem nao foi encontrado ainda
  cols_not_null <- paste(
    glue::glue("{x}.{key_cols} IS NOT NULL"),
    collapse = ' AND '
  )

  # whether to keep all columns in the result
  colunas_encontradas <- ""
  additional_cols <- ""

  if (isTRUE(resultado_completo)) {
    colunas_encontradas <- paste0(
      glue::glue("{key_cols}_encontrado"),
      collapse = ', '
    )

    colunas_encontradas <- gsub(
      'localidade_encontrado',
      'localidade_encontrada',
      colunas_encontradas
    )
    colunas_encontradas <- paste0(", ", colunas_encontradas)

    additional_cols <- paste0(
      glue::glue("{y}.{key_cols} AS {key_cols}_encontrado"),
      collapse = ', '
    )

    additional_cols <- gsub(
      'localidade_encontrado',
      'localidade_encontrada',
      additional_cols
    )
    additional_cols <- paste0(", ", additional_cols)

    # adiciona codigo do setor censitario
    additional_cols <- paste0(additional_cols, glue::glue(", {y}.cod_setor AS cod_setor"))
    colunas_encontradas <- paste0(colunas_encontradas, ", cod_setor")

  }



  # summarize query
  query_match <- glue::glue(
    "INSERT INTO output_db (tempidgeocodebr, lat, lon, endereco_encontrado, tipo_resultado,
                            desvio_metros, log_causa_confusao, contagem_cnefe {colunas_encontradas})
      SELECT {x}.tempidgeocodebr,
        {y}.lat,
        {y}.lon,
        {y}.endereco_completo AS endereco_encontrado,
        '{match_type}' AS tipo_resultado,
        {y}.desvio_metros,
        {x}.log_causa_confusao,
        {y}.n_casos AS contagem_cnefe {additional_cols}
      FROM {x}
      INNER JOIN {y}
      ON {join_condition}
      WHERE {cols_not_null};"
  )

  DBI::dbExecute(con, query_match)
  # a <- DBI::dbReadTable(con, 'output_db')
  # summary(a$desvio_metros)
  # summary(a$lat)

  # UPDATE input_padrao_db: Remove observations found in previous step
  temp_n <- update_input_db(
    con,
    update_tb = x,
    reference_tb = output_tb
  )

  return(temp_n)
} # nocov end

devtools::load_all('.')

library(ipeadatalake)
library(dplyr)
library(data.table)
library(enderecobr)
# library(mapview)
# library(sfheaders)
# library(sf)
# options(scipen = 999)
# mapview::mapviewOptions(platform = 'leafgl')
set.seed(42)

#' take-away
#' 1) a performance do geocodebr fica muito proxima do arcgis
#' 2) o que precisa fazer eh checar os casos em q a gente encontra com baixa
#' precisao e arcgis com alta. O que a gente pode fazer para melhorar o match?
#' Usar o LIKE logradouro na join ja melhorou muito, mas ainda daria pra melhorar?
#'
#' t <- subset(rais_like, match_type=='case_09' & Addr_type==	'PointAddress')

2+2
# stop()




# cad unico --------------------------------------------------------------------
sample_size <- 10000000

cad_con <- ipeadatalake::ler_cadunico(
  data = 202312,
  base = 'familia',
  as_data_frame = F,
  colunas = c("co_familiar_fam", "co_uf", "cd_ibge_cadastro",
              "no_localidade_fam", "no_tip_logradouro_fam",
              "no_tit_logradouro_fam", "no_logradouro_fam",
              "nu_logradouro_fam", "ds_complemento_fam",
              "ds_complemento_adic_fam",
              "nu_cep_logradouro_fam", "co_unidade_territorial_fam",
              "no_unidade_territorial_fam", "co_local_domic_fam")
  )

# a <- tail(cad, n = 100) |> collect()

# compose address fields
df <- cad_con |>
  mutate(no_tip_logradouro_fam = ifelse(is.na(no_tip_logradouro_fam), '', no_tip_logradouro_fam),
         no_tit_logradouro_fam = ifelse(is.na(no_tit_logradouro_fam), '', no_tit_logradouro_fam),
         no_logradouro_fam = ifelse(is.na(no_logradouro_fam), '', no_logradouro_fam)
         ) |>
  mutate(abbrev_state = co_uf,
          code_muni = cd_ibge_cadastro,
          logradouro = paste(no_tip_logradouro_fam, no_tit_logradouro_fam, no_logradouro_fam),
          numero = nu_logradouro_fam,
          cep = nu_cep_logradouro_fam,
          bairro = no_localidade_fam) |>
  select(co_familiar_fam,
         abbrev_state,
         code_muni,
         logradouro,
         numero,
         cep,
         bairro) |>
  dplyr::compute() |>
  #dplyr::slice_sample(n = sample_size) |> # sample 20K
  dplyr::collect()

df$id <- 1:nrow(df)

campos <- geocodebr::definir_campos(
  logradouro = 'logradouro',
  numero = 'numero',
  cep = 'cep',
  localidade = 'bairro',
  municipio = 'code_muni',
  estado = 'abbrev_state'
)

stop()


gc(T,T,T)
#bench::system_time(
 bench::mark(iterations = 1,
  cadgeo <- geocodebr::geocode(
    enderecos  = df,
    campos_endereco = campos,
    n_cores = 7, # 7
    verboso = T,
    resultado_completo = F,
    resolver_empates = T,
    #resultado_sf = F
    #, h3_res = 9
    padronizar_enderecos = T
    )
)


# 10 milhoes
# args: n_cores = 7, resultado_completo = F resolver_empates = T
# expression        min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory
# v0.3.0 CRAN     29.7m  29.7m  0.000562    18.3GB   0.0725     1   129      29.7m <NULL> <Rprofmem>
# v0.4.0 CRAN     33.5m  33.5m  0.000497    8.06GB  0.00746     1    15      33.5m <NULL> <Rprofmem>
# v0.5.0 CRAN     6.04m  6.04m   0.00276     916MB  0.00276     1     1      6.04m <df>   <Rprofmem> <bench_tm> <tibble>
# v0.6.0 dev      4.91m  4.91m   0.00339    1014MB   0.0102     1     3      4.91m <df>   <Rprofmem>

# v0.5.0 CRAN     2.39m !!!! em paralelo
# v0.6.0 dev      2.16m !!!! em paralelo



# 43 milhoes
# args: n_cores = 7, resultado_completo = F resolver_empates = T
# expression        min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory
# v0.3.0 CRAN        2h     2h  0.000139    79.3GB   0.0176     1   127         2h <dt>
# v0.4.0 CRAN      3.3h   3.3h 0.0000843    34.5GB  0.00244     1    29       3.3h <dt>   <Rprofmem> <bench_tm> <tibble>
# v0.5.0 CRAN     24.9m  24.9m  0.000670    4.12GB  0.00134     1     2      24.9m <df>
# v0.6.0 dev      17.9m  17.9m  0.000931    4.11GB 0.000931     1     1      17.9m <df>   <Rprofmem>



 # v0.5.0  8.99m  !!!! em paralelo por uf
 # v0.6.0  7.24m  !!!! em paralelo por uf

# encontra setor censitario para % do cad unico
1- sum(is.na(cadgeo$cod_setor)) / nrow(cadgeo)
# v0.6.0: 0.6438606


# nao era para ser empate
# 5 "da02" "da02"   mesmo rua e cep
# [1] "RUA PAULO SIMOES DA COSTA, 32 (aprox) - JARDIM ANGELA, SAO PAULO - SP, 04929-140"
# [2] "RUA PAULO SIMOES DA COSTA, 32 (aprox) - ALTO DO RIVIERA, SAO PAULO - SP, 04929-140"



## cadunico cada passo ----------------

gc(T,T,T)
bench::bench_time(
 #bench::mark(
 callr::r(function(df, campos) {
   rfiles <- list.files("R", full.names = TRUE)
   invisible(lapply(rfiles, source))
   library("data.table")
   # library("geocodebr")
   cadgeo <- geocode_core(
     enderecos = df,
     campos_endereco = campos,
     n_cores = 7,
     resultado_completo = F,
     verboso = T,
     resolver_empates = T,
     resultado_sf = F,
     padronizar_enderecos = T,
     h3_res = NULL,
     cache = T
   )
 },
 args = list(df = df, campos = campos),
 show = TRUE
 )
)

# 10 milhoes
# v0.6.0 dev
#                             step_sec total_sec step_relative
#                       Start     0.00      0.00           0.0
#                Padronizacao    29.00     29.00          11.9
# Register standardized input    22.14     51.14           9.1
#                    Matching   159.42    210.56          65.2
#             Resolve empates     7.39    217.95           3.0
#   Write original input back     3.82    221.77           1.6
#               Add precision     0.36    222.13           0.1
#               Merge results    22.40    244.53           9.2
#                                          4.07m


# 43 milhoes
# v0.6.0 dev
#                             step_sec total_sec step_relative
#                       Start     0.02      0.02           0.0
#                Padronizacao   102.76    102.78          12.5
# Register standardized input    85.88    188.66          10.5
#                    Matching   499.81    688.47          60.9
#             Resolve empates    31.94    720.41           3.9
#   Write original input back    17.32    737.73           2.1
#               Add precision     1.94    739.67           0.2
#               Merge results    80.92    820.59           9.9
#                                         17.27m


## cadunico parallel callr ----------------

# library(future.callr)
library(future)
library(furrr)

df$abbrev_state <- enderecobr::padronizar_estados(df$abbrev_state)
df$code_muni <- enderecobr::padronizar_municipios(df$code_muni)

future::plan(future::multisession)


gc(T,T,T)
bench::bench_time(
 a <-   split(df, f = df$abbrev_state) |>
   furrr::future_map(
     .progress = TRUE,
     .f = function(x){
       geocode(
         enderecos = x,
         campos_endereco = campos,
         resultado_completo = F,
         resolver_empates = T
       )
     }
   )

)

bench::bench_time(

a <- data.table::rbindlist(a)
)

quantile(a$desvio_metros, na.rm = T, probs = c(0.5, 0.7, 0.75, 0.8, 0.85, 0.9))
  # 50%   70%   75%   80%   85%   90%
  #   7   228   451   864  2418 10763


# rais --------------------------------------------------------------------

rais <- ipeadatalake::ler_rais(
  ano = 2019,
  tipo = 'estabelecimento',
  as_data_frame = F,
  geoloc = T) |>
  select("id_estab", "logradouro", "bairro", "codemun", "uf", "cep",
         'lat', 'lon', 'Addr_type', 'Match_addr') |>
  compute() |>
  dplyr::slice_sample(n = 1000000) |> # sample 10 million
  filter(uf != "IG") |>
  filter(uf != "") |>
  collect()


# rais <- head(rais, n = 1000) |> collect() |> dput()
data.table::setDT(rais)

# create column number
rais[, numero := gsub("[^0-9]", "", logradouro)]

# remove numbers from logradouro
rais[, logradouro_no_numbers := gsub("//d+", "", logradouro)]
rais[, logradouro_no_numbers := gsub(",", "", logradouro_no_numbers)]

rais[, id := 1:nrow(rais)]

data.table::setnames(
  rais,
  old = c('lat', 'lon'),
  new = c('lat_arcgis', 'lon_arcgis')
)


head(rais)


fields <- geocodebr::definir_campos(
  logradouro = 'logradouro_no_numbers',
  numero = 'numero',
  cep = 'cep',
  bairro = 'bairro',
  municipio = 'codemun',
  estado = 'uf'
)



rais_geo <- geocodebr::geocode(
    addresses_table = rais,
    address_fields = fields,
    n_cores = 7,
    full_results =  T,
    progress = T
  )



# censo escolar ---------------------------------

censo_escolar <- ipeadatalake::ler_censo_escolar(
  ano = 2022,
  base = 'basica'
  )  |>
  select(
    c("NU_ANO_CENSO",                "NO_REGIAO",
      "CO_REGIAO",                   "NO_UF",
      "SG_UF",                       "CO_UF",
      "NO_MUNICIPIO",                "CO_MUNICIPIO",
      "NO_MESORREGIAO",              "CO_MESORREGIAO",
      "NO_MICRORREGIAO",             "CO_MICRORREGIAO",
      "CO_DISTRITO",                 "NO_ENTIDADE",
      "CO_ENTIDADE",                 "TP_DEPENDENCIA",
      "TP_CATEGORIA_ESCOLA_PRIVADA", "TP_LOCALIZACAO",
      "TP_LOCALIZACAO_DIFERENCIADA", "DS_ENDERECO",
      "NU_ENDERECO",                 "DS_COMPLEMENTO",
      "NO_BAIRRO",                   "CO_CEP",
      "NU_DDD")
    ) |>
  dplyr::collect()


censo_escolar$id <- 1:nrow(censo_escolar)

fields_cad <- geocodebr::definir_campos(
  logradouro = 'DS_ENDERECO',
  numero = 'NU_ENDERECO',
  cep = 'CO_CEP',
  localidade = 'NO_BAIRRO',
  municipio = 'NO_MUNICIPIO',
  estado = 'NO_UF'
)

# bench::mark( iterations = 1,
bench::system_time(
  geo <- geocodebr::geocode(
    enderecos  = censo_escolar,
    campos_endereco = fields_cad,
    resultado_completo = T,
    n_cores = 25, # 7
    verboso = T,
    resultado_sf = F,
    resolver_empates = F
  )
)

# prolema atual é que a pacote recalcula distancias de string TODA VEZ
# seria melhor criar uma tabela de distancia e ir populando

# 1st step: create small table with unique logradouros
# 2nd step: update input_padrao_db with the most probable logradouro
# 3rd step: deterministic match to update output

match_cases_probabilistic <- function(
  con = con,
  x = 'input_padrao_db',
  output_tb = "output_db",
  key_cols = key_cols,
  match_type = match_type,
  resultado_completo
) {
  # nocov start

  # match_type = "pn01"

  # get corresponding parquet table and key columns
  cnefe_table_name <- get_reference_table(match_type)
  y <- cnefe_table_name
  key_cols <- get_key_cols(match_type)

  # write cnefe table to db
  register_cnefe_table(con, match_type)

  # 1st step: create small table with unique logradouros -----------------------
  unique_logradouros_tbl <- register_unique_logradouros_table(con, match_type)

  # 2nd step: update input_padrao_db with the most probable logradouro ---------

  calculate_string_dist(con, match_type, unique_logradouros_tbl)

  # 3rd step: update output table com match deterministico --------------------------------------------------------

  # update join condition to use probable logradouro and deterministic number
  join_condition_match <- paste(
    glue::glue("{y}.{key_cols} = {x}.{key_cols}"),
    collapse = ' AND '
  )

  join_condition_match <- gsub(
    'input_padrao_db.logradouro',
    'input_padrao_db.temp_lograd_determ',
    join_condition_match
  )

  # update cols that cannot be null
  cols_not_null <- paste(
    glue::glue("{x}.{key_cols} IS NOT NULL"),
    collapse = ' AND '
  )

  # cols that cannot be null
  cols_not_null <- gsub('.logradouro', '.temp_lograd_determ', cols_not_null)

  # whether to keep all columns in the result
  colunas_encontradas <- ""
  additional_cols <- ""

  if (isTRUE(resultado_completo)) {
    colunas_encontradas <- paste0(
      glue::glue("{key_cols}_encontrado"),
      collapse = ', '
    )

    colunas_encontradas <- gsub(
      'localidade_encontrado',
      'localidade_encontrada',
      colunas_encontradas
    )
    colunas_encontradas <- paste0(", ", colunas_encontradas)
    colunas_encontradas <- paste0(
      colunas_encontradas,
      ", similaridade_logradouro"
    )

    additional_cols <- paste0(
      glue::glue("{y}.{key_cols} AS {key_cols}_encontrado"),
      collapse = ', '
    )

    additional_cols <- gsub(
      'localidade_encontrado',
      'localidade_encontrada',
      additional_cols
    )
    additional_cols <- paste0(
      ", ",
      additional_cols,
      ", input_padrao_db.similaridade_logradouro AS similaridade_logradouro"
    )

    # adiciona codigo do setor censitario
    additional_cols <- paste0(additional_cols, glue::glue(", {y}.cod_setor AS cod_setor"))
    colunas_encontradas <- paste0(colunas_encontradas, ", cod_setor")

  }

  # summarize query
  query_update_db <- glue::glue(
    "INSERT INTO output_db (tempidgeocodebr, lat, lon, endereco_encontrado, tipo_resultado,
                            desvio_metros, log_causa_confusao, contagem_cnefe {colunas_encontradas} )
      SELECT {x}.tempidgeocodebr,
        {y}.lat,
        {y}.lon,
        {y}.endereco_completo AS endereco_encontrado,
        '{match_type}' AS tipo_resultado,
        {y}.desvio_metros,
        {x}.log_causa_confusao,
        {y}.n_casos AS contagem_cnefe {additional_cols}
      FROM {x}
      INNER JOIN {y}
      ON {join_condition_match}
      WHERE {cols_not_null};"
  )

  DBI::dbExecute(con, query_update_db)
  # DBI::dbExecute(con, query_update_db)
  # b <- DBI::dbReadTable(con, 'output_db')

  # UPDATE input_padrao_db: Remove observations found in previous step
  temp_n <- update_input_db(
    con,
    update_tb = x,
    reference_tb = output_tb
  )

  return(temp_n)
} # nocov end

# test
df_rp <- data.frame(
  estado = 'df',
  # cep=70355030,
  municipio='brasilia',
  logradouro = "sqs 308",
  bairro = 'asa sul'
  )

campos <- geocodebr::definir_campos(
  logradouro = 'logradouro',
  # numero = 'numero',
  # cep = 'cep',
  localidade = 'bairro',
  municipio = 'municipio',
  estado = 'estado'
)

test <- geocodebr::geocode(
  enderecos = df_rp,
  campos_endereco = campos,
  resultado_completo = T,
  verboso = T,
  resultado_sf = T
  )

mapview::mapview(test)



# NEXT CHANGES  ---------------------------
#
# #> using cache_dir and data_release as global variables
# https://stackoverflow.com/questions/12598242/global-variables-in-packages-in-r
#
# censobr_env <- new.env()
# censobr_env$data_release <- 'v0.1.0'
#


#> cache function delete data from previous package versions

# current cache
pkgv <- paste0('censobr_', 'v0.1.0' )
cache_dir <- tools::R_user_dir(pkgv, which = 'cache')

# determine old cache
dir_above <- dirname(cache_dir)
all_cache <- list.files(dir_above, pattern = 'censobr',full.names = TRUE)
old_cache <- all_cache[!grepl(pkgv, all_cache)]

# delete
unlink(old_cache, recursive = TRUE)


library(censobr)

censobr_cache(delete_file = 'all')

interview_manual(year = 2010)
interview_manual(year = 2000)
interview_manual(year = 1991) #
interview_manual(year = 1980) #
interview_manual(year = 1970)



# github actions ------------------------

usethis::use_github_action("test-coverage")
usethis::use_github_action("check-standard")
usethis::use_github_action("pkgdown")

# Coverage ------------------------
# usethis::use_coverage()
# usethis::use_github_action("test-coverage")

library(testthat)
library(covr)
Sys.setenv(NOT_CRAN = "true")



# each function separately
t1 <- covr::function_coverage(fun=geocode, test_file("tests/testthat/test-geocode.R"))
t1 <- covr::function_coverage(fun=definir_campos, test_file("tests/testthat/test-definir_campos.R"))
t1 <- covr::function_coverage(fun=download_cnefe, test_file("tests/testthat/test-download_cnefe.R"))
t1 <- covr::function_coverage(fun=listar_dados_cache, test_file("tests/testthat/test-cache.R"))
# t1 <- covr::function_coverage(fun=deletar_pasta_cache, test_file("tests/testthat/test-cache.R"))
t1 <- covr::function_coverage(fun=busca_por_cep, test_file("tests/testthat/test-busca_por_cep.R"))
t1 <- covr::function_coverage(fun=geocodebr_message, test_file("tests/testthat/test-message.R"))

t1

rep <- covr::report(t1)



# nocov start

# nocov end

# the whole package
Sys.setenv(NOT_CRAN = "true")
cov <- covr::package_coverage(path = ".", type = "tests", clean = FALSE)
cov

rep <- covr::report(t1)

x <- as.data.frame(cov)
covr::codecov( coverage = cov, token ='aaaaa' )






# checks spelling ----------------
library(spelling)
devtools::spell_check(pkg = ".", vignettes = TRUE, use_wordlist = TRUE)


### Check URL's ----------------

urlchecker::url_update()





# CMD Check --------------------------------
# Check package errors

# linux ubuntu-clang for CRAN
rhub::rhub_check(platforms = 'ubuntu-clang')

# run only the tests
Sys.setenv(NOT_CRAN = "true")
testthat::test_local()

# LOCAL
Sys.setenv(NOT_CRAN = "true")
devtools::check(pkg = ".",  cran = FALSE, env_vars = c(NOT_CRAN = "true"))

# detecta problema de cpu < 2
devtools::check(remote = TRUE, manual = TRUE)

devtools::check(remote = TRUE, manual = TRUE, env_vars = c(NOT_CRAN = "false"))


# CRAN
Sys.setenv(NOT_CRAN = "false")
devtools::check(pkg = ".",  cran = TRUE, env_vars = c(NOT_CRAN = "false"))


# extrachecks -----------------
#' https://github.com/JosiahParry/extrachecks
#' remotes::install_github("JosiahParry/extrachecks")

library(extrachecks)
extrachecks::extrachecks()


# submit to CRAN -----------------
# usethis::use_cran_comments()


devtools::submit_cran()


# build binary -----------------
system("R CMD build . --resave-data") # build tar.gz




###
library(pkgdown)
library(usethis)
# usethis::use_pkgdown_github_pages() # only once
## coverage
usethis::use_coverage()
usethis::use_github_action("test-coverage")

pkgdown::build_site()
#
#
# ### cache
# tigris
# https://github.com/walkerke/tigris/blob/0f0d7992e0208b4c55a9fe8ac6c52f9e438a3b0c/R/helpers.R#L78
#
# https://github.com/walkerke/tigris/blob/0f0d7992e0208b4c55a9fe8ac6c52f9e438a3b0c/R/helpers.R#L78
#
# tidycensus
# https://github.com/walkerke/tidycensus/blob/master/R/search_variables.R



library(geobr)
library(censobr)
library(dplyr)
library(ggplot2)

fort_df <- geobr::lookup_muni(name_muni = 'Sao Paulo')
fort_code <- fort_df$code_muni
fort_wa <- read_weighting_area(code_weighting = fort_code,
                               year = 2010,
                               simplified = FALSE)



ggplot() +
  geom_sf(data=fort_wa)


# download household data
hs <- read_households(year = 2010,
                      showProgress = FALSE)


rent <- hs |>
  mutate( V0011 = as.character(V0011)) |>
  filter(V0011 %in% fort_wa$code_weighting) |>
  collect() |>
  group_by(V0011) |>                                                 # (a)
  summarize(avgrent=weighted.mean(x=V2011, w=V0010, na.rm=TRUE)) |>  # (b)
  collect()                                                          # (c)

head(rent)


for_sf <- left_join(fort_wa, rent, by = c('code_weighting'='V0011'))


ggplot() +
  geom_sf(data=for_sf, aes(fill = avgrent), color=NA)






# had CPU time 3 times elapsed time ----
# https://stackoverflow.com/questions/77323811/r-package-to-cran-had-cpu-time-5-times-elapsed-time

# Flavor: r-devel-linux-x86_64-debian-gcc
# Check: tests, Result: NOTE
# Running 'testthat.R' [161s/53s]
# Running R code in 'testthat.R' had CPU time 3 times elapsed time

library(testthat)
Sys.setenv(NOT_CRAN = "true")

files = list.files("tests/testthat", "test-", full.names = TRUE)

for (file in files[3]) {

  time = system.time(test_file(file))
  ratio = time[1] / time[3]
  if (ratio > 2) {
    print(time)
    stopf("Test file %s had CPU time %f times elapsed time", file, ratio)
  }
}

devtools::load_all('.')

# open input data
data_path <- system.file("extdata/large_sample.parquet", package = "geocodebr")
input_df <- arrow::read_parquet(data_path)


ncores <- 7


campos <- geocodebr::definir_campos(
  logradouro = 'logradouro',
  numero = 'numero',
  cep = 'cep',
  localidade = 'bairro',
  municipio = 'municipio',
  estado = 'uf'
)


bench::mark(iterations = 3,
  a <- geocodebr::geocode(
    enderecos = input_df,
    campos_endereco = campos,
    n_cores = 7,
    resultado_completo = F,
    verboso = T,
    resultado_sf = F,
    resolver_empates = T,
    cache = T
  )
)

#          expression    min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory
# streetmap 0.6.0 dev  7.10s  7.26s     0.136    5.47MB        0     5     0      36.7s <df>   <Rprofmem>
# laptop    0.6.0 CRAN  5.2s  5.53s     0.174    7.46MB        0     5     0      28.8s <df>
# load 1 a <- geoc…  8.1s  8.79s     0.116    3.03MB        0     3     0        26s
# sem 1 a <- geoc… 10.3s  10.5s    0.0944    3.03MB        0     3     0      31.8s

#' @details Precisão dos resultados:
#'
#' A precisão dos resultados do **{geocodebr}** são apresentadas em 3 colunas,
#' `precisao`, `tipo_resultado` e `desvio_metros`, explicadas abaixo.
#'
#' # Precisão
#'
#' Os resultados são classificados em seis amplas categorias de `precisao`:
#'
#' 1. "numero"
#' 2. "numero_aproximado"
#' 3. "logradouro"
#' 4. "cep"
#' 5. "localidade"
#' 6. "municipio"
#'
#' Cada nível de precisão pode ser desagregado em tipos de resultado mais
#' refinados.
#'
#' # Tipos de resultados
#'
#' A coluna `tipo_resultado` fornece informações mais detalhadas sobre como
#' exatamente cada endereço de entrada foi encontrado no CNEFE. Em cada
#' categoria,  o **{geocodebr}** calcula a média da latitude e longitude dos
#' endereços incluídos no CNEFE que correspondem ao endereço de entrada, com
#' base em combinações de diferentes campos. No caso mais rigoroso, por exemplo,
#' a função encontra uma correspondência determinística para todos os campos de
#' um dado endereço (`"estado"`, `"municipio"`, `"logradouro"`, `"numero"`,
#' `"cep"`, `"localidade"`). Pense, por exemplo, em um prédio com vários
#' apartamentos que correspondem ao mesmo endereço de rua e número. Nesse caso,
#' as coordenadas dos apartamentos podem diferir ligeiramente, e o
#' **{geocodebr}** calcula a média dessas coordenadas. Em um caso menos rigoroso,
#' no qual apenas os campos (`"estado"`, `"municipio"`, `"logradouro"`,
#' `"localidade"`) são encontrados, o **{geocodebr}** calcula as coordenadas
#' médias de todos os endereços no CNEFE ao longo daquela rua e que se encontram
#' na mesma localidade/bairro. Assim, as coordenadas de resultado tendem a ser o
#' ponto médio do trecho daquela rua que passa dentro daquela localidade/bairro.
#'
#' A coluna `tipo_resultado` fornece informações mais detalhadas sobre os campos de
#' endereço utilizados no cálculo das coordenadas de cada endereço de entrada. Cada
#' categoria é nomeada a partir de um código de quatro caracteres:
#'
#' - o primeiro caracter, sempre `d` ou `p`, determina se a correspondência foi
#' feita de forma determinística (`d`) ou probabilística (`p`);
#' - o segundo faz menção à categoria de `precisao` na qual o resultado foi
#' classificado (`n` para `"numero"`, `a` para `"numero_aproximado"`, `r` para
#' `"logradouro"`, `c` para `"cep"`, `b` para `"localidade"` e `m` para `"municipio"`);
#' - o terceiro e o quarto caracteres designam a classificação de cada categoria
#' dentro de seu grupo - via de regra, quanto menor o número formado por esses
#' caracteres, mais precisa são as coordenadas calculadas.
#'
#' As categorias de `tipo_resultado` são listadas abaixo, junto às categorias de
#' `precisao` a qual elas estão associadas:
#'
#' - precisao `"numero"`
#'   - `dn01` - logradouro, numero, cep e localidade
#'   - `dn02` - logradouro, numero e cep
#'   - `dn03` - logradouro, numero e localidade
#'   - `dn04` - logradouro e numero
#'   - `pn01` - logradouro, numero, cep e localidade
#'   - `pn02` - logradouro, numero e cep
#'   - `pn03` - logradouro, numero e localidade
#'   - `pn04` - logradouro e numero
#'
#' - precisao `"numero_aproximado"`
#'   - `da01` - logradouro, numero, cep e localidade
#'   - `da02` - logradouro, numero e cep
#'   - `da03` - logradouro, numero e localidade
#'   - `da04` - logradouro e numero
#'   - `pa01` - logradouro, numero, cep e localidade
#'   - `pa02` - logradouro, numero e cep
#'   - `pa03` - logradouro, numero e localidade
#'   - `pa04` - logradouro e numero
#'
#' - precisao `"logradouro"` (quando o número de entrada está faltando 'S/N')
#'   - `dl01` - logradouro, cep e localidade
#'   - `dl02` - logradouro e cep
#'   - `dl03` - logradouro e localidade
#'   - `dl04` - logradouro
#'   - `pl01` - logradouro, cep e localidade
#'   - `pl02` - logradouro e cep
#'   - `pl03` - logradouro e localidade
#'   - `pl04` - logradouro
#'
#' - precisao `"cep"`
#'   - `dc01` - municipio, cep, localidade
#'   - `dc02` - municipio, cep
#'
#' - precisao `"localidade"`
#'   - `db01` - municipio, localidade
#'
#' - precisao `"municipio"`
#'   - `dm01` - municipio
#'
#' Endereços não encontrados são retornados com latitude, longitude, precisão e
#' tipo de resultado `NA`.
#'
#' # Desvio em metros
#'
#' A coluna `desvio_metros` apresenta uma forma intuitiva e prática de saber o
#' grau de incerteza do resultado encontrado. Essa coluna informa que pelo menos
#' 95% de todos os pontos do CNEFE que possuem corrêspondência com o endereço de
#' input estão num raio de `x` metros da localização encontrada.
#'
#' Um desvio de até `30` metros, por exemplo, tende a representar um resultado
#' muito confiável. A depender de como o dado geolocalizado será utilizado, até
#' mesmos resultados com `desvio_metros` de até 500 ou 900 metros podem ser ser
#' aceitáveis.
#'
#' A coluna `desvio_metros` é particularmente útil para decidir por exemplo se
#' um resultado encontrado com a `precisao` de CEP, localidade ou logradouro
#' deveria ser aceitável. Por exemplo, muitas cidades do Brasil possuem um CEP
#' único, o que tende a gerar resultados com alto grau de incerteza. Em várias
#' cidades, no entanto, um CEP pode ser circunscrito a uma área muito pequena e
#' as vezes até um único edifício. Nesses casos, o valor do `desvio_metros`
#' tende a ser bem pequeno.
#'
#' # Busca probabilistica
#'
#' Os tipos de resultado com busca probabilitisca usam como base o algoritmo de
#' semelhança de Jaro para comparar as strings de 'logradouro' dos dados de
#' input e da base de endereços do geocodebr. O pacote considera como match o
#' logradouro da base de endereços que apresenta a maior semelhança de Jaro
#' condicionado a uma semelhança mínima, e desde que também haja match
#' determinístico em ao menos um dos campos "cep" e "localidade".
#' O geocodebr utiliza uma semelhança mínima de `0.85` nos casos de match
#' probabilistico, e de `0.90` nos demais casos.
#'
#' # Código do setor censitário
#'
#' Quando o usuário passa o argumento `resultado_completo = TRUE`, a função
#' `geocode()` também retorna a coluna `cod_setor` com o código do setor
#' censitário do endereço encontrado. Atualmente, a função somente retorna o
#' código do setor dos casos em que todos os pontos do CNEFE correspondentes estão
#' 100% dentro de um único setor censitário. Quando os dados do CNEFE
#' correspondentes ao endereço buscado estão em mais de um setor, o resultado da
#' coluna `cod_setor` é `NA`.

#' @param cache Um valor lógico. Indica se os dados do CNEFE devem ser salvos ou
#'   lidos do cache, reduzindo o tempo de processamento em chamadas futuras. O
#'   padrão é `TRUE`. Quando `FALSE`, os dados do CNEFE são baixados para um
#'   diretório temporário.

# ===================================================================
# SCRIPT 1: ANÁLISE DE PARIDADE (LÓGICA E VISUAL)
#
# Objetivo: Comparar os resultados das funções de geocodificação
# original e em DuckDB para garantir que a lógica é a mesma.
# ===================================================================

# --- 1. SETUP E CONFIGURAÇÃO ---
library(devtools)
library(dplyr)
library(sf)
library(cli)
library(waldo)
library(ggplot2)
library(geobr)
library(arrow)
library(writexl)

devtools::load_all()
source("R/geocode_duckdb.R")
source("R/trata_empates_geocode_duckdb.R")

# --- 2. DEFINIÇÃO DE PARÂMETROS E DIRETÓRIOS ---

# Diretório de output para os mapas
output_dir <- "tests/tests_pedro"
cli_alert_info("Os mapas serão salvos em: '{.path {output_dir}}'")

# Parâmetros para salvar as figuras
img_width <- 40
img_height <- 20
img_dpi <- 300

# Carrega os dados de entrada
input_df <- read.csv(system.file("extdata/small_sample.csv", package = "geocodebr"))

# Definição dos campos de endereço
campos <- geocodebr::definir_campos(
  logradouro = "nm_logradouro", numero = "Numero", cep = "Cep",
  localidade = "Bairro", municipio = "nm_municipio", estado = "nm_uf"
)

# --- 3. EXECUÇÃO E COMPARAÇÃO LÓGICA ---
cli::cli_h1("Análise de paridade dos resultados")
cli_alert_info("Geocodificando com ambas as funções para análise de paridade (pode demorar)...")

# Executa as 4 combinações
df_orig_trat <- geocodebr::geocode(input_df, campos, resolver_empates = TRUE, resultado_sf = TRUE, cache = TRUE, n_cores = 4)
df_orig_nao_trat <- geocodebr::geocode(input_df, campos, resolver_empates = FALSE, resultado_sf = TRUE, cache = TRUE, n_cores = 4)
df_duckdb_trat <- geocode_duckdb(input_df, campos, resolver_empates = TRUE, resultado_sf = TRUE, cache = TRUE, n_cores = 4)
df_duckdb_nao_trat <- geocode_duckdb(input_df, campos, resolver_empates = FALSE, resultado_sf = TRUE, cache = TRUE, n_cores = 4)
cli_alert_success("Geocodificação para paridade concluída.")

# Comparação Lógica com Waldo
#cli_alert_info("Comparando resultados lógicos com waldo::compare()")
#waldo::compare(df_orig_nao_trat, df_duckdb_nao_trat)
#waldo::compare(df_orig_trat, df_duckdb_trat)

# --- 4. COMPARAÇÃO VISUAL (MAPAS) ---
cli::cli_h2("Geração de mapas para comparação visual")

# Preparação dos dados para o mapa
df_orig_trat$funcao <- "Original"
df_orig_nao_trat$funcao <- "Original"
df_duckdb_trat$funcao <- "DuckDB"
df_duckdb_nao_trat$funcao <- "DuckDB"
mapa_df_tratado <- rbind(df_orig_trat, df_duckdb_trat) %>% filter(!sf::st_is_empty(geometry))
mapa_df_nao_tratado <- rbind(df_orig_nao_trat, df_duckdb_nao_trat) %>% filter(!sf::st_is_empty(geometry))

# Baixar a camada de fundo
cli_alert_info("Baixando a malha de estados do Brasil com 'geobr'...")
estados_br <- geobr::read_state(year = 2020)
cli_alert_success("Malha de estados carregada.")

# --- Figura 1: Comparação COM Resolução de Empates ---
cli_alert_info("Gerando e exportando mapa para resultados COM resolução...")
plot_tratado <- ggplot() +
  geom_sf(data = estados_br, fill = "gray95", color = "gray80", size = 0.1) +
  geom_sf(data = mapa_df_tratado, size = 2.5, shape = 21, aes(fill = funcao), color = 'black', alpha = 0.8) +
  facet_wrap(~ funcao) +
  scale_fill_manual(values = c("Original" = "#F8766D", "DuckDB" = "#00BFC4"), name = "Função") +
  labs(title = "Comparação de geocodificação com resolução de empates",
       subtitle = "Distribuição dos pontos geocodificados sobre a malha de estados (amostra pequena)") +
  theme_minimal(base_size = 16) +
  theme(axis.title = element_blank(), axis.text = element_blank(), axis.ticks = element_blank(),
        panel.grid = element_line(color = "#f0f0f0"),
        plot.title = element_text(hjust = 0.5, size = rel(1.5), face = "bold"),
        plot.subtitle = element_text(hjust = 0.5, size = rel(1.1)),
        legend.position = "none")
plot_tratado
ggsave(file.path(output_dir, "mapa_comparacao_tratado.png"), plot = plot_tratado, width = img_width, height = img_height, units = "cm", dpi = img_dpi, bg = "white")
ggsave(file.path(output_dir, "mapa_comparacao_tratado.pdf"), plot = plot_tratado, width = img_width, height = img_height, units = "cm")


# --- Figura 2: Comparação SEM Resolução de Empates ---
cli_alert_info("Gerando e exportando mapa para resultados SEM resolução...")
plot_nao_tratado <- ggplot() +
  geom_sf(data = estados_br, fill = "gray95", color = "gray80", size = 0.1) +
  geom_sf(data = mapa_df_nao_tratado, size = 2.5, shape = 21, aes(fill = funcao), color = 'black', alpha = 0.8) +
  facet_wrap(~ funcao) +
  scale_fill_manual(values = c("Original" = "#F8766D", "DuckDB" = "#00BFC4"), name = "Função") +
  labs(title = "Comparação de geocodificação sem resolução de empates",
       subtitle = "Distribuição dos pontos geocodificados sobre a malha de estados (amostra pequena)") +
  theme_minimal(base_size = 16) +
  theme(axis.title = element_blank(), axis.text = element_blank(), axis.ticks = element_blank(),
        panel.grid = element_line(color = "#f0f0f0"),
        plot.title = element_text(hjust = 0.5, size = rel(1.5), face = "bold"),
        plot.subtitle = element_text(hjust = 0.5, size = rel(1.1)),
        legend.position = "none")
plot_nao_tratado
ggsave(file.path(output_dir, "mapa_comparacao_nao_tratado.png"), plot = plot_nao_tratado, width = img_width, height = img_height, units = "cm", dpi = img_dpi, bg = "white")
ggsave(file.path(output_dir, "mapa_comparacao_nao_tratado.pdf"), plot = plot_nao_tratado, width = img_width, height = img_height, units = "cm")

cli_alert_success("Análise de paridade e geração de mapas concluída.")

#' @param n_cores Um número. O número de núcleos de CPU a serem utilizados no
#'   processamento dos dados. Por padrão, `n_cores = NULL` e o pacote utiliza o
#'   número máximo de cores físicos disponíveis.

# 1st step: create small table with unique logradouros
# 2nd step: update input_padrao_db with the most probable logradouro
# 3rd step: deterministic match
# 4th step: aggregate

match_weighted_cases_probabilistic <- function(
  # nocov start
  con = con,
  x = 'input_padrao_db',
  output_tb = "output_db",
  key_cols = key_cols,
  match_type = match_type,
  resultado_completo) {

  # match_type = "pn01"

  # get corresponding parquet table
  cnefe_table_name <- get_reference_table(match_type)
  y <- cnefe_table_name
  key_cols <- get_key_cols(match_type)

  # write cnefe table to db
  register_cnefe_table(con, match_type)

  # 1st step: create small table with unique logradouros -----------------------
  unique_logradouros_tbl <- register_unique_logradouros_table(con, match_type)

  # 2nd step: update input_padrao_db with the most probable logradouro ---------
  calculate_string_dist(con, match_type, unique_logradouros_tbl)

  # 3rd step: match deterministico --------------------------------------------------------

  # cols that cannot be null
  cols_not_null <- paste(
    glue::glue("{x}.{key_cols} IS NOT NULL"),
    collapse = ' AND '
  )

  key_cols <- key_cols[key_cols != 'numero']

  # Create the JOIN condition by concatenating the key columns
  join_condition_determ <- paste(
    glue::glue("{y}.{key_cols} = {x}.{key_cols}"),
    collapse = ' AND '
  )

  # update join condition to use probable logradouro
  join_condition_determ <- gsub(
    'input_padrao_db.logradouro',
    'input_padrao_db.temp_lograd_determ',
    join_condition_determ
  )

  # cols that cannot be null
  cols_not_null_match <- gsub(
    '.logradouro',
    '.temp_lograd_determ',
    cols_not_null
  )

  # whether to keep all columns in the result
  colunas_encontradas <- ""
  additional_cols_first <- ""
  additional_cols_second <- ""

  if (isTRUE(resultado_completo)) {
    colunas_encontradas <- paste0(
      glue::glue("{key_cols}_encontrado"),
      collapse = ', '
    )

    colunas_encontradas <- gsub(
      'localidade_encontrado',
      'localidade_encontrada',
      colunas_encontradas
    )
    colunas_encontradas <- paste0(", ", colunas_encontradas)

    # additonal cols for the first part of the query
    additional_cols_first <- paste0(
      glue::glue("{y}.{key_cols} AS {key_cols}_encontrado"),
      collapse = ', '
    )
    additional_cols_first <- gsub(
      'localidade_encontrado',
      'localidade_encontrada',
      additional_cols_first
    )
    additional_cols_first <- paste0(", ", additional_cols_first)

    # additonal cols for the second part of the query
    additional_cols_second <- paste0(
      glue::glue("FIRST({key_cols}_encontrado) AS {key_cols}_encontrado"),
      collapse = ', '
    )
    additional_cols_second <- gsub(
      'localidade_encontrado',
      'localidade_encontrada',
      additional_cols_second
    )
    additional_cols_second <- paste0(", ", additional_cols_second)

    # adiciona codigo do setor censitario
    additional_cols_first <- paste0(additional_cols_first, glue::glue(", {y}.cod_setor AS cod_setor"))
    additional_cols_second <- paste0(additional_cols_second, glue::glue(", FIRST(cod_setor)"))
    colunas_encontradas <- paste0(colunas_encontradas, ", cod_setor")

  }

  # Match query  --------------------------------------------------------

  query_match <- glue::glue(
    "
  -- PART 1) inner join to get all cases that match
  WITH temp_db AS (
      SELECT {x}.tempidgeocodebr,
             {x}.numero,
             {y}.numero AS numero_cnefe,
             {y}.lat, {y}.lon,
             REGEXP_REPLACE( {y}.endereco_completo, ', \\d+ -', CONCAT(', ', {x}.numero, ' (aprox) -')) AS endereco_encontrado,
             {x}.similaridade_logradouro,
             {y}.desvio_metros,
             {x}.log_causa_confusao,
             {y}.n_casos AS contagem_cnefe {additional_cols_first}
          FROM {x}
          INNER JOIN {y}
          ON {join_condition_determ}
          WHERE {cols_not_null_match}
          )

  -- PART 2: aggregate and interpolate get aprox location

  INSERT INTO output_db (tempidgeocodebr, lat, lon, endereco_encontrado, tipo_resultado, desvio_metros,
                         log_causa_confusao, similaridade_logradouro, contagem_cnefe {colunas_encontradas})
       SELECT tempidgeocodebr,
         SUM((1/ABS(numero - numero_cnefe) * lat)) / SUM(1/ABS(numero - numero_cnefe)) AS lat,
         SUM((1/ABS(numero - numero_cnefe) * lon)) / SUM(1/ABS(numero - numero_cnefe)) AS lon,
         FIRST(endereco_encontrado) AS endereco_encontrado,
         '{match_type}' AS tipo_resultado,
         AVG(desvio_metros) AS desvio_metros,
         FIRST(log_causa_confusao) AS log_causa_confusao,
         FIRST(similaridade_logradouro) AS similaridade_logradouro,
         FIRST(contagem_cnefe) AS contagem_cnefe {additional_cols_second}
      FROM temp_db
      GROUP BY tempidgeocodebr, endereco_encontrado;"
  )

  DBI::dbExecute(con, query_match)
  # DBI::dbExecute(con, query_aggregate)
  # d <- DBI::dbReadTable(con, 'output_db')
  # d <- DBI::dbReadTable(con, 'aaa')

  # UPDATE input_padrao_db: Remove observations found in previous step
  temp_n <- update_input_db(
    con,
    update_tb = x,
    reference_tb = output_tb
  )

  return(temp_n)
} # nocov end

geocodebr_message <- function(message, .envir = parent.frame()) {
  message_call <- sys.call(-1)
  message_function <- as.name(message_call[[1]])

  message_classes <- c(
    paste0("geocodebr_message_", sub("^message_", "", message_function)),
    "geocodebr_message"
  )

  cli::cli_inform(message, class = message_classes, .envir = .envir)
}

message_standardizing_addresses <- function() {
  geocodebr_message(c("i" = "Padronizando endere\u00e7os de entrada"))
}

message_baixando_cnefe <- function() {
  geocodebr_message(c("i" = "Baixando dados do CNEFE"))
}

message_usando_cnefe_local <- function() {
  geocodebr_message(c("i" = "Utilizando dados do CNEFE armazenados localmente"))
}

message_looking_for_matches <- function() {
  geocodebr_message(c("i" = "Geolocalizando endere\u00e7os"))
}

message_preparando_output <- function() {
  geocodebr_message(c("i" = "Preparando resultados"))
}

message_cache <- function() {
  geocodebr_message(c("i" = "Nenhum dado em cache local"))
}

parent_function <- function() message_test()

message_test <- function() geocodebr_message(c("*" = "info"))

test_that("message_works_correctly", {
  expect_message(parent_function(), class = "geocodebr_message_test")
  expect_message(parent_function(), class = "geocodebr_message")

  expect_snapshot(parent_function(), cnd_class = TRUE)

  expect_message(message_standardizing_addresses())
  expect_message(message_baixando_cnefe())
  expect_message(message_looking_for_matches())
  expect_message(message_preparando_output())
})


# error
test_that("message_does_NOT_work_correctly", {
  expect_error(message_test("banana"))
})

# Find cases nearby

serch_nearby_addresses <- function(con, dist){

  # update coordinates range
  margin_lat <- round(dist / 111320, digits = 8)
  # AND lon_max = lon + ({dist} * COS( lat * 0.017453 ))

  query_update_coords_range <- glue::glue(
    "UPDATE input_table_db
      SET lat_min = lat - {margin_lat},
          lat_max = lat + {margin_lat},
          lon_min = lon - (({dist} / 111320) * COS(lat)),
          lon_max = lon + (({dist} / 111320) * COS(lat));"
    )

  DBI::dbExecute(con, query_update_coords_range)
  # a <- DBI::dbReadTable(con, 'input_table_db')

  # spatial search
  query_filter_cases_nearby <- glue::glue(
    "INSERT INTO output_db (tempidgeocodebr, lat_cnefe, lon_cnefe, endereco_completo,
                            estado, municipio, logradouro, numero, cep, localidade)
      SELECT
          input_table_db.tempidgeocodebr,
          filtered_cnefe.lat AS lat_cnefe,
          filtered_cnefe.lon AS lon_cnefe,
          filtered_cnefe.endereco_completo,
          filtered_cnefe.estado,
          filtered_cnefe.municipio,
          filtered_cnefe.logradouro,
          filtered_cnefe.numero,
          filtered_cnefe.cep,
          filtered_cnefe.localidade
      FROM
        input_table_db, filtered_cnefe
      WHERE
            filtered_cnefe.lat BETWEEN input_table_db.lat_min AND input_table_db.lat_max
        AND filtered_cnefe.lon BETWEEN input_table_db.lon_min AND input_table_db.lon_max
        AND filtered_cnefe.lon IS NOT NULL;"
    )

      # FROM input_table_db
      # LEFT JOIN filtered_cnefe
      #     ON filtered_cnefe.lon BETWEEN input_table_db.lon_min AND input_table_db.lon_max
      #     AND filtered_cnefe.lat BETWEEN input_table_db.lat_min AND input_table_db.lat_max
      # WHERE filtered_cnefe.lon IS NOT NULL;"

  DBI::dbExecute(con, query_filter_cases_nearby)
  # b <- DBI::dbReadTable(con, 'output_db')

          # a <- DBI::dbGetQuery(con, query_filter_cases_nearby)
          # data.table::setDT(a)
          # a[, distancia_metros := dt_haversine(lat,lon , lat_cnefe, lon_cnefe)]
          # summary(a$distancia_metros)


  # UPDATE input_padrao_db: Remove observations found in previous step
  temp_n <- update_input_db(
    con,
    update_tb = 'input_table_db',
    reference_tb = 'output_db'
  )

  return(temp_n)
}

devtools::load_all('.')
library(data.table)
library(dplyr)

# open input data
data_path <- system.file("extdata/large_sample.parquet", package = "geocodebr")
input_df <- arrow::read_parquet(data_path)


input_df$municipio <- enderecobr::padronizar_municipios(input_df$municipio)
input_df$estado <- enderecobr::padronizar_estados(input_df$uf)

# df = input_df
# N_groups = 4

balanced_grouping_dt <- function(df, N_groups) {

  # Convert to data.table (without copying if already DT)
  dt <- as.data.table(df)
  #setDT(df)

  # 1. Count rows per municipality block
  # 2. Sort blocks by size (largest first for best balancing)
  muni_sizes <- df[, .(n_rows = .N), by = .(estado, municipio)][order(-n_rows)]


  # 3. Greedy assignment of blocks to groups
  group_loads <- integer(N_groups)
  muni_sizes[, group := NA_integer_]


  for (i in seq_len(nrow(muni_sizes))) {
    # choose group with fewest rows so far
    g <- which.min(group_loads)
    muni_sizes$group[i] <- g
    group_loads[g] <- group_loads[g] + muni_sizes$n_rows[i]
  }


  # 4. Join group assignment back to original data
  df[muni_sizes, on = .(estado, municipio), "group" := i.group]

  return(df)
}

balanced_grouping_dplyr <- function(df, N_groups) {
  stopifnot(all(c("estado", "municipio") %in% names(df)))

  # 1. Count rows per municipality block
  # 2. Sort blocks by size (descending)
  muni_sizes <- df %>%
    count(estado, municipio, name = "n_rows") %>%
    arrange(desc(n_rows))

  # 3. Balanced assignment of blocks to groups
  # Track group loads
  group_loads <- rep(0, N_groups)
  muni_sizes$group <- NA_integer_


  for (i in seq_len(nrow(muni_sizes))) {
    # choose group with fewest rows so far
    g <- which.min(group_loads)
    muni_sizes$group[i] <- g
    group_loads[g] <- group_loads[g] + muni_sizes$n_rows[i]
  }


  # 4. Join assignment back to original data
  df_out <- df %>%
    left_join(muni_sizes %>% select(estado, municipio, group),
              by = c("estado", "municipio"))

  return(df_out)
}




df_dplyr <- balanced_grouping_dplyr(df = input_df, N_groups = 4)
df_dt <- balanced_grouping_dt(df = input_df, N_groups = 4)

n = 10
bench::mark(check = F, relative = T,

  dpylr =  balanced_grouping_dplyr(df = input_df, N_groups = n),
  dt = balanced_grouping_dt(df = input_df, N_groups = n)
)


n <- parallelly::availableCores() /2
df <- balanced_grouping_dt(df = input_df, N_groups = n)


# parallel callr --------------------------------------


library(future)
library(furrr)

future::plan(future::multisession(workers = n))


bench::bench_time(
  a <-   split(df, f = "group") |>
    furrr::future_map(
      .f = function(x){
        geocode(
          enderecos = x,
          campos_endereco = campos,
          n_cores = 1,
          resultado_completo = F,
          resolver_empates = T
        )
      }
    )
)

#' Especifica as colunas que descrevem os campos dos endereços
#'
#' Cria um vetor de caracteres especificando as colunas que representam cada
#' campo do endereço na tabela de endereços. Os campos `estado` e `municipio`
#' são obrigatórios.

#' @param estado Uma string. O nome da coluna que representa o estado do
#'   endereço. Campo obrigatório. Na tabela de endereços, essa coluna pode
#'   conter os nomes dos estados por extenso, ou a abrevição oficial dos estados
#'   com duas letras, e.g. "AM", "SP", "DF", "RJ".
#' @param municipio Uma string. O nome da coluna que representa o município do
#'   endereço. Campo obrigatório. Na tabela de endereços, essa coluna pode
#'   conter o nome dos municípios, ou o seu código IBGE de 7 dígitos.
#' @param logradouro Uma string. O nome da coluna que representa o *logradouro*
#'   (endereço da rua) do endereço. Pode ser `NULL` se o campo não estiver
#'   especificado na tabela de endereços. Na tabela de endereços, essa coluna
#'   deve incluir o *tipo* do lograoduro, indicando se trata-se de uma "Rua" ou
#'   "Avenida" etc, por exemplo "Avenida Presidente Getúlio Vargas". Além disso,
#'   essa coluna *não* deve incluir o `numero` do endereço, pois o número deve
#'   ser indicado numa coluna separada.
#' @param numero Uma string. O nome da coluna que representa o número do endereço.
#'    Pode ser `NULL` se o campo não estiver especificado na tabela de endereços.
#'    Na tabela de endereços, valores como `0` ou caracteres não numerciso como
#'    `"S/N"` ou `"10a"` são considerados como `NA`.
#' @param cep Uma string. O nome da coluna que representa o *CEP* (Código de
#'    Endereçamento Postal) do endereço. Pode ser `NULL` se o campo não estiver
#'    especificado na tabela de endereços.
#' @param localidade Uma string. O nome da coluna que representa a localidade
#'   (equivalente ao 'bairro' em áreas urbanas) do endereço. Pode ser `NULL` se
#'   esse campo não estiver presente na tabela de endereços.
#'
#' @return Um vetor de caracteres no qual os nomes são os campos do endereço e os
#'   valores são as colunas que os representam na tabela de endereços.
#'
#' @examples
#' definir_campos(
#'   logradouro = "Nome_logradouro",
#'   numero = "Numero",
#'   cep = "CEP",
#'   localidade = "Bairro",
#'   municipio = "Cidade",
#'   estado = "UF"
#' )
#'
#' @export
definir_campos <- function(
  estado,
  municipio,
  logradouro = NULL,
  numero = NULL,
  cep = NULL,
  localidade = NULL
) {
  col <- checkmate::makeAssertCollection()
  checkmate::assert_string(logradouro, null.ok = TRUE, add = col)
  checkmate::assert_string(numero, null.ok = TRUE, add = col)
  checkmate::assert_string(cep, null.ok = TRUE, add = col)
  checkmate::assert_string(localidade, null.ok = TRUE, add = col)
  checkmate::assert_string(municipio, null.ok = TRUE, add = col)
  checkmate::assert_string(estado, null.ok = TRUE, add = col)
  checkmate::reportAssertions(col)

  address_fields <- c(
    logradouro = logradouro,
    numero = numero,
    cep = cep,
    localidade = localidade,
    municipio = municipio,
    estado = estado
  )

  if (is.null(address_fields)) {
    error_null_address_fields()
  }

  return(address_fields)
}

error_null_address_fields <- function() {
  geocodebr_error(
    "Pelo menos um campo n\u00e3o pode ser nulo {.code NULL}.",
    call = rlang::caller_env()
  )
}

tester <- function(logradouro = NULL,
                   numero = NULL,
                   cep = NULL,
                   localidade = NULL,
                   municipio = NULL,
                   estado = NULL) {
  definir_campos(
    logradouro,
    numero,
    cep,
    localidade,
    municipio,
    estado
  )
}

test_that("definir_campos expected behavior", {

  testthat::expect_vector(
    test <- tester(
    logradouro = 'nm_logradouro',
    numero = 'Numero',
    cep = 'Cep',
    localidade = 'Bairro',
    municipio = 'nm_municipio',
    estado = 'nm_uf')
  )

  testthat::expect_true(
    all(names(test) == c("logradouro","numero", "cep",
                         "localidade", "municipio", "estado"))
  )


})

test_that("errors with incorrect input", {
  expect_error(tester(logradouro = 1))
  expect_error(tester(logradouro = c("aaa", "bbb")))

  expect_error(tester(numero = 1))
  expect_error(tester(numero = c("aaa", "bbb")))

  expect_error(tester(cep = 1))
  expect_error(tester(cep = c("aaa", "bbb")))

  expect_error(tester(localidade = 1))
  expect_error(tester(localidade = c("aaa", "bbb")))

  expect_error(tester(municipio = 1))
  expect_error(tester(municipio = c("aaa", "bbb")))

  expect_error(tester(estado = 1))
  expect_error(tester(estado = c("aaa", "bbb")))
})

test_that("errors when all fields are NULL", {
  expect_error(tester(), class = "geocodebr_error_null_address_fields")

  expect_snapshot(tester(), error = TRUE, cnd_class = TRUE)
})

# test_that("returns a character vector", {
#   expect_identical(
#     tester(
#       logradouro = "Nome_logradouro",
#       numero = "Numero",
#       cep = "CEP",
#       localidade = "Bairro",
#       municipio = "Cidade",
#       estado = "UF"
#     ),
#     c(
#       logradouro = "Nome_logradouro",
#       numero = "Numero",
#       cep = "CEP",
#       localidade = "Bairro",
#       municipio = "Cidade",
#       estado = "UF"
#     )
#   )
#
#   expect_identical(
#     tester(logradouro = "oi", numero = "ola"),
#     c(logradouro = "oi", numero = "ola")
#   )
# })

devtools::load_all('.')
library(tictoc)
library(dplyr)
library(data.table)
library(ipeadatalake)
set.seed(42)

sample_size <- 10000

data_path <- system.file("extdata/small_sample.csv", package = "geocodebr")
input_df <- read.csv(data_path)

# select and reorder columns
input_df <- input_df |>
  select(
    id = id,
    logradouro = nm_logradouro,
    numero = Numero,
    bairro = Bairro,
    cep = Cep,
    municipio = nm_municipio,
    uf = nm_uf
  )

head(input_df)

# rais --------------------------------------------------------------------

rais <- ipeadatalake::ler_rais(
  ano = 2019,
  tipo = 'estabelecimento',
  as_data_frame = T,
  colunas = c("id_estab", "logradouro",
              "bairro", "codemun", "uf", "cep")) |>
  filter(uf %in% c(12, 27, 33)) |> # only states of Acre, Alagoas and RJ
  compute() |>
  dplyr::slice_sample(n = sample_size) |> # sample 50K
  filter(uf != "IG") |>
  filter(uf != "") |>
  collect()

data.table::setDT(rais)

# create column number
rais[, numero := gsub("[^0-9]", "", logradouro)]

# remove numbers from logradouro
rais[, logradouro := gsub("\\d+", "", logradouro)]
rais[, logradouro := gsub(",", "", logradouro)]


# select and reorder columns
rais <- rais |>
  select(
    id = id_estab,
    logradouro = logradouro,
    numero = numero,
    bairro = bairro,
    cep = cep,
    municipio = codemun,
    uf = uf
  )

head(rais)






# cad unico --------------------------------------------------------------------

cad <- ipeadatalake::ler_cadunico(
  data = 201912,
  tipo = 'familia',
  as_data_frame = F,
  colunas = c("co_familiar_fam", "co_uf", "cd_ibge_cadastro",
              "no_localidade_fam", "no_tip_logradouro_fam",
              "no_tit_logradouro_fam", "no_logradouro_fam",
              "nu_logradouro_fam", "ds_complemento_fam",
              "ds_complemento_adic_fam",
              "nu_cep_logradouro_fam", "co_unidade_territorial_fam",
              "no_unidade_territorial_fam", "co_local_domic_fam")
)

# a <- tail(cad, n = 100) |> collect()

# compose address fields
cad <- cad |>
  filter(co_uf %in% c(12, 27, 33)) |> # only states of Acre, Alagoas and RJ
  compute() |>
  dplyr::slice_sample(n = sample_size) |> # sample 20K
  mutate(no_tip_logradouro_fam = ifelse(is.na(no_tip_logradouro_fam), '', no_tip_logradouro_fam),
         no_tit_logradouro_fam = ifelse(is.na(no_tit_logradouro_fam), '', no_tit_logradouro_fam),
         no_logradouro_fam = ifelse(is.na(no_logradouro_fam), '', no_logradouro_fam)
  ) |>
  mutate(abbrev_state = co_uf,
         code_muni = cd_ibge_cadastro,
         logradouro = paste(no_tip_logradouro_fam, no_tit_logradouro_fam, no_logradouro_fam),
         numero = nu_logradouro_fam,
         cep = nu_cep_logradouro_fam,
         bairro = no_localidade_fam) |>
  select(co_familiar_fam,
         abbrev_state,
         code_muni,
         logradouro,
         numero,
         cep,
         bairro) |>
  dplyr::collect()



cad <- cad |>
  select(
    id = co_familiar_fam,
    logradouro = logradouro,
    numero = numero,
    bairro = bairro,
    cep = cep,
    municipio = code_muni,
    uf = abbrev_state
  )



df <- rbind(input_df, cad, rais)

setDT(df)
df[, id := 1:nrow(df)]
head(df)

arrow::write_parquet(df, './inst/extdata/large_sample.parquet')

create_progress_bar <- function(standard_locations, .envir = parent.frame()) {
  cli::cli_progress_bar(
    total = nrow(standard_locations),
    format = " Casos processados: {formatC(cli::pb_current, big.mark = ',', format = 'd')}/{formatC(cli::pb_total, big.mark = ',', format = 'd')} {cli::pb_bar} {cli::pb_percent} - {cli::pb_status}",
    clear = FALSE,
    .envir = .envir
  )
}

update_progress_bar <- function(
  matched_rows,
  formatted_case,
  .envir = parent.frame()
) {
  cli::cli_progress_update(
    set = matched_rows,
    status = glue::glue("{formatted_case}"),
    force = TRUE,
    .envir = .envir
  )
}

finish_progress_bar <- function(matched_rows, .envir = parent.frame()) {
  cli::cli_progress_update(
    set = matched_rows,
    status = "Fim!",
    force = TRUE,
    .envir = .envir
  )
}

#' @details Lidando com casos de empate:
#'
#' No processo de geolocalização de dados, é possível que para alguns endereços
#' de input sejam encontrados diferentes coordenadas possíveis (e.g. duas ruas
#' diferentes com o mesmo nome, mas em bairros distintos em uma mesma cidade).
#' Esses casos são trados como empate'. Quando a função `geocode()` recebe o
#' o parâmetro `resolver_empates = TRUE`, os casos de empate são resolvidos
#' automaticamente pela função. A solução destes empates é feita da seguinte
#' maneira:
#'
#' 1) Quando se encontra diferente coordenadas possíveis para um mesmo endereço
#' de input, nós assumimos que essas coordendas pertencem provavelmente a endereços
#' diferentes se (a) estas coordenadas estão a mais de 1Km entre si, ou (b) estão
#' associadas a um logradouro 'ambíguo', i.e. que costumam se repetir em muitos
#' bairros (e.g. "RUA A", "RUA QUATRO", "RUA 10", etc). Nestes casos, a solução
#' de desempate é retornar o ponto com maior número de estabelecimentos no CNEFE,
#' valor indicado na coluna `"contagem_cnefe"`.
#'
#' 2) Quando as coordenadas possivelmente associadas a um endereço estão a menos
#' de 1Km entre si e não se trata de um logradouro 'ambíguo', nós assumimos que
#' os pontos pertencem provavelmente ao mesmo logradouro (e.g. diferentes CEPs
#' ao longo de uma mesma rua). Nestes casos, a solução de desempate é retornar
#' um ponto que resulta da média das coordenadas dos pontos possíveis ponderada
#' pelo valor de `"contagem_cnefe"`. Nesse caso, a coluna de output
#' `"endereco_encontrado"` recebe valor do ponto com maior `"contagem_cnefe"`.
#'

#' Geocode reverso de coordenadas espaciais no Brasil
#'
#' @description
#' Geocode reverso de coordenadas geográficas para endereços. A função recebe um
#' `sf data frame` com pontos e retorna o endereço mais próximo dando uma
#' distância máxima de busca.
#'
#' @param pontos Uma tabela de dados com classe espacial `sf data frame` no
#'        sistema de coordenadas geográficas SIRGAS 2000, EPSG 4674.
#' @param dist_max Integer. Distancia máxima aceitável (em metros) entre os
#'        pontos de input e o endereço Por padrão, a distância é de 1000 metros.
#' @template verboso
#' @template cache
#' @template n_cores
#'
#' @return Retorna o `sf data.frame` de input adicionado das colunas do endereço
#'         encontrado. O output inclui uma coluna "distancia_metros" que indica
#'         a distância entre o ponto de input e o endereço mais próximo
#'         encontrado.
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' library(geocodebr)
#' library(sf)
#'
#' # ler amostra de dados
#' pontos <- readRDS(
#'     system.file("extdata/pontos.rds", package = "geocodebr")
#'     )
#'
#' pontos <- pontos[1:3,]
#'
#' # geocode reverso
#' df_enderecos <- geocodebr::geocode_reverso(
#'   pontos = pontos,
#'   dist_max = 800,
#'   verboso = TRUE
#'   )
#'
#'head(df_enderecos)
#'
#' @export
geocode_reverso <- function(
    pontos,
    dist_max = 1000,
    verboso = TRUE,
    cache = TRUE,
    n_cores = NULL
  ){

  # check input
  checkmate::assert_class(pontos, 'sf')
  checkmate::assert_number(dist_max, lower = 500, upper = 100000) # max 100 Km
  checkmate::assert_logical(verboso)
  checkmate::assert_logical(cache)

  # check if geometry type is POINT
  if (any(sf::st_geometry_type(pontos) != 'POINT')) {
    cli::cli_abort(
      "Input precisa ser um sf data frame com geometria do tipo POINT."
    )
  }

  epsg <- sf::st_crs(pontos)$epsg
  if (epsg != 4674) {
    cli::cli_abort(
      "Dados de input precisam estar com sistema de coordenadas geogr\u00e1ficas SIRGAS 2000, EPSG 4674."
    )
  }

  # pontos <- sf::st_transform(pontos, 4674)


  # prep input -------------------------------------------------------

  # converte pontos de input para data.frame
  pontos$tempidgeocodebr <- 1:nrow(pontos)


  # check if input falls within Brazil
  bbox <- sf::st_bbox(pontos)

  bbox_brazil <- data.frame(
    xmin = -73.99044997,
    ymin = -33.75208127,
    xmax = -28.83594354,
    ymax = 5.27184108
  )

  error_msg <- 'Coordenadas de input localizadas fora do bounding box do Brasil.'
  if (
    bbox[[1]] < bbox_brazil$xmin |
    bbox[[3]] > bbox_brazil$xmax |
    bbox[[2]] < bbox_brazil$ymin |
    bbox[[4]] > bbox_brazil$ymax
  ) {
    cli::cli_abort(error_msg)
  }

  # download cnefe  -------------------------------------------------------

  # downloading cnefe
  cnefe_dir <- geocodebr::download_cnefe(
    tabela = 'municipio_logradouro_numero_cep_localidade',
    verboso = verboso,
    cache = cache
  )

  # creating a temporary db and register the input table data
  conn <- create_geocodebr_db(
    n_cores = n_cores,
    load_spatial = TRUE
  )

  # limita escopo de busca aos municipios  -------------------------------------------------------
  # determine potential municipalities
  munis <- system.file("extdata/munis_bbox_2022.parquet", package = "geocodebr") |>
    arrow::open_dataset() |>
    sf::st_as_sf()

  # place holder to use geoarrow becaue:
  #   Namespace in Imports field not imported from: 'geoarrow'
  #        All declared Imports should be used.
  geoarrow::as_geoarrow_vctr("POINT (0 1)")

  # munis_path <- system.file("extdata/munis_2022.parquet", package = "geocodebr")
  #
  # query_register_muni <- glue::glue(
  #   "CREATE OR REPLACE TEMP VIEW munis AS
  #       SELECT *,
  #       geometry::GEOMETRY AS geometry
  #   FROM read_parquet('{munis_path}');"
  # )
  #
  # DBI::dbExecute(conn, query_register_muni)

  potential_munis <- duckspatial::ddbs_join(
    x = pontos,
    y = munis,
    join = "within",
    quiet = TRUE
  ) |>
    dplyr::pull(code_muni) |>
    unique()

  potential_munis <- enderecobr::padronizar_municipios(potential_munis)

  # lida com munis com apostrofe no nome tipo Olho d'agua
  potential_munis <- gsub("'", "''", potential_munis, fixed = TRUE)

  unique_munis <- paste(glue::glue("'{potential_munis}'"), collapse = ",")


  # build path to local file
  path_to_parquet <- fs::path(
    listar_pasta_cache(),
    glue::glue("geocodebr_data_release_{data_release}"),
    paste0("municipio_logradouro_numero_cep_localidade.parquet")
  )

  # create filtered_cnefe table, filter on the fly
  cols_to_keep <- c(
    "estado",
    "municipio",
    "logradouro",
    "numero",
    "cep",
    "localidade"
  )

  cols_to_keep_string <- paste0(cols_to_keep, collapse = ", ")

  # Load CNEFE data and filter it to narrow search global scope of cnefe to the
  # bounding box of input points
  query_filter_cnefe <- glue::glue(
    "CREATE OR REPLACE TEMP VIEW cnefe_tb AS
        SELECT {cols_to_keep_string},
               ST_Point(lon, lat)::GEOMETRY('EPSG:4674') AS geometry
     FROM read_parquet('{path_to_parquet}') m
          WHERE m.municipio IN ({unique_munis});"
  )

  DBI::dbExecute(conn, query_filter_cnefe)
  # b <- duckspatial::ddbs_read_table(conn = conn, name = "cnefe_tb")
  # duckspatial::ddbs_crs(con, "filtered_cnefe")
  # ST_Point(lon, lat)::GEOMETRY('EPSG:4674') AS geom


  cnefe_utm_duck <-  duckspatial::ddbs_transform(
    x = 'cnefe_tb',
    y = 'EPSG:31983',conn = conn,
    quiet = TRUE
  )

  # input to UTM
  input_utm_duck <-  duckspatial::ddbs_transform(
    x = pontos,
    y = 'EPSG:31983',
    quiet = TRUE
  )

  # buffers around input points
  buff <- duckspatial::ddbs_buffer(
    x = input_utm_duck,
    distance = dist_max,
    quiet = TRUE
  )

  # Lazy join - computation stays in DuckDB
  suppressWarnings(
    result <- duckspatial::ddbs_join(
      x = cnefe_utm_duck,
      y = buff,
      join = "within",
      quiet = TRUE
    )
  )

  # write to connection
  duckspatial::ddbs_write_table(
    conn = conn,
    data = input_utm_duck,
    name = "pontos_utm",
    overwrite = T,
    temp_view = T,
    quiet = TRUE
  )

  duckspatial::ddbs_write_table(
    conn = conn,
    data = result,
    name = "join_result",
    overwrite = T,
    temp_view = T,
    quiet = TRUE
  )

  # Get column names from both tables
  cols_a <- DBI::dbGetQuery(conn, "SELECT column_name FROM (DESCRIBE pontos_utm)")$column_name
  cols_b <- DBI::dbGetQuery(conn, "SELECT column_name FROM (DESCRIBE join_result)")$column_name

  # Find overlapping columns (plus the ones you want to exclude anyway)
  exclude_from_b <- unique(c(
    "tempidgeocodebr", "geometry",
    intersect(cols_a, cols_b)
  ))

  exclude_clause <- paste0("EXCLUDE (", paste(exclude_from_b, collapse = ", "), ")")

  query <- glue::glue("
    CREATE OR REPLACE TEMP TABLE output AS
    SELECT * EXCLUDE (rn)
    FROM (
      SELECT
        a.*,
        b.* {exclude_clause},
        ST_Distance(a.geometry, b.geometry) AS distancia_metros,
        ROW_NUMBER() OVER (
          PARTITION BY a.id
          ORDER BY ST_Distance(a.geometry, b.geometry)
        ) AS rn
      FROM pontos_utm AS a
      JOIN join_result AS b
        ON a.tempidgeocodebr = b.tempidgeocodebr
    ) t
    WHERE rn = 1
    ORDER BY tempidgeocodebr
    ")

  DBI::dbExecute(conn, query)
  # output <- duckspatial::ddbs_read_table(conn, name = "output")
  # head(output)

  output <- duckspatial::ddbs_transform(
    x = "output",
    conn = conn,
    y = 'EPSG:4674',
    quiet = TRUE
  )

  output <- duckspatial::ddbs_collect(output)


  # TODO
  if (nrow(output)==0){
    stop("Nenhum endereco proximo foi encontrados")
  }


  output <- output |>
    dplyr::select(-tempidgeocodebr) |>
    dplyr::relocate(geometry, .after = dplyr::last_col())

  duckdb::dbDisconnect(conn)

  return(output)
}

# skip tests because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("arrow")


# amostra de pontos espaciais
points <- readRDS(
  system.file("extdata/pontos.rds", package = "geocodebr")
)

points <- points[1:10,]


tester <- function(pontos = points,
                   dist_max = 1000,
                   verboso = FALSE,
                   cache = TRUE,
                   n_cores = 1) {
  geocode_reverso(
    pontos = pontos,
    dist_max = dist_max,
    verboso = verboso,
    cache = cache,
    n_cores = n_cores
  )
}

test_that("expected output", {

  # radio de busca de 1 Km
  testthat::succeed(std_output <- tester())
  testthat::expect_true(nrow(std_output) == 1)

  # radio de busca de 5 Km
  testthat::succeed(std_output_5K <- tester(dist_max = 5000))
  testthat::expect_true(nrow(std_output_5K) == 8)

  # output in sf format
  testthat::expect_true(is(std_output , 'sf'))
})





test_that("errors with incorrect input", {

  # input nao eh sf
  testthat::expect_error(tester(unclass(points)))

  # input tem geometry diferente de POINT
  testthat::expect_error( tester(pontos = sf::st_cast(points, "LINESTRING")) )

  # input na projecao espacial errada
  testthat::expect_error(tester(sf::st_transform(points, 4326)))

  # input fora do bbox do Brazil
  p <- sf::st_sfc(sf::st_point(c(53.12682, 25.61657)))
  p <- sf::st_sf(p)
  sf::st_crs(p) <- 4674
  testthat::expect_error(tester(p))



  testthat::expect_error(tester(dist_max = 'A'))
  testthat::expect_error(tester(dist_max = NA))
  testthat::expect_error(tester(dist_max = TRUE))

  testthat::expect_error(tester(n_cores = "a"))
  testthat::expect_error(tester(n_cores = 0))
  testthat::expect_error(tester(n_cores = Inf))

  testthat::expect_error(tester(verboso = 1))
  testthat::expect_error(tester(verboso = NA))
  testthat::expect_error(tester(verboso = c(TRUE, TRUE)))

  testthat::expect_error(tester(cache = 1))
  testthat::expect_error(tester(cache = NA))
  testthat::expect_error(tester(cache = c(TRUE, TRUE)))
})

devtools::load_all('.')
library(dplyr)
library(geoarrow)


# input data

pontos <- readRDS(
   system.file("extdata/pontos.rds", package = "geocodebr")
   )

pontos <- pontos[1:500,]



# reverse geocode
bench::system_time(
 out <-  geocode_reverso(
   pontos = pontos,
   dist_max = 1000
   )
)
View(out)


# ttt <- data.frame(id=1, lat=-15.814192047159876, lon=-47.90534614672923)
# reverse_geocode(df = ttt)

# take aways
# ok reverse_geocode_filter  # mais rapdido e eficiente, mas sem progress bar
# ok reverse_geocode_join    # igual o _filter, mas usa join
# ok reverse_geocode_hybrid  # com progress bar mas um pouco mais lento e bem mais memoria
# ok reverse_geocode_arrow   # tempo igual a _hybrid, mas usa bem mais memoria
# ok           filterloop    # disparado o mais lento, com progress e memoria media

# essa funcao pode fica muito mais rapida / eficiente se usarmos a biblioteca de
# dados espaciais do duckdb

b5 <- bench::mark(
  current = geocode_reverso(pontos = pontos, dist_max = 1000),
  iterations = 5,
  check = F
)

b5
#

# # 500 pontos
#     expression           min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory
#     <bch:expr>        <bch:> <bch:>     <dbl> <bch:byt>    <dbl> <int> <dbl>   <bch:tm> <list> <list>
#   1 duck_filter4       1.54m  1.62m   0.0101    221.5MB  0.0423      5    21      8.28m <NULL> <Rprofmem>
#   2 duck_filter_loop4  5.35m  6.88m   0.00255    14.5MB  0.00357     5     7      32.7m <NULL> <Rprofmem>
#   3 hybrid4            2.41m  2.47m   0.00663    34.5MB  0.302       5   228     12.56m <NULL> <Rprofmem>

# # 1000 pontos
#     expression   min median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory
#     <bch:expr> <bch> <bch:>     <dbl> <bch:byt>    <dbl> <int> <dbl>   <bch:tm> <list> <list>
#   1 duck_filt… 2.97m  2.97m   0.00560    11.3MB    0         1     0      2.97m <NULL> <Rprofmem>
#   2 duck_join4 3.03m  3.03m   0.00550    11.5MB    0         1     0      3.03m <NULL> <Rprofmem>
#   3 arrow4     4.27m  4.27m   0.00391   240.1MB    0.316     1    81      4.27m <NULL> <Rprofmem>
#   4 hybrid4    4.24m  4.24m   0.00393   122.4MB    0.110     1    28      4.24m <NULL> <Rprofmem>
#   4 filterloop 10.7m 11.27m   0.00146    19.8MB  0.00195     3     4     34.19m <NULL> <Rprofmem> <bench_tm [3]> <tibble>
# 1 current      1.92s  1.98s     0.496  247.62MB    3.17      5    32     10.08s <NULL> <Rprofmem [22,026 × 3]> <bench_tm [5]> <tibble>
# 2 geocrev2     1.28s  1.33s     0.730    8.03MB    0.730     5     5      6.85s <NULL> <Rprofmem [13,876 × 3]> <bench_tm [5]> <tibble>
# 3 geocrev3     1.38s  1.61s     0.626    8.29MB    0.752     5     6      7.98s <NULL> <Rprofmem [15,238 × 3]> <bench_tm [5]> <tibble>
#



# aternativas da funcao de geocode reverso -----------------------------------------------------------

#' muni join + haversine
#' (usa spatial join para detectar munis candidatos, e depois calcular haversine na unha)
#' a diferenca dessa para a versao implementada é que a implementada calcula distancias dentro do duckspatial com ST_DIST

geocode_reverso2 <- function(
    pontos,
    dist_max = 1000,
    verboso = TRUE,
    cache = TRUE,
    n_cores = NULL
) {
  # check input
  checkmate::assert_class(pontos, 'sf')
  checkmate::assert_number(dist_max, lower = 500, upper = 100000) # max 100 Km
  checkmate::assert_logical(verboso)
  checkmate::assert_logical(cache)

  # check if geometry type is POINT
  if (any(sf::st_geometry_type(pontos) != 'POINT')) {
    cli::cli_abort(
      "Input precisa ser um sf data frame com geometria do tipo POINT."
    )
  }

  epsg <- sf::st_crs(pontos)$epsg
  if (epsg != 4674) {
    cli::cli_abort(
      "Dados de input precisam estar com sistema de coordenadas geogr\u00e1ficas SIRGAS 2000, EPSG 4674."
    )
  }

  # prep input -------------------------------------------------------

  # converte para data.frame
  coords <- sfheaders::sf_to_df(pontos, fill = TRUE)
  data.table::setDT(coords)
  coords[, c('sfg_id', 'point_id') := NULL]
  data.table::setnames(coords, old = c('x', 'y'), new = c('lon', 'lat'))

  # create temp id
  coords[, tempidgeocodebr := 1:nrow(coords)]

  # convert max_dist to degrees
  # 1 degree of latitude is always 111320 meters
  margin_lat <- dist_max / 111320

  # 1 degree of longitude is 111320 * cos(lat)
  coords[, c("lat_min", "lat_max") := .(lat - margin_lat, lat + margin_lat)]

  coords[,
         c("lon_min", "lon_max") := .(
           lon - dist_max / 111320 * cos(lat),
           lon + dist_max / 111320 * cos(lat)
         )
  ]

  # get bounding box around input points
  # using a range of max dist around input points
  bbox_lat_min <- min(coords$lat_min)
  bbox_lat_max <- max(coords$lat_max)
  bbox_lon_min <- min(coords$lon_min)
  bbox_lon_max <- max(coords$lon_max)

  # check if input falls within Brazil
  bbox_brazil <- data.frame(
    xmin = -73.99044997,
    ymin = -33.75208127,
    xmax = -28.83594354,
    ymax = 5.27184108
  )

  error_msg <- 'Coordenadas de input localizadas fora do bounding box do Brasil.'
  if (
    bbox_lon_min < bbox_brazil$xmin |
    bbox_lon_max > bbox_brazil$xmax |
    bbox_lat_min < bbox_brazil$ymin |
    bbox_lat_max > bbox_brazil$ymax
  ) {
    cli::cli_abort(error_msg)
  }

  # download cnefe  -------------------------------------------------------

  # downloading cnefe
  cnefe_dir <- geocodebr::download_cnefe(
    tabela = 'municipio_logradouro_numero_cep_localidade',
    verboso = verboso,
    cache = cache
  )

  # creating a temporary db and register the input table data
  con <- create_geocodebr_db(n_cores = n_cores)


  # limita escopo de busca aos municipios  -------------------------------------------------------
  # determine potential municipalities
  munis <- system.file("extdata/munis_bbox_2022.parquet", package = "geocodebr") |>
    arrow::open_dataset() |>
    sf::st_as_sf()
  # munis_path <- system.file("extdata/munis_2022.parquet", package = "geocodebr")
  #
  # query_register_muni <- glue::glue(
  #   "CREATE OR REPLACE TEMP VIEW munis AS
  #       SELECT *,
  #       geometry::GEOMETRY AS geometry
  #   FROM read_parquet('{munis_path}');"
  # )
  #
  # DBI::dbExecute(conn, query_register_muni)

  potential_munis <- duckspatial::ddbs_join(
    x = pontos,
    y = munis,
    join = "within",
    quiet = TRUE
  ) |>
    dplyr::pull(code_muni) |>
    unique()

  potential_munis <- enderecobr::padronizar_municipios(potential_munis)

  # lida com munis com apostrofe no nome tipo Olho d'agua
  potential_munis <- gsub("'", "''", potential_munis, fixed = TRUE)

  unique_munis <- paste(glue::glue("'{potential_munis}'"), collapse = ",")

  # build path to local file
  path_to_parquet <- fs::path(
    listar_pasta_cache(),
    glue::glue("geocodebr_data_release_{data_release}"),
    paste0("municipio_logradouro_numero_cep_localidade.parquet")
  )

  # create filtered_cnefe table, filter on the fly
  cols_to_keep <- c(
    "estado",
    "municipio",
    "logradouro",
    "numero",
    "cep",
    "localidade",
    "endereco_completo",
    "lon",
    "lat"
  )
  cols_to_keep <- paste0(cols_to_keep, collapse = ", ")

  # Load CNEFE data and filter it to include only municipalities
  # present in the input table, reducing the search scope
  # Narrow search global scope of cnefe to bounding box
  query_filter_cnefe <- glue::glue(
    "CREATE TEMP VIEW filtered_cnefe AS
        SELECT {cols_to_keep}
        FROM read_parquet('{path_to_parquet}') m
          WHERE m.municipio IN ({unique_munis});"
  )

  DBI::dbExecute(con, query_filter_cnefe)
  # DBI::dbExecute(con, query_filter_cnefe)
  # b <- DBI::dbReadTable(con, "filtered_cnefe")

  # Convert input data frame to DuckDB table
  duckdb::dbWriteTable(con, "input_table_db", coords, temporary = TRUE)

  # Haversine macro (kept for speed; consider spatial extension later)
  DBI::dbExecute(
    con,
    "
    CREATE MACRO IF NOT EXISTS haversine(lat1, lon1, lat2, lon2) AS (
      6378137 * 2 * ASIN(
        SQRT(
          POWER(l(lat2 - lat1) / 2), 2) +
          COS(RADIANS(lat1)) * COS(RADIANS(lat2)) *
          POWER(SIN(RADIANS(lon2 - lon1) / 2), 2)
        )
      )
    );
  "
  )

  # TO OPTMIZE 666666666666666666666 -------------------------------------------
  # 1.1) calc dist mantendo apenas tempidgeocodebr
  # 1.2) calc de dist sucessivo: primeiro 1k, dpeois a distancia q o usuario passa
  # 1.3) dist usando duckspatial
  # 2) depois dar um left join do input_table_db com resultado da dist para retornar input original


  # Find cases nearby -------------------------------------------------------
  query_filter_cases_nearby <- glue::glue(
    "WITH dist_data AS (
        SELECT
              input_table_db.* EXCLUDE (lon_min, lon_max, lat_min, lat_max),
              filtered_cnefe.endereco_completo,
              filtered_cnefe.estado,
              filtered_cnefe.municipio,
              filtered_cnefe.logradouro,
              filtered_cnefe.numero,
              filtered_cnefe.cep,
              filtered_cnefe.localidade,
              haversine(
                    input_table_db.lat, input_table_db.lon,
                    filtered_cnefe.lat, filtered_cnefe.lon
              ) AS distancia_metros
        FROM
              input_table_db, filtered_cnefe
        WHERE
              input_table_db.lat_min < filtered_cnefe.lat
          AND input_table_db.lat_max > filtered_cnefe.lat
          AND input_table_db.lon_min < filtered_cnefe.lon
          AND input_table_db.lon_max > filtered_cnefe.lon
    ),

    ranked AS (
        SELECT
            *,
            RANK() OVER (
                PARTITION BY tempidgeocodebr
                ORDER BY distancia_metros ASC
            ) AS ranking
        FROM dist_data
    )

    SELECT * EXCLUDE(tempidgeocodebr, ranking)
    FROM ranked
    WHERE ranking = 1;"
  )



  output <- DBI::dbGetQuery(con, query_filter_cases_nearby)

  # TODO 6666666
  if (nrow(output)==0){
    stop("Nenhum endereco proximo foi encontrados")
  }

  # organize output -------------------------------------------------

  # convert df to simple feature
  output_sf <- sfheaders::sf_point(
    obj = output,
    x = 'lon',
    y = 'lat',
    keep = TRUE
  )

  sf::st_crs(output_sf) <- 4674

  duckdb::dbDisconnect(con)

  return(output_sf)
}















pontos <- readRDS(
  system.file("extdata/pontos.rds", package = "geocodebr")
)

bench::mark(
  # duck_filter1 = reverse_geocode_filter(coordenadas = pontos, dist_max = 2000, n_cores = 1),
  # duck_filter8 = reverse_geocode_filter(coordenadas = pontos, dist_max = 2000, n_cores = 8),
  # duck_join1 =  reverse_geocode_join(coordenadas = pontos, dist_max = 2000, n_cores = 1),
  # duck_join8 =  reverse_geocode_join(coordenadas = pontos, dist_max = 2000, n_cores = 8),
  hybrid1 = reverse_geocode_hybrid(coordenadas = pontos, dist_max = 2000, n_cores = 1),
  hybrid8 = reverse_geocode_hybrid(coordenadas = pontos, dist_max = 2000, n_cores = 8),
  iterations = 5,
  check = F
)

# 1000 pontos
#     expression      min   median `itr/sec` mem_alloc `gc/sec` n_itr  n_gc total_time result memory
#     <bch:expr>   <bch:>   <bch:>     <dbl> <bch:byt>    <dbl> <int> <dbl>   <bch:tm> <list> <list>
# 1 duck_filter1    3.19m    3.31m   0.00490      75MB  0.0108      5    11        17m <NULL> <Rprofmem>
# 2 duck_filter8    2.93m    3.04m   0.00516      51MB  0.00723     5     7      16.1m <NULL> <Rprofmem>
#
# 1 duck_join1      2.93m    3.54m   0.00475    76.2MB  0.00854     5     9      17.6m <NULL> <Rprofmem>
# 2 duck_join8      3.62m    4.05m   0.00407    51.2MB  0.00651     5     8      20.5m <NULL> <Rprofmem>
#
# 1 hybrid1         5.13m     5.9m   0.00277    88.3MB    0.262     5   473      30.1m <NULL> <Rprofmem>
# 2 hybrid8         5.09m    5.19m   0.00321      66MB    0.307     5   478      25.9m <NULL> <Rprofmem>

# # open input data
# data_path <- system.file("extdata/small_sample.csv", package = "geocodebr")
# input_df <- read.csv(data_path, encoding = 'Latin-1')
#
# campos <- geocodebr::definir_campos(
#   logradouro = 'nm_logradouro',
#   numero = 'Numero',
#   cep = 'Cep',
#   localidade = 'Bairro',
#   municipio = 'nm_municipio',
#   estado = 'nm_uf'
# )
#
# df_geo <- geocodebr::geocode(
#     enderecos = input_df,
#     campos_endereco = campos,
#     n_cores = 7,
#     resultado_completo = T,
#     verboso = T
#   )
#
#
# df_rafaF$lat |> nchar() |> summary()
#
# lapply(X=df_rafaF$lat, FUN=decimalplaces) |>  unlist() |> summary()
#
# a <- subset(df_rafaF, nchar(lat) <= 6 )
#
#
#   filtered_cnefedf$lat |> nchar() |> summary()
#   lapply(X=filtered_cnefedf$lat, FUN=decimalplaces) |>  unlist() |> as.numeric() |> summary()
#
#
#
#
#
library(dplyr)
library(mapview)
library(sfheaders)
library(data.table)
mapviewOptions(platform = 'leafgl')

sp_muni <- geobr::read_municipality(code_muni = 3550308)


filtered_cnefe_sp <- arrow::open_dataset( geocodebr::listar_dados_cache()[11] ) |>
    dplyr::filter(estado == 'SP') |>
    dplyr::filter(municipio == "SAO PAULO") |>
    dplyr::filter(logradouro_sem_numero == "RUA IPE ROXO") |>
    dplyr::collect()

a <- subset(filtered_cnefe_sp, cep == "04896-360")
a <- subset(filtered_cnefe_sp, cep == "04896-360")

table(a$localidade)
#> RUA IPE ROXO
#>          243

a <- sfheaders::sf_point(
  obj = a,
  x = 'lon',
  y = 'lat',
  keep = TRUE
)
sf::st_crs(a) <- 4674

mapview::mapview(a, zcol='localidade') + sp_muni

#   table(a$localidade)
#   table(a$logradouro_sem_numero)
#   table(a$cep)
#
#     a <- subset(filtered_cnefe_sp, localidade %like% "COLONIA")
#

library(geocodebr)
library(mapview)

df <- data.frame(
  estado = 'sp',
  municipio = 'sao paulo',
  logradouro = 'rua ipe roxo',
  numero = 1:5000,
  bairro = NA, #'vargem grande',
  cep = "04896-360"
  )

sp_muni <- geobr::read_municipality(code_muni = 3550308)

campos <- geocodebr::definir_campos(
  logradouro = 'logradouro',
  numero = 'numero',
  cep = 'cep',
  localidade = 'bairro',
  municipio = 'municipio',
  estado = 'estado'
)

df_geo <- geocodebr::geocode(
  enderecos = df,
  campos_endereco = campos,
  resultado_completo = T,
  verboso = T,
  resultado_sf = T
)
# nchar(df_geo$lat) # -23.85401
# df_geo

mapview::mapview(df_geo, zcol='numero') + sp_muni

df_geo$lat



# teste TSE ---------------


df <- data.table::fread("teste_geocodebr_1km.csv")
head(df)
i <- 7
dfi <- df[i,]
dfi$tipo_resultado

bairroi <- df[i, 'NM_BAIRRO']$NM_BAIRRO
logr <- df[i, 'LOGRADOURO']$LOGRADOURO |> enderecobr::padronizar_logradouros()
num <- df[i, 'NUMERO']$NUMERO
cepi <- df[i, 'NR_CEP']$NR_CEP |> enderecobr::padronizar_ceps()
muni <- df[i, 'nm_municipio']$nm_municipio
uf <- df[i, 'nm_uf']$nm_uf

tudo <- geocodebr::listar_dados_cache()[11]
tudo <- geocodebr::listar_dados_cache()[11]


filtered_cnefe_sp <- arrow::open_dataset( tudo ) |>
  dplyr::filter(estado == 'SP') |>
  dplyr::filter(municipio == "SAO PAULO") |>
   dplyr::filter(logradouro_sem_numero == logr) |>
  # dplyr::filter(cep == cepi) |>
  #  dplyr::filter(localidade == bairroi) |>
  # dplyr::filter(lat > -23.574453 & lat < -23.567316 &    # parque sao domingos
  #               lon > -46.719479 & lon < -46.704553) |> # parque sao domingos
  dplyr::collect()


# cnefe_original <- ipeadatalake::ler_cnefe(ano = 2022, as_data_frame = F) |> dplyr::compute()
#
# filtered_cnefe_sp <- cnefe_original |>
#   dplyr::compute() |>
#   dplyr::filter(cep == 08461620) |>
#   dplyr::collect()


# -23.566701, -46.717479       -23.567316, -46.704360
#
# -23.574453, -46.719479      -23.574429, -46.704553

unique(filtered_cnefe_sp$logradouro_sem_numero)
unique(filtered_cnefe_sp$cep)
unique(filtered_cnefe_sp$localidade)
unique(filtered_cnefe_sp$numero)

a <- subset(filtered_cnefe_sp, cep == "04896-360")


campos <- geocodebr::definir_campos(
  logradouro = 'LOGRADOURO',
  numero = 'NUMERO',
  cep = 'NR_CEP',
  localidade = 'NM_BAIRRO',
  municipio = 'nm_municipio',
  estado = 'nm_uf'
)

df_geo <- geocodebr::geocode(
  enderecos = select( df, -c("lat", "lon", "tipo_resultado",'precisao')),
  campos_endereco = campos,
  resultado_completo = T,
  verboso = T,
  resultado_sf = F
)


a <- sfheaders::sf_point(
  obj = filtered_cnefe_sp,
  x = 'lon',
  y = 'lat',
  keep = TRUE
)
sf::st_crs(a) <- 4674

mapview::mapview(a, zcol='localidade') + df_geo



#' notas internas:
#' casos (1) e (5) criar o probabilistico e ver se corrige
#'
#' 2) se tem dois lograd com mesmo cep, pq o left join só retorna um e nao os dois?
#'  porque na hora de gerar os parquets, a gente tira a media sem a coluna de localidade
#'  e dai a media dá um resultado bizarro no meio do caminho.
#'  SE alternativamente a gente fizer pra categoria er02 o merge com o parquet
#'  que inclui o bairro, daí o left join retorna mais de um resultado. E dai
#'  poderiamos indicar caso de empate
#'

#' 1) COLÉGIO EQUIPE
#' ok: google e TSE
#' O logradouro está com uma letra errada no dado de input. O nome correto da é
#' 'Sao Vicente de PaulO', mas o input é de PaulA. TSE e Google maps concordam.
#' geocodebr erra feio porque ainda nao tem match probabilistico por string.
#' Se corrigir o nome da rua no input, o geocodebr acerta. Mas uma proxima etapa
#' é implementar o match probabilistico mesmo.
#'
#' 2) E.E. JOSÉ DE SAN MARTIN
#' ok: google e TSE
#' Caso parecido com o da  'RUA IPE ROXO'. O CNEFE aponta que existem endereços
#' no logradouro "RUA DELTA" e com o mesmo CEP "08071-060" em áreas completamente
#' distantes entre si. Nos bairros "JABAQUARA" e "UNIAO VILA NOVA". Como nessa
#' categoria o geocodebr tira a media das coordenadas de todos enderecos com mesmo
#' lograouro e cep, o resultado sai num lugar no meio do caminho e sem sentido.
#' To pensando uma solucao aqui q recupera a info do bairro, mesmo quando ele nao
#' bate com o input do usuario.
#'
#' 3) EMEF. AMADEU MENDES
#' O dado de input parece correto e o cnefe errado.
#' CNEFE: localidade "SAO DOMINGOS" mas deveria ser "PARQUE SAO DOMINGOS". E o cenfe
#' tem uns pontos soltos mais perto do centro com esse mesmo bairro. estranho
#' CNEFE: logradouro "RUA JOAO FERREIRA" mas deveria ser "RUA TOMAS LOPES FERREIRA"
#'
# dplyr::filter(lat > -23.547470 & lat < -23.476933 &    # parque sao domingos
#               lon > -46.780367 & lon < -46.710332 ) |> # parque sao domingos
#'
#'
#' 4) ESCOLA ESTADUAL ALBERTO TORRES
#' input correto. Erro no cnefe onde a avenida muda de nome de "AVENIDA VITAL BRASIL" (correto)
#' para "AVENIDA DOUTOR VITAL"
#'
#  dplyr::filter(lat > -23.574453 & lat < -23.567316 &    # parque sao domingos
#                  lon > -46.719479 & lon < -46.704553) |> # parque sao domingos
#
#'
#'5) EE  AQUILINO RIBEIRO
#' ok: TSE e geocodebr
#' As coordenadas do Google map estao erradas. Mas as coordenadas do TSE e
#' geocodebr sao bem proximas! E isso acontece apesar de ter um erro no input.
#' No input esta "RUA ONOFRE LEITE MEIRELES" e deveria ser 'MEIRELLES' com dois 'L'
#' geocodebr: deve pegar quando tiver match probabilistico, mas mesmo assim acerta
#' pelo cep
#'
#'
#'
#' 6) EE PROF. MARIO ARMINANTE
#' ok: TSE e geocodebr
#' as coordenadas do TSE e o geocodebr estao bem proximas (170 metros). Parece
#' correto mesmo olhando o google street view. As coordenadas do google maps
#' levam para uma outra instituição chamada 'CEU EMEF MANOEL VIEIRA DE QUEIROZ
#' FILHO'
#'
#'
#' 7) EE AYRTON SENNA DA SILVA
#' caso da RUA IPE ROXO. Dado de input parece q tem alguns erros:
#'  - input tem bairro "COLONIA (ZONA SUL)" q nao condiz com google maps
#'  - input tem cep "04896-260" q é "errado" e que joga na "RUA DAS ARARAS", que
#'  é 4 ruas pro lado da 'R IPE ROXO'. O CEP correto da escola seria '04896-360',
#'  MAS o CNEFE aponta q tem RUA IPE ROXO com esse cep "04896-360" em regioes muito
#'  diferentes de SP.
#'



# rua ipe roxo ---------------

tudo <- geocodebr::listar_dados_cache()[11]
log_e_cep <- geocodebr::listar_dados_cache()[6]
log_e_cep_e_localidade <- geocodebr::listar_dados_cache()[7]


filtered_cnefe_sp <- arrow::open_dataset( tudo  ) |> # tudo
  dplyr::filter(estado == 'SP') |>
  dplyr::filter(municipio == "SAO PAULO") |>
 dplyr::filter(logradouro_sem_numero == "RUA IPE ROXO") |>
 # dplyr::filter(cep == "04896-360") |>   # ESSE EH O CEP PROBLEMATICO
  dplyr::collect()

setDT(filtered_cnefe_sp)[, . (lat=mean(lat), lon=mean(lon)), by= c("logradouro_sem_numero", "cep",'localidade')]


a <- sfheaders::sf_point(
  obj = filtered_cnefe_sp,
  x = 'lon',
  y = 'lat',
  keep = TRUE
)
sf::st_crs(a) <- 4674

mapview::mapview(a, zcol='localidade')



filtered_cnefe_sp <- arrow::open_dataset( log_e_cep ) |>
  dplyr::filter(estado == 'SP') |>
  dplyr::filter(municipio == "SAO PAULO") |>
  dplyr::filter(logradouro_sem_numero == "RUA IPE ROXO") |>
  dplyr::filter(cep == "04896-360") |>
  dplyr::collect()


setDT(filtered_cnefe_sp)[, . (lat=mean(lat), lon=mean(lon)), by= c("logradouro_sem_numero", "cep")]



# rua delta ---------------

tudo <- geocodebr::listar_dados_cache()[11]
log_e_cep <- geocodebr::listar_dados_cache()[6]
log_e_cep <- geocodebr::listar_dados_cache()[7]

filtered_cnefe_sp <- arrow::open_dataset( tudo ) |>
  dplyr::filter(estado == 'SP') |>
  dplyr::filter(municipio == "SAO PAULO") |>
  dplyr::filter(logradouro_sem_numero == "RUA DELTA") |>
  dplyr::filter(cep == "08071-060") |>
  dplyr::collect()


filtered_cnefe_sp <- arrow::open_dataset( log_e_cep ) |>
  dplyr::filter(estado == 'SP') |>
  dplyr::filter(municipio == "SAO PAULO") |>
  dplyr::filter(logradouro_sem_numero == "RUA DELTA") |>
  dplyr::filter(cep == "08071-060") |>
  dplyr::collect()


seDT(filtered_cnefe_sp)

filtered_cnefe_sp[, . (lat=mean(lat), lon=mean(lon)), by= c("logradouro_sem_numero", "cep",'localidade'  )]







# rua nove, mage sp ---------------


filtered_cnefe_sp <- arrow::open_dataset( geocodebr::listar_dados_cache()[11] ) |>
  dplyr::filter(estado == 'RJ') |>
  dplyr::filter(municipio == "MAGE") |>
  dplyr::filter(logradouro_sem_numero == "RUA NOVE") |>
  # dplyr::filter(localidade == "PRAIA DA ESPERANCA") |>
  dplyr::collect()

a <- subset(filtered_cnefe_sp, cep == "04896-360")
a <- subset(filtered_cnefe_sp, cep == "04896-360")

table(a$localidade)
#> RUA IPE ROXO
#>          243

a <- sfheaders::sf_point(
  obj = filtered_cnefe_sp,
  x = 'lon',
  y = 'lat',
  keep = TRUE
)
sf::st_crs(a) <- 4674

mapview::mapview(a, zcol='localidade')







# rua padre vieira em sao goncalo ---------------------------------------------------------------------------
# numero cresce e depois diminui
# mesmo numero em ceps diferentes em potas opostas da rua

cidade <- arrow::open_dataset( tudo ) |>
  dplyr::filter(estado == 'RJ') |>
  dplyr::filter(municipio == "SAO GONCALO") |>
  dplyr::collect()


cidade_case <- cidade |>
  dplyr::filter(logradouro_sem_numero == "AVENIDA PADRE VIEIRA") |>
  # dplyr::filter(cep == "23065-110")
  dplyr::filter(localidade =="JARDIM CATARINA")


unique(cidade_case$localidade)
unique(cidade_case$logradouro_sem_numero)
unique(cidade_case$cep)
'27253-360'

# direto via cnefe


cnef <- ipeadatalake::ler_cnefe(ano = 2022, as_data_frame = F) |>
  filter(code_muni==3304904) |> collect()


cnef_case <- cnef |>
  filter(desc_localidade == 'JARDIM CATARINA') |>
  filter( nom_tipo_seglogr == 'AVENIDA' &
            nom_titulo_seglogr == 'PADRE' &
            nom_seglogr =='VIEIRA' &
            num_adress %in% 1:5)


cidade_case_sf <- sfheaders::sf_point(
  obj = cnef_case,
  x = 'lon',
  y = 'lat',
  keep = TRUE
)
sf::st_crs(cidade_case_sf) <- 4674

mapview::mapview(cidade_case_sf, zcol='num_adress')






Existem 6 endereços de Avenida Brasil, número 7 no RJ. A MESMA AVENIDA!
cada um com cep diferente.


# cep estranho no rio de janeiro -----------------------------------------------

tudo <- geocodebr::listar_dados_cache()
tudo <- tudo[7]

dt <- arrow::open_dataset( tudo ) |>
  dplyr::filter(estado == 'RJ') |>
  dplyr::filter(municipio == "RIO DE JANEIRO") |>
  dplyr::filter(cep %in% c( "22440-035")) |>
  # dplyr::filter(cep %in% c("22620-110", "20521-470")) |>
  # dplyr::filter(localidade %in% c("LEBLON")) |>
  dplyr::collect()


dt_sf <- sfheaders::sf_point(
  obj = dt,
  x = 'lon',
  y = 'lat',
  keep = TRUE
)

sf::st_crs(dt_sf) <- 4674

mapview::mapview(dt_sf, zcol='cep')

calculate_string_dist <- function(con, match_type, unique_logradouros_tbl) {
  # message("calculate_string_dist")

  # match_type = "pl01"

  key_cols <- get_key_cols(match_type)

  # cols that cannot be null
  cols_not_null <- paste(
    glue::glue("input_padrao_db.{key_cols} IS NOT NULL"),
    collapse = ' AND '
  )

  # remove numero and logradouro from key cols to allow for the matching
  key_cols_string_dist <- key_cols[!key_cols %in% c("numero", "logradouro")]

  join_condition_lookup <- paste(
    glue::glue(
      "{unique_logradouros_tbl}.{key_cols_string_dist} = input_padrao_db.{key_cols_string_dist}"
    ),
    collapse = ' AND '
  )

  # min cutoff for string match
  min_cutoff <- get_prob_match_cutoff(match_type)

  #-----------------------------------------------------------------------------

  # query novo
  query_calc_dist <- glue::glue(
    "
    -- STEP 1: pick only rows that do NOT have similarity in temp table
    WITH to_compute AS (
      SELECT
          input_padrao_db.tempidgeocodebr,
          input_padrao_db.logradouro AS logradouro_input,
          {unique_logradouros_tbl}.logradouro AS logradouro_cnefe
      FROM input_padrao_db
      JOIN {unique_logradouros_tbl}
        ON {join_condition_lookup}
      WHERE input_padrao_db.similaridade_logradouro IS NULL
        AND input_padrao_db.log_causa_confusao = FALSE
        AND {cols_not_null}
        ),

    -- STEP 2: calculate similarity only for missing pairs
    computed AS (
      SELECT
          tempidgeocodebr,
          logradouro_cnefe,
          CAST(jaro_similarity(logradouro_input, logradouro_cnefe) AS NUMERIC(5,3)) AS similarity,
          RANK() OVER (PARTITION BY tempidgeocodebr ORDER BY similarity DESC, logradouro_cnefe) AS rank
      FROM to_compute
      WHERE similarity > {min_cutoff}
      )

    -- STEP 3: distances to input db
    UPDATE input_padrao_db
      SET temp_lograd_determ = computed.logradouro_cnefe,
          similaridade_logradouro = similarity
      FROM computed
      WHERE input_padrao_db.tempidgeocodebr = computed.tempidgeocodebr
            AND computed.rank = 1;"
  )

  DBI::dbExecute(con, query_calc_dist)
  #-----------------------------------------------------------------------------

  # # query antigo
  # query_lookup <- glue::glue(
  #   "WITH ranked_data AS (
  #       SELECT
  #         input_padrao_db.tempidgeocodebr,
  #         {unique_logradouros_tbl}.logradouro AS logradouro_cnefe,
  #         CAST(jaro_similarity(input_padrao_db.logradouro, {unique_logradouros_tbl}.logradouro) AS NUMERIC(5,3)) AS similarity,
  #         RANK() OVER (PARTITION BY input_padrao_db.tempidgeocodebr ORDER BY similarity DESC, logradouro_cnefe) AS rank
  #       FROM input_padrao_db
  #       JOIN {unique_logradouros_tbl}
  #         ON {join_condition_lookup}
  #      WHERE {cols_not_null}
  #            AND input_padrao_db.log_causa_confusao is false
  #            AND input_padrao_db.similaridade_logradouro IS NULL
  #            AND similarity > {min_cutoff}
  #     )
  #
  #     UPDATE input_padrao_db
  #        SET temp_lograd_determ = ranked_data.logradouro_cnefe,
  #            similaridade_logradouro = similarity
  #      FROM ranked_data
  #     WHERE input_padrao_db.tempidgeocodebr = ranked_data.tempidgeocodebr
  #           AND ranked_data.similarity > {min_cutoff}
  #           AND ranked_data.rank = 1;"
  #     )
  #
  # DBI::dbExecute(con, query_lookup)

  # a <- DBI::dbReadTable(con, 'input_padrao_db')
  # sum(is.na(a$similaridade_logradouro))
}

# gera bbox de estados e municipios

library(sf)
library(geobr)
library(dplyr)
library(geoarrow)

# Load the state polygons
df <- geobr::read_municipality(year = 2022, simplified = T)

# Calculate bounding boxes of states
bounding_boxes <- df |>
  st_as_sf() |>                           # Ensure df is an sf object
  rowwise() |>                            # Process each polygon individually
  mutate(
    xmin = st_bbox(geometry)["xmin"],      # Extract xmin from the bounding box
    ymin = st_bbox(geometry)["ymin"],      # Extract ymin from the bounding box
    xmax = st_bbox(geometry)["xmax"],      # Extract xmax from the bounding box
    ymax = st_bbox(geometry)["ymax"]       # Extract ymax from the bounding box
  ) |>
  ungroup() |>                            # Unrowwise after rowwise operations
  select(code_muni, xmin, ymin, xmax, ymax) |> # Select desired columns
  st_drop_geometry()

# View the resulting bounding box data.frame
head(bounding_boxes)

data.table::fwrite(bounding_boxes, './inst/extdata/munis_bbox.csv')


head(input_table)

candidate_states <-
  subset(x = bounding_boxes,
         (xmin < bbox_lon_min | xmax > bbox_lon_max) &
           (ymin < bbox_lat_min | ymax > bbox_lat_max)
  )









# Calculate bounding boxes of municipalities
bounding_boxes <- df |>
  rowwise() |>                            # Process each polygon individually
  mutate(
    geometry = st_as_sfc(st_bbox(geometry))  # Create bbox geometry
  ) |>
  ungroup() |>                            # Unrowwise after rowwise operations
  select(code_muni, geometry) |>
  st_set_geometry("geometry")            # Set bbox_geom as the active geometry

# View the resulting bounding box sf object
head(bounding_boxes)

# mapview::mapview(bounding_boxes) + df
bounding_boxes <- bounding_boxes |>
  mutate(code_muni = as.integer(code_muni))
  # select(code_muni, geometry = bbox_geom, - geom) |>
  # st_set_geometry("geometry")            # Set bbox_geom as the active geometry

head(bounding_boxes)

# arrow::write_parquet(bounding_boxes2, "munis_bbox_2022.parquet")
arrow::write_parquet(bounding_boxes, "munis_bbox_2022.parquet",
                     compression='zstd',
                     compression_level = 7)


path <- "//storage6/usuarios/Proj_acess_oport/git_rafa/prep_roger/data/municipality/2022/municipalities_2022_simplified.parquet"
df <- arrow::open_dataset(path) |>
  dplyr::select(code_muni, geometry) |>
  sf::st_as_sf()

arrow::write_parquet(df, "munis_2022.parquet",
                     compression='zstd',
                     compression_level = 7)

calculate_string_dist <- function(con, match_type, unique_logradouros_tbl){

  # match_type = "pn01"

  key_cols <- get_key_cols(match_type)

  # cols that cannot be null
  cols_not_null <-  paste(
    glue::glue("input_padrao_db.{key_cols} IS NOT NULL"),
    collapse = ' AND '
  )

  # remove numero and logradouro from key cols to allow for the matching
  key_cols_string_dist <- key_cols[!key_cols %in%  c("numero", "logradouro")]

  # # name of cache table
  # cache_tbl <- paste0("similarity_cache_", paste0(key_cols_string_dist, collapse = "_"))

  join_condition_lookup <- paste(
    glue::glue("{unique_logradouros_tbl}.{key_cols_string_dist} = input_padrao_db.{key_cols_string_dist}"),
    collapse = ' AND '
  )

  # min cutoff for string match
  min_cutoff <- get_prob_match_cutoff(match_type)


  # query create NEW cache for logradouro distances
  query_create_dist_cache <- glue::glue(
    "
    -- STEP 1: pick only rows that do NOT have similarity in temp table
   CREATE OR REPLACE TEMP TABLE similarity_cache AS
   -- CREATE TEMP TABLE IF NOT EXISTS similarity_cache AS
      SELECT
          input_padrao_db.tempidgeocodebr,
          input_padrao_db.logradouro AS logradouro_input,
          {unique_logradouros_tbl}.logradouro AS logradouro_cnefe,
          NULL AS logr_similarity
      FROM input_padrao_db
      JOIN {unique_logradouros_tbl}
        ON {join_condition_lookup}
      WHERE {cols_not_null}
        AND input_padrao_db.log_causa_confusao = FALSE
        AND input_padrao_db.similaridade_logradouro IS NULL;"
    )

  query_calc_dist_cache <- glue::glue(
      "UPDATE similarity_cache
        SET logr_similarity =
         CAST(jaro_similarity(logradouro_input, logradouro_cnefe) AS NUMERIC(5,3))
      WHERE logr_similarity IS NULL;"
  )

  DBI::dbExecute(con, query_create_dist_cache)
  similarity_cache0 <- DBI::dbReadTable(con, 'similarity_cache')

  DBI::dbExecute(con, query_calc_dist_cache)
  similarity_cache0 <- DBI::dbReadTable(con, 'similarity_cache')

  # check if there are  create NEW cache for logradouro distances
  query_add_rows_to_dist_cache <- glue::glue(
    "
    -- STEP 2: check if there are more distance pairs to calculate and add them to cache
   WITH to_compute AS (
      SELECT
          input_padrao_db.tempidgeocodebr,
          input_padrao_db.logradouro AS logradouro_input,
          {unique_logradouros_tbl}.logradouro AS logradouro_cnefe,
          input_padrao_db.similaridade_logradouro AS logr_similarity,
      FROM input_padrao_db
      JOIN {unique_logradouros_tbl}
        ON {join_condition_lookup}
      WHERE input_padrao_db.log_causa_confusao = FALSE
        AND input_padrao_db.similaridade_logradouro IS NULL
        AND input_padrao_db.tempidgeocodebr NOT IN (SELECT DISTINCT(tempidgeocodebr) FROM similarity_cache)
        )

  INSERT INTO similarity_cache (tempidgeocodebr, logradouro_input, logradouro_cnefe, logr_similarity)
      SELECT tempidgeocodebr, logradouro_input, logradouro_cnefe, logr_similarity
      FROM to_compute;"
  )

  DBI::dbExecute(con, query_add_rows_to_dist_cache)
  similarity_cache1 <- DBI::dbReadTable(con, 'similarity_cache')

   # compute similarity and update the cache
    query_calc_dist <- glue::glue(
    "
    -- STEP 2: calculate similarity only for missing pairs
    WITH computed AS (
      SELECT
          tempidgeocodebr,
          logradouro_input,
          logradouro_cnefe,
          CAST(jaro_similarity(logradouro_input, logradouro_cnefe) AS NUMERIC(5,3)) AS similarity,
          RANK() OVER (PARTITION BY tempidgeocodebr ORDER BY similarity DESC, logradouro_cnefe) AS rank
      FROM similarity_cache
      WHERE logr_similarity IS NULL
      )

    -- STEP 3: update cache
    UPDATE similarity_cache
      SET logr_similarity = computed.similarity
      FROM computed
      WHERE similarity_cache.tempidgeocodebr = computed.tempidgeocodebr
            AND similarity_cache.logradouro_cnefe = computed.logradouro_cnefe
            AND computed.rank = 1
          --  AND computed.similarity > {min_cutoff};"
    )

    query_dist_to_inputdb <- glue::glue(
    "
    -- STEP 4: pass distances to input db
    UPDATE input_padrao_db
      SET temp_lograd_determ =similarity_cache.logradouro_cnefe,
          similaridade_logradouro = logr_similarity
      FROM similarity_cache
      WHERE input_padrao_db.tempidgeocodebr = similarity_cache.tempidgeocodebr
            -- AND similarity_cache.rank = 1
            AND similarity_cache.logr_similarity > {min_cutoff}

     ;"
  )


  DBI::dbExecute(con, query_create_dist_cache)
  similarity_cache0 <- DBI::dbReadTable(con, 'similarity_cache')

  DBI::dbExecute(con, query_calc_dist)
  similarity_cache2 <- DBI::dbReadTable(con, 'similarity_cache')

  DBI::dbExecute(con, query_dist_to_inputdb)
  input_padrao_db1 <- DBI::dbReadTable(con, 'input_padrao_db')
  # sum(is.na(a$similaridade_logradouro))
}

# solucao chat ----------------------------------------

  # match_type = "pn02"

  key_cols <- get_key_cols(match_type)

  # cols that cannot be null
  cols_not_null <-  paste(
    glue::glue("input_padrao_db.{key_cols} IS NOT NULL"),
    collapse = ' AND '
  )

  # remove numero and logradouro from key cols to allow for the matching
  key_cols_string_dist <- key_cols[!key_cols %in%  c("numero", "logradouro")]

  # # name of cache table
  # cache_tbl <- paste0("similarity_cache_", paste0(key_cols_string_dist, collapse = "_"))

  join_condition_lookup <- paste(
    glue::glue("{unique_logradouros_tbl}.{key_cols_string_dist} = input_padrao_db.{key_cols_string_dist}"),
    collapse = ' AND '
  )

  # min cutoff for string match
  min_cutoff <- get_prob_match_cutoff(match_type)


  # STEP 0 — TEMP similarity cache (created once per geocode() call)
  DBI::dbExecute(
    con,
    "
     CREATE TEMP TABLE IF NOT EXISTS similarity_cache (
      tempidgeocodebr     INTEGER,
      logradouro_input    TEXT,
      logradouro_cnefe    TEXT,
      logr_similarity     DOUBLE,
      PRIMARY KEY (tempidgeocodebr, logradouro_cnefe)
      );"
    )

  similarity_cache1 <-  DBI::dbReadTable(con, 'similarity_cache')

  # 6666666666666666666666666 o erro esta aqui na hora de roda pn02
  # STEP 1 — Insert NEW pairs that still need similarity
  query1_update_cache <- glue::glue(
    "
    WITH to_compute AS (
      SELECT
          input_padrao_db.tempidgeocodebr,
          input_padrao_db.logradouro AS logradouro_input,
          {unique_logradouros_tbl}.logradouro AS logradouro_cnefe,
          NULL AS logr_similarity
      FROM input_padrao_db
      JOIN {unique_logradouros_tbl}
        ON {join_condition_lookup}
      WHERE {cols_not_null}
        AND input_padrao_db.log_causa_confusao = FALSE
        AND input_padrao_db.similaridade_logradouro IS NULL
        AND input_padrao_db.tempidgeocodebr NOT IN (
              SELECT tempidgeocodebr FROM similarity_cache)
    )

      MERGE INTO similarity_cache AS sc
      USING to_compute AS tc
      ON sc.tempidgeocodebr = tc.tempidgeocodebr
         AND sc.logradouro_cnefe = tc.logradouro_cnefe
      WHEN NOT MATCHED THEN
          INSERT (tempidgeocodebr, logradouro_input, logradouro_cnefe, logr_similarity)
          VALUES (tc.tempidgeocodebr, tc.logradouro_input, tc.logradouro_cnefe, NULL);
    "
      )

  DBI::dbExecute(con, query1_update_cache)

  similarity_cache2 <-  DBI::dbReadTable(con, 'similarity_cache')


  # STEP 2 — Compute similarity only for rows where it's still NULL

  query2_calc_dist <- glue::glue(
    "
    WITH computed AS (
    SELECT
        tempidgeocodebr,
        logradouro_cnefe,
        CAST(jaro_similarity(logradouro_input, logradouro_cnefe) AS NUMERIC(5,3)) AS similarity,
    FROM similarity_cache
    WHERE logr_similarity IS NULL
    )

    UPDATE similarity_cache AS sc
      SET logr_similarity = c.similarity
      FROM computed AS c
      WHERE sc.tempidgeocodebr = c.tempidgeocodebr
            AND sc.logradouro_cnefe = c.logradouro_cnefe;
    ")

  DBI::dbExecute(con, query2_calc_dist)

  similarity_cache1 <-  DBI::dbReadTable(con, 'similarity_cache')

  # step 3 - keep best per input
  query3_dedup_cache <- glue::glue(
    "
    WITH ranked AS (
    SELECT
        tempidgeocodebr,
        logradouro_cnefe,
        ROW_NUMBER() OVER (PARTITION BY tempidgeocodebr ORDER BY logr_similarity DESC, logradouro_cnefe) AS rank
    FROM similarity_cache
    )

    DELETE FROM similarity_cache
    WHERE (tempidgeocodebr, logradouro_cnefe) NOT IN (
        SELECT tempidgeocodebr, logradouro_cnefe
        FROM ranked
        WHERE rank = 1);
    ")

  DBI::dbExecute(con, query3_dedup_cache)

  similarity_cache2 <-  DBI::dbReadTable(con, 'similarity_cache')
  nrow(similarity_cache2)


  # STEP 4 - Rank similarity matches & pick best candidate per input
  query4_dist_to_inputdb <- glue::glue(
  "
  UPDATE input_padrao_db AS x
    SET
        temp_lograd_determ = sc.logradouro_cnefe,
        similaridade_logradouro = sc.logr_similarity
    FROM similarity_cache AS sc
    WHERE sc.logr_similarity > {min_cutoff}
      AND x.tempidgeocodebr = sc.tempidgeocodebr
      AND x.similaridade_logradouro is NULL;
    ")

  DBI::dbExecute(con, query4_dist_to_inputdb)

  input_padrao_db1 <-  DBI::dbReadTable(con, 'input_padrao_db')

#' Package: geocodebr: Geolocalização De Endereços Brasileiros (Geocoding
#' Brazilian Addresses)
#'
#' @name geocodebr
#' @aliases geocodebr-package
#'
#' @importFrom dplyr mutate select across case_when all_of
#' @importFrom data.table := .I .SD %chin% .GRP .N %like%
#' @importFrom stats weighted.mean na.omit
#'
#' @keywords internal
"_PACKAGE"

utils::globalVariables(
  c(
    "year",
    "temp_local_file",
    "lon_min",
    "lon_max",
    "lat_min",
    "lat_max",
    ".",
    "lon_diff",
    "lat_diff",
    "lon",
    "lat",
    "estado",
    "municipio",
    "tempidgeocodebr",
    "input_padrao",
    "dist_geocodebr",
    "empate",
    "contagem_cnefe",
    "temp_lograd_determ",
    "similaridade_logradouro",

    # due to reverse geocoding draft
    "cep",
    "code_muni",
    "geometry",
    "lat_cnefe",
    "localidade",
    "lon_cnefe",
    "distancia_metros",
    "logradouro",
    "numero",
    "xmax",
    "xmin",
    "ymax",
    "ymin",

    # due to empates
    "endereco_encontrado",
    "logradouro_encontrado"
  )
)

create_geocodebr_db <- function(
  # nocov start
  db_path = "tempdir",
  n_cores = parent.frame()$n_cores,
  load_spatial = FALSE
) {
  # check input
  checkmate::assert_number(n_cores, lower = 1, finite = TRUE, null.ok = TRUE)

  # checkmate::assert_string(db_path, pattern = "tempdir|memory")

  # this creates a local database which allows DuckDB to
  # perform **larger-than-memory** workloads
  if (db_path == 'tempdir') {
    db_path <- tempfile(pattern = 'geocodebr', fileext = '.duckdb')
  }

  if (db_path == 'memory') {
    con <- duckdb::dbConnect(duckdb::duckdb(), dbdir = ":memory:")
  }

  # create db connection
  con <- duckdb::dbConnect(
    duckdb::duckdb(bigint = "integer64"),
    dbdir = db_path
  ) # db_path ":memory:"

  # Set Number of cores for parallel operation
  if (is.null(n_cores)) {
    n_cores <- min(
      parallelly::availableCores(),
      parallelly::freeConnections()
    )
  }

  # Set threads
  DBI::dbExecute(con, sprintf("SET threads = %s;", n_cores))

  # Silence progress bar from duckdb
  DBI::dbExecute(con, "SET enable_progress_bar = false")

  # load spatial extension
  if (isTRUE(load_spatial)) {
    duckspatial::ddbs_install(conn = con, upgrade = FALSE, quiet = TRUE)
    duckspatial::ddbs_load(conn = con, quiet = TRUE)
  }


  # load parquet extension
  # DBI::dbExecute(con, "INSTALL parquet; LOAD parquet;")


  # Set Memory limit
  # DBI::dbExecute(con, "SET memory_limit = '8GB'")

  # DBI::dbExecute(con, "INSTALL arrow FROM community; LOAD arrow;")
  # DBI::dbExecute(con, "LOAD arrow;")

  return(con)
} # nocov end

# gera bbox de estados e municipios

library(sf)
library(dplyr)

# Load the state polygons
df <- geobr::read_state()

# Calculate bounding boxes of states
bounding_boxes <- df |>
  st_as_sf() |>                           # Ensure df is an sf object
  rowwise() |>                            # Process each polygon individually
  mutate(
    xmin = st_bbox(geom)["xmin"],      # Extract xmin from the bounding box
    ymin = st_bbox(geom)["ymin"],      # Extract ymin from the bounding box
    xmax = st_bbox(geom)["xmax"],      # Extract xmax from the bounding box
    ymax = st_bbox(geom)["ymax"]       # Extract ymax from the bounding box
  ) |>
  ungroup() |>                            # Unrowwise after rowwise operations
  select(abbrev_state, xmin, ymin, xmax, ymax) |> # Select desired columns
  st_drop_geometry()

# View the resulting bounding box data.frame
bounding_boxes

data.table::fwrite(bounding_boxes, './inst/extdata/states_bbox.csv')


head(input_table)

candidate_states <-
  subset(x = bounding_boxes,
         (xmin < bbox_lon_min | xmax > bbox_lon_max) &
         (ymin < bbox_lat_min | ymax > bbox_lat_max)
         )





# gerate states bbox -----------------------------------------------------------

states <- geobr::read_state(simplified = FALSE)

bbox_states2 <- states |>
  group_by(abbrev_state) |>
  mutate(xmin = sf::st_bbox(geom)[1] |> round(7),
         ymin = sf::st_bbox(geom)[2] |> round(7),
         xmax = sf::st_bbox(geom)[3] |> round(7),
         ymax = sf::st_bbox(geom)[4] |> round(7)
  ) |>
  sf::st_drop_geometry() |>
  select(abbrev_state, xmin, ymin, xmax, ymax)

data.table::setDT(bbox_states2)
bbox_states2$xmin |> nchar()

saveRDS(bbox_states2, './inst/extdata/states_bbox.rds')

---
title: "Introdução ao geocodebr"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
code-annotations: hover
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Introdução ao geocodebr}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---
  
```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)
```

Geolocalização refere-se ao ato de encontrar um ponto no espaço, geralmente
representado por um par de coordenadas, a partir de um determinado endereço. O
**geocodebr** permite geolocalizar endereços brasileiros de forma simples e
eficiente e sem limite de número de consultas, a partir de dados públicos de 
endereços do Brasil. A principal base de referência é o Cadastro Nacional de 
Endereços para Fins Estatísticos (CNEFE), um conjunto de dados coletado e
[publicado](https://www.ibge.gov.br/estatisticas/sociais/populacao/38734-cadastro-nacional-de-enderecos-para-fins-estatisticos.html) pelo Instituto Brasileiro de Geografia e Estatística 
(IBGE) que contém os endereços de mais de 110 milhões de domicílios e 
estabelecimentos do país.


## Instalação

A versão estável do pacote pode ser baixada do CRAN com o comando a seguir:
  
```{r, eval = FALSE}
install.packages("geocodebr")
```

Caso prefira, a versão em desenvolvimento:
  
```{r, eval = FALSE}
# install.packages("remotes")
remotes::install_github("ipea/geocodebr")
```



## Utilização

O **{geocodebr}** possui três funções principais para geolocalização de dados:
  
1. `geocode()`
2. `geocode_reverso()`
3. `busca_por_cep()`


### 1. Geolocalização: de endereços para coordenadas espaciais

Uma vez que você possui uma tabela de dados (`data.frame`) com endereços no Brasil, a geolocalização desses dados pode ser feita em apenas dois passos:
  
1. O primeiro passo é usar a função `definir_campos()` para indicar os nomes das colunas no seu `data.frame` que correspondem a cada campo dos endereços.
2. O segundo passo é usar a função `geocode()` para encontrar as coordenadas geográficas dos endereços de input.

```{r}
library(geocodebr)
library(sf)

# carregando uma amostra de dados
input_df <- read.csv(system.file("extdata/small_sample.csv", package = "geocodebr"))

# Primeiro passo: inidicar o nome das colunas com cada campo dos enderecos
campos <- geocodebr::definir_campos(
  logradouro = "nm_logradouro",
  numero = "Numero",
  cep = "Cep",
  localidade = "Bairro",
  municipio = "nm_municipio",
  estado = "nm_uf"
)

# Segundo passo: geolocalizar
df <- geocodebr::geocode(
  enderecos = input_df,
  campos_endereco = campos,
  resultado_completo = FALSE,
  resolver_empates = FALSE,
  h3_res = NULL,
  resultado_sf = FALSE,
  verboso = FALSE,
  cache = TRUE,
  n_cores = 1
)

```

**Nota:** A função `geocode()` requer que os dados do CNEFE estejam armazenados
localmente. A primeita vez que a função é executada, ela baixa os dados do CNEFE 
e salva em um cache local na sua máquina. No total, esses dados somam cerca de 
1.4 GB, o que pode fazer com que a primeira execução da função demore. Esses dados, 
no entanto, são salvos de forma  persistente, logo eles são baixados uma única vez.
Ver abaixo mais informações sobre o cache de dados.

Os resultados do **{geocodebr}** são classificados em seis categorias gerais de `precisao`, dependendo do nível de exatidão com que cada endereço de input foi encontrado nos dados do CNEFE. s resultados trazem ainda uma estimativa da 
incerteza da localização encontrado como um `desvio_metros`. Para mais 
informações, consulte a documentação da função ou a [**vignette "geocode"**](https://ipea.github.io/geocodebr/articles/geocode.html).


### 2. Geolocalização reversa: de coordenadas espaciais para endereços

A função `geocode_reverso()`, por sua vez, permite a geolocalização reversa, ou seja, a busca de endereços próximos a um conjunto de coordenadas geográficas. Mais detalhes na [**vignette "geocode"**](https://ipea.github.io/geocodebr/articles/geocode_reverso.html).

```{r}
# amostra de pontos espaciais
pontos <- readRDS(
  system.file("extdata/pontos.rds", package = "geocodebr")
)

# seleciona somente os primeiros 20 pontos
pontos <- pontos[1:20,]

# geocode reverso
df_enderecos <- geocodebr::geocode_reverso(
  pontos = pontos,
  dist_max = 1000,
  verboso = FALSE,
  n_cores = 1
)

```


### 3. Busca por CEPs

Por fim, a função `busca_por_cep()` permite fazer consultas de CEPs para encontrar endereços associados a cada CEP. A função recebe um vetor de CEPs e retorna um `data.frame` com os endereços e as coordenadas geográficas de cada CEP. O parâmetro
`h3_res` pode ser utilizado para incluir no output o id das células H3 na resolução
espacial desejada.

```{r}
# amostra de CEPs
ceps <- c("70390-025", "20071-001")

df_ceps <- geocodebr::busca_por_cep(
  cep = ceps,
  h3_res = 9,
  resultado_sf = FALSE,
  verboso = FALSE
)

```


## Cache de dados

Como comentado anteriormente, os dados do CNEFE são baixados na primeira vez que
a `geocode()` é executada. Esses dados ficam salvos no *cache* do pacote e não
precisam ser baixados novamente. O pacote inclui algumas funções que ajudam a
gerenciar o *cache*:
  
- `listar_pasta_cache()` - retorna o endereço do *cache* na sua máquina, onde os
dados do CNEFE estão salvos;
- `definir_pasta_cache()` - define uma pasta personalizada para ser usada como
*cache*. Essa configuração é persistente entre diferentes sessões do R;
- `listar_dados_cache()` - lista todos os arquivos armazenados no *cache*;
- `deletar_pasta_cache()` - exclui a pasta de *cache*, bem como todos os
arquivos que estavam armazenados dentro dela.

Após rodar o código desta *vignette*, é provável que o seu *cache* esteja
configurado como a seguir:
  
```{r}
geocodebr::listar_pasta_cache()

geocodebr::listar_dados_cache()
```

---
title: "Geocode reverso"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
code-annotations: hover
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Geocode reverso}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)
```

# Geolocalização reversa: de coordenadas espaciais para endereços

A função `geocode_reverso()` permite fazer geolocalização reversa, isto é,  a partir de um conjunto de coordenadas geográficas, encontrar os endereços correspondentes ou próximos. Essa funcionalidade pode ser útil, por exemplo, para identificar endereços próximos a pontos de interesse, como escolas, hospitais, ou locais de acidentes.

A função recebe como *input* um objeto espacial `sf` com geometria do tipo `POINT`. O resultado é um *data frame* com o endereço encontrado mais próximo de cada ponto de *input*, onde a coluna `"distancia_metros"` indica a distância entre coordenadas originais e os endereços encontrados.


```{r}
library(geocodebr)
library(sf)

# amostra de pontos espaciais
pontos <- readRDS(
  system.file("extdata/pontos.rds", package = "geocodebr")
)

pontos <- pontos[1:20,]

# geocode reverso
df_enderecos <- geocodebr::geocode_reverso(
  pontos = pontos,
  dist_max = 1000,
  verboso = FALSE,
  n_cores = 1
)

head(df_enderecos)
```

Por padrão, a função busca pelo endereço mais próximo num raio aproximado de 1000 metros. No entanto,  o usuário pode ajustar esse valor usando o parâmetro `dist_max` para definir a distância máxima (em metros) de busca. Se um ponto de *input* não tiver nenhum endereço próximo dentro do raio de busca, o ponto não é incluído no *output*.

**Nota:** A função `geocode_reverso()` requer que os dados do CNEFE estejam armazenados
localmente. A primeita vez que a função é executada, ela baixa os dados do CNEFE 
e salva em um cache local na sua máquina. No total, esses dados somam cerca de 
3 GB, o que pode fazer com que a primeira execução da função demore. Esses dados, 
no entanto, são salvos de forma  persistente, logo eles são baixados uma única vez. Mais informações sobre o cache de dados [aqui](https://ipea.github.io/geocodebr/articles/geocodebr.html#cache-de-dados).

---
title: "Geocode"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
code-annotations: hover
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Geocode}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = identical(tolower(Sys.getenv("NOT_CRAN")), "true"),
  out.width = "100%"
)

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)
```

## Geolocalização: de endereços para coordenadas espaciais

A principal função do pacote {geocodebr} é a `geocode()`, que recebe uma tabela 
(`data.frame`) de endereços como entrada e retorna a mesma tabela geolocalizada
como saída. Para demonstrar essa função, utilizamos no exemplo abaixo pequeno 
conjunto de dados que contém endereços com problemas comuns, como informações 
ausentes e campos digitados incorretamente. 

A geolocalização desses dados com **{geocodebr}** pode ser feita em apenas dois 
passos:

1. O primeiro passo é usar a função `definir_campos()` para indicar os nomes das 
colunas no seu `data.frame` que correspondem a cada campo dos endereços. No 
exemplo abaixo, nós indicamos que coluna que contém a informação de logradouro se 
chama `"nm_logradouro"`, que a coluna de número se chama `"Numero"`, etc. 

obs. Note que as colunas indicando o `"estado"` e `"município"` são obrigatórias.

```{r}
library(geocodebr)

# leitura de amostra de dados
ends <- read.csv(system.file("extdata/small_sample.csv", package = "geocodebr"))

# definição dos campos de endereço
campos <- definir_campos(
  estado = "nm_uf",
  municipio = "nm_municipio",
  logradouro = "nm_logradouro",
  numero = "Numero",
  cep = "Cep",
  localidade = "Bairro"
)
```


2. O segundo passo é usar a função `geocode()` para encontrar as coordenadas 
geográficas dos dados de input.

**Nota:** A função `geocode()` requer que os dados do CNEFE estejam armazenados
localmente. A primeita vez que a função é executada, ela baixa os dados do CNEFE 
e salva em um cache local na sua máquina. No total, esses dados somam cerca de 
1.2 GB, o que pode fazer com que a primeira execução da função demore. Esses dados, 
no entanto, são salvos de forma  persistente, logo eles são baixados uma única vez.
Mais informações sobre o cache de dados [aqui](https://ipea.github.io/geocodebr/articles/geocodebr.html#cache-de-dados).

```{r}
# geolicalização
ends_geo <- geocode(
  enderecos = ends, 
  campos_endereco = campos, 
  resultado_completo = FALSE,
  resolver_empates = TRUE,
  h3_res = 9,
  resultado_sf = FALSE,
  verboso = FALSE
  )

head(ends_geo)
```


Por padrão, a tabela de *output* é igual à tabela de input do usuário acrescida de colunas com a latitude e longitude encontradas, bem como de colunas indicando o nível de precisão dos resultados e o endereço encontrado. Quando `resultado_completo = TRUE`, o output é acrescido de algumas colunas extras discriminando separadamente cada componente do endereço que teria sido encontrado.

Cabe também destacar aqui outros três argumentos da função `geocode()`:

- `h3_res` : que permite o usuário inserir uma coluna no output indicando o id 
da célula H3 na resolução espacial desejada. Detalhes sobre as resoluções 
disponíveis em \code{\link{https://h3geo.org/docs/core-library/restable/};
- `resolver_empates`: serve para indicar se o usuário quer que a função resolva automaticamente casos de empate, i.e. casos que o endereço de input do usuário 
pode se referir a diferentes localidades na cidade (e.g. logradouros diferentes 
com mesmo nome mas em bairros distintos). Quando `TRUE`, a função resolve os 
empates selecioando os endereços com maior número de visitas do CNEFE. Quando 
`FALSE`, a função retorna todos os resultados indicando os casos empatados na 
coluna 'empate' para que o usuário possa inspecionar cada caso coluna 'endereco_encontrado'.
- `resultado_sf`: quando `TRUE`, o output é retornado como um objeto espacial de classe `sf` simple feature.

As coordendas espaciais do resultado usam o sistema de referência SIRGAS2000 
(EPSG 4674.), padrão adotado pelo IBGE em todo o Brasil. 



## Processo de matching de endereços

As coordenadas incluídas no resultado da `geocode()` são calculadas a partir da
média das coordenadas dos endereços do CNEFE que correspondem a cada um dos
endereços de *input*. Esse cálculo é feito em duas etapas, e retorna o ponto 
médio dos 95% pontos mais próximos entre si, evitando eventual distorção por 
devido a poucos pontos muito isolados. 

A correspondência entre os endereços de entrada e os do CNEFE pode ser feita com 
base em diferentes combinações de campos, impactando, assim, na precisão do 
resultado retornado. No caso mais rigoroso, a função encontra uma correspondência determinística para todos os campos do endereço (estado, município, logradouro, número, CEP e localidade). Pense, por exemplo, em um prédio com vários apartamentos, cuja única variação no endereço se dá a nível de apartamento: o resultado, nesse
caso, é a média das coordenadas dos apartamentos, que podem diferir ligeiramente.

Em um caso menos rigoroso, no qual são encontradas correspondências apenas para
os campos de estado, município, logradouro e localidade, a função calcula as
coordenadas médias de todos os endereços do CNEFE que se encontram na mesma rua
e na mesma localidade. O resultado, portanto, é agregado a nível de rua,
tendendo para a extremidade do logradouro com maior concentração de endereços.


## Grau de precisão dos resultados

A precisão dos resultados do **{geocodebr}** são apresentadas em 3 colunas, `precisao`, `tipo_resultado` e `desvio_metros`, explicadas abaixo.


### Precisão

A coluna `precisao` se refere ao nível de agregação das coordenadas do CNEFE
utilizadas no processo de geolicalização. A função `geocode()` sempre retorna o resultado de maior precisão possível - ou seja, ela só vai procurar endereços com precisão `"numero_aproximado"` (ver a seguir) caso não tenha encontrado correspondência de precisão `"numero"`. As coordenadas calculadas podem ser classificadas em seis diferentes categorias de precisão:

1. `"numero"` - calculadas a partir de endereços que compartilham o mesmo
logradouro e número;
2. `"numero_aproximado"` - calculadas a partir de endereços que compartilham o
mesmo logradouro, mas número de *input* não encontra correspondência exata no
CNEFE e sua localização é calculada a partir de uma interpolação espacial;
3. `"logradouro"` - calculadas a partir de endereços que compartilham o mesmo
logradouro (número de *input* está ausente ou é S/N);
4. `"cep"` - calculadas a partir de endereços que compartilham o mesmo CEP;
5. `"localidade"` - calculadas a partir de endereços que compartilham a mesma
localidade;
6. `"municipio"` - calculadas a partir de endereços que compartilham o mesmo
município.

### Tipos de resultados

A coluna `tipo_resultado` fornece informações mais detalhadas sobre os campos de
endereço utilizados no cálculo das coordenadas de cada endereço de entrada. Cada
categoria é nomeada a partir de um código de quatro caracteres:

- o primeiro, sempre `d` ou `p`, determina se a correspondência foi feita de
forma determinística (`d`) ou probabilística (`p`);
- o segundo faz menção à categoria de `precisao` na qual o resultado foi
classificado (`n` para `"numero"`, `a` para `"numero_aproximado"`, `l` para
`"logradouro"`, `c` para `"cep"`, `b` para `"localidade"` e `m` para
`"municipio"`);
- o terceiro e o quarto designam a classificação de cada categoria dentro de seu
grupo - via de regra, quanto menor o número formado por esses caracteres, mais
precisa são as coordenadas calculadas.

As categorias de `tipo_resultado` são listadas abaixo, junto às categorias de
`precisao` a qual elas estão associadas:

- precisao `"numero"`
  - `dn01` - logradouro, numero, cep e localidade
  - `dn02` - logradouro, numero e cep
  - `dn03` - logradouro, numero e localidade
  - `dn04` - logradouro e numero
  - `pn01` - logradouro, numero, cep e localidade
  - `pn02` - logradouro, numero e cep
  - `pn03` - logradouro, numero e localidade
  - `pn04` - logradouro e numero

- precisao `"numero_aproximado"`
  - `da01` - logradouro, numero, cep e localidade
  - `da02` - logradouro, numero e cep
  - `da03` - logradouro, numero e localidade
  - `da04` - logradouro e numero
  - `pa01` - logradouro, numero, cep e localidade
  - `pa02` - logradouro, numero e cep
  - `pa03` - logradouro, numero e localidade
  - `pa04` - logradouro e numero

- precisao `"logradouro"` (quando o número de entrada está faltando 'S/N')
  - `dl01` - logradouro, cep e localidade
  - `dl02` - logradouro e cep
  - `dl03` - logradouro e localidade
  - `dl04` - logradouro
  - `pl01` - logradouro, cep e localidade
  - `pl02` - logradouro e cep
  - `pl03` - logradouro e localidade
  - `pl04` - logradouro

- precisao `"cep"`
  - `dc01` - municipio, cep, localidade
  - `dc02` - municipio, cep

- precisao `"localidade"`
  - `db01` - municipio, localidade

- precisao `"municipio"`
  - `dm01` - municipio

Endereços não encontrados são retornados com latitude, longitude, precisão e tipo de resultado `NA`.

### Desvio em metros

A coluna `desvio_metros` apresenta uma forma intuitiva e prática de saber o grau 
de incerteza do resultado encontrado. Essa coluna informa que pelo menos 95% de 
todos os pontos do CNEFE que possuem corrêspondência com o endereço de input 
estão num raio de x metros da localização encontrada. 

Um desvio de `30` metros, por exemplo, tende a representar um resultado muito confiável. A depender de como o dado geolocalizado será utilizado, até mesmos resultados com um `desvio_metros` de até 500 ou 900 metros podem ser ser aceitáveis.

A coluna `desvio_metros` pode ser particularmente útil para decidir por exemplo 
se um resultado encontrado com a `precisao` de CEP deveria ser aceitável. Muitas cidades do Brasil possuem um CEP único, o que tende a gerar resultados com 
altíssimo grau de incerteza. Em várias cidades, no entanto, um CEP pode ser circunscrito a uma área muito pequena e as vezes até um único edifício. Nesses 
casos, o valor do `desvio_metros` tende a ser bem pequeno.


## Código do setor censitário

- Quando o usuário passa o argumento `resultado_completo = TRUE`, a função `geocode()` também retorna a coluna `cod_setor` com o código do setor censitário do endereço encontrado. Atualmente, a função somente retorna o código do setor dos casos em 
que todos os pontos do CNEFE correspondentes estão 100% dentro de um único setor 
censitário. Quando os dados do CNEFE correspondentes ao endereço buscado estão 
em mais de um setor, o resultado da coluna `cod_setor` é `NA`.

# errors if could not download one or more files

    Code
      tester(cache = FALSE)
    Condition <geocodebr_error_cnefe_download_failed>
      Error in `download_cnefe()`:
      ! Could not download one or more CNEFE data files.
      i Please try again later.

# messages are formatted correctly

    Code
      definir_pasta_cache(path = NULL)
    Message <geocodebr_cache_dir>
      i Definido como pasta de cache '<path_to_default_dir>'.

---

    Code
      definir_pasta_cache("aaa")
    Message <geocodebr_cache_dir>
      i Definido como pasta de cache 'aaa'.

# behaves correctly

    Code
      listar_dados_cache(print_tree = TRUE)
    Output
      <path_to_cache_dir>
      +-- hello.parquet
      \-- oie.parquet

# deletar_pasta_cache behaves correctly

    Code
      res <- deletar_pasta_cache()
    Message <geocodebr_message_removed_cache_dir>
      v Deletada a pasta de cache que se encontrava em '<path_to_cache_dir>'.

# errors when all fields are NULL

    Code
      tester()
    Condition <geocodebr_error_null_address_fields>
      Error in `definir_campos()`:
      ! Pelo menos um campo não pode ser nulo `NULL`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{update_input_db}
\alias{update_input_db}
\title{Update input_padrao_db to remove observations previously matched}
\usage{
update_input_db(con, update_tb = "input_padrao_db", reference_tb)
}
\arguments{
\item{con}{A db connection}

\item{update_tb}{String. Name of a table to be updated in con}

\item{reference_tb}{A table written in con used as reference}
}
\value{
Drops observations from input_padrao_db
}
\description{
Update input_padrao_db to remove observations previously matched
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geocode.R
\name{geocode}
\alias{geocode}
\title{Geolocaliza endereços no Brasil}
\usage{
geocode(
  enderecos,
  campos_endereco = definir_campos(),
  resultado_completo = FALSE,
  resolver_empates = TRUE,
  resultado_sf = FALSE,
  h3_res = NULL,
  padronizar_enderecos = TRUE,
  verboso = TRUE,
  cache = TRUE,
  n_cores = NULL
)
}
\arguments{
\item{enderecos}{Um \code{data.frame}. Os endereços a serem geolocalizados. Cada
coluna deve representar um campo do endereço.}

\item{campos_endereco}{Um vetor de caracteres. A correspondência entre cada
campo de endereço e o nome da coluna que o descreve na tabela \code{enderecos}.
A função \code{\link[=definir_campos]{definir_campos()}} auxilia na criação deste vetor e realiza
algumas verificações nos dados de entrada. Campos de endereço passados
como \code{NULL} serão ignorados, e a função deve receber pelo menos um campo
não nulo, além  dos campos \code{"estado"} e \code{"municipio"}, que são
obrigatórios. Note que o campo  \code{"localidade"} é equivalente a 'bairro'.}

\item{resultado_completo}{Lógico. Indica se o output deve incluir colunas
adicionais, como o endereço encontrado de referência. Por padrão, é \code{FALSE}.}

\item{resolver_empates}{Lógico. Alguns resultados da geolocalização podem
indicar diferentes coordenadas possíveis (e.g. duas ruas diferentes com o
mesmo nome em uma mesma cidade). Esses casos são trados como 'empate' e o
parâmetro \code{resolver_empates} indica se a função deve resolver esses empates
automaticamente. Por padrão, é \code{TRUE}, e a função retorna apenas o caso
mais provável. Para mais detalhes sobre como é feito o processo de
desempate, consulte abaixo a seção "Detalhes".}

\item{resultado_sf}{Lógico. Indica se o resultado deve ser um objeto espacial
da classe \code{sf}. Por padrão, é \code{FALSE}, e o resultado é um \code{data.frame} com
as colunas \code{lat} e \code{lon}.}

\item{h3_res}{Um número que indica a resolução espacial da célula hexagonal
H3 da localização dos pontos retornados. Também aceita um vetor de números,
e.g. \code{c(8, 9)} Por padrão, é \code{NULL}. Detalhes sobre as resoluções disponíveis
em \url{https://h3geo.org/docs/core-library/restable/}}

\item{padronizar_enderecos}{Lógico. Indica se os dados de endereço de entrada
devem ser padronizados. Por padrão, é \code{TRUE}. Essa padronização é essencial
para uma geolocalizaçao correta. Alerta! Apenas utilize
\code{padronizar_enderecos = FALSE} caso os dados de input já tenham sido
padronizados anteriormente com \code{enderecobr::padronizar_enderecos(..., formato_estados = 'sigla', formato_numeros = 'integer')}.}

\item{verboso}{Um valor lógico. Indica se barras de progresso e mensagens
devem ser exibidas durante o download dos dados do CNEFE e a geocodificação
dos endereços. O padrão é \code{TRUE}.}

\item{cache}{Um valor lógico. Indica se os dados do CNEFE devem ser salvos ou
lidos do cache, reduzindo o tempo de processamento em chamadas futuras. O
padrão é \code{TRUE}. Quando \code{FALSE}, os dados do CNEFE são baixados para um
diretório temporário.}

\item{n_cores}{Um número. O número de núcleos de CPU a serem utilizados no
processamento dos dados. Por padrão, \code{n_cores = NULL} e o pacote utiliza o
número máximo de cores físicos disponíveis.}
}
\value{
Retorna o \code{data.frame} de input \code{enderecos} adicionado das colunas de
latitude (\code{lat}) e longitude (\code{lon}), bem como as colunas (\code{precisao} e
\code{tipo_resultado}) que indicam o nível de precisão e o tipo de resultado.
Alternativamente, o resultado pode ser um objeto \code{sf}.
}
\description{
Geocodifica endereços brasileiros com base nos dados do CNEFE. Os endereços
de input devem ser passados como um \code{data.frame}, no qual cada coluna
descreve um campo do endereço (logradouro, número, cep, etc). Os resuldos dos
endereços geolocalizados podem seguir diferentes níveis de precisão. Consulte
abaixo a seção "Detalhes" para mais informações. As coordenadas de output
utilizam o sistema de coordenadas geográficas SIRGAS 2000, EPSG 4674.
}
\details{
Precisão dos resultados:

A precisão dos resultados do \strong{{geocodebr}} são apresentadas em 3 colunas,
\code{precisao}, \code{tipo_resultado} e \code{desvio_metros}, explicadas abaixo.

Lidando com casos de empate:

No processo de geolocalização de dados, é possível que para alguns endereços
de input sejam encontrados diferentes coordenadas possíveis (e.g. duas ruas
diferentes com o mesmo nome, mas em bairros distintos em uma mesma cidade).
Esses casos são trados como empate'. Quando a função \code{geocode()} recebe o
o parâmetro \code{resolver_empates = TRUE}, os casos de empate são resolvidos
automaticamente pela função. A solução destes empates é feita da seguinte
maneira:
\enumerate{
\item Quando se encontra diferente coordenadas possíveis para um mesmo endereço
de input, nós assumimos que essas coordendas pertencem provavelmente a endereços
diferentes se (a) estas coordenadas estão a mais de 1Km entre si, ou (b) estão
associadas a um logradouro 'ambíguo', i.e. que costumam se repetir em muitos
bairros (e.g. "RUA A", "RUA QUATRO", "RUA 10", etc). Nestes casos, a solução
de desempate é retornar o ponto com maior número de estabelecimentos no CNEFE,
valor indicado na coluna \code{"contagem_cnefe"}.
\item Quando as coordenadas possivelmente associadas a um endereço estão a menos
de 1Km entre si e não se trata de um logradouro 'ambíguo', nós assumimos que
os pontos pertencem provavelmente ao mesmo logradouro (e.g. diferentes CEPs
ao longo de uma mesma rua). Nestes casos, a solução de desempate é retornar
um ponto que resulta da média das coordenadas dos pontos possíveis ponderada
pelo valor de \code{"contagem_cnefe"}. Nesse caso, a coluna de output
\code{"endereco_encontrado"} recebe valor do ponto com maior \code{"contagem_cnefe"}.
}
}
\section{Precisão}{
Os resultados são classificados em seis amplas categorias de \code{precisao}:
\enumerate{
\item "numero"
\item "numero_aproximado"
\item "logradouro"
\item "cep"
\item "localidade"
\item "municipio"
}

Cada nível de precisão pode ser desagregado em tipos de resultado mais
refinados.
}

\section{Tipos de resultados}{
A coluna \code{tipo_resultado} fornece informações mais detalhadas sobre como
exatamente cada endereço de entrada foi encontrado no CNEFE. Em cada
categoria,  o \strong{{geocodebr}} calcula a média da latitude e longitude dos
endereços incluídos no CNEFE que correspondem ao endereço de entrada, com
base em combinações de diferentes campos. No caso mais rigoroso, por exemplo,
a função encontra uma correspondência determinística para todos os campos de
um dado endereço (\code{"estado"}, \code{"municipio"}, \code{"logradouro"}, \code{"numero"},
\code{"cep"}, \code{"localidade"}). Pense, por exemplo, em um prédio com vários
apartamentos que correspondem ao mesmo endereço de rua e número. Nesse caso,
as coordenadas dos apartamentos podem diferir ligeiramente, e o
\strong{{geocodebr}} calcula a média dessas coordenadas. Em um caso menos rigoroso,
no qual apenas os campos (\code{"estado"}, \code{"municipio"}, \code{"logradouro"},
\code{"localidade"}) são encontrados, o \strong{{geocodebr}} calcula as coordenadas
médias de todos os endereços no CNEFE ao longo daquela rua e que se encontram
na mesma localidade/bairro. Assim, as coordenadas de resultado tendem a ser o
ponto médio do trecho daquela rua que passa dentro daquela localidade/bairro.

A coluna \code{tipo_resultado} fornece informações mais detalhadas sobre os campos de
endereço utilizados no cálculo das coordenadas de cada endereço de entrada. Cada
categoria é nomeada a partir de um código de quatro caracteres:
\itemize{
\item o primeiro caracter, sempre \code{d} ou \code{p}, determina se a correspondência foi
feita de forma determinística (\code{d}) ou probabilística (\code{p});
\item o segundo faz menção à categoria de \code{precisao} na qual o resultado foi
classificado (\code{n} para \code{"numero"}, \code{a} para \code{"numero_aproximado"}, \code{r} para
\code{"logradouro"}, \code{c} para \code{"cep"}, \code{b} para \code{"localidade"} e \code{m} para \code{"municipio"});
\item o terceiro e o quarto caracteres designam a classificação de cada categoria
dentro de seu grupo - via de regra, quanto menor o número formado por esses
caracteres, mais precisa são as coordenadas calculadas.
}

As categorias de \code{tipo_resultado} são listadas abaixo, junto às categorias de
\code{precisao} a qual elas estão associadas:
\itemize{
\item precisao \code{"numero"}
\itemize{
\item \code{dn01} - logradouro, numero, cep e localidade
\item \code{dn02} - logradouro, numero e cep
\item \code{dn03} - logradouro, numero e localidade
\item \code{dn04} - logradouro e numero
\item \code{pn01} - logradouro, numero, cep e localidade
\item \code{pn02} - logradouro, numero e cep
\item \code{pn03} - logradouro, numero e localidade
\item \code{pn04} - logradouro e numero
}
\item precisao \code{"numero_aproximado"}
\itemize{
\item \code{da01} - logradouro, numero, cep e localidade
\item \code{da02} - logradouro, numero e cep
\item \code{da03} - logradouro, numero e localidade
\item \code{da04} - logradouro e numero
\item \code{pa01} - logradouro, numero, cep e localidade
\item \code{pa02} - logradouro, numero e cep
\item \code{pa03} - logradouro, numero e localidade
\item \code{pa04} - logradouro e numero
}
\item precisao \code{"logradouro"} (quando o número de entrada está faltando 'S/N')
\itemize{
\item \code{dl01} - logradouro, cep e localidade
\item \code{dl02} - logradouro e cep
\item \code{dl03} - logradouro e localidade
\item \code{dl04} - logradouro
\item \code{pl01} - logradouro, cep e localidade
\item \code{pl02} - logradouro e cep
\item \code{pl03} - logradouro e localidade
\item \code{pl04} - logradouro
}
\item precisao \code{"cep"}
\itemize{
\item \code{dc01} - municipio, cep, localidade
\item \code{dc02} - municipio, cep
}
\item precisao \code{"localidade"}
\itemize{
\item \code{db01} - municipio, localidade
}
\item precisao \code{"municipio"}
\itemize{
\item \code{dm01} - municipio
}
}

Endereços não encontrados são retornados com latitude, longitude, precisão e
tipo de resultado \code{NA}.
}

\section{Desvio em metros}{
A coluna \code{desvio_metros} apresenta uma forma intuitiva e prática de saber o
grau de incerteza do resultado encontrado. Essa coluna informa que pelo menos
95\% de todos os pontos do CNEFE que possuem corrêspondência com o endereço de
input estão num raio de \code{x} metros da localização encontrada.

Um desvio de até \code{30} metros, por exemplo, tende a representar um resultado
muito confiável. A depender de como o dado geolocalizado será utilizado, até
mesmos resultados com \code{desvio_metros} de até 500 ou 900 metros podem ser ser
aceitáveis.

A coluna \code{desvio_metros} é particularmente útil para decidir por exemplo se
um resultado encontrado com a \code{precisao} de CEP, localidade ou logradouro
deveria ser aceitável. Por exemplo, muitas cidades do Brasil possuem um CEP
único, o que tende a gerar resultados com alto grau de incerteza. Em várias
cidades, no entanto, um CEP pode ser circunscrito a uma área muito pequena e
as vezes até um único edifício. Nesses casos, o valor do \code{desvio_metros}
tende a ser bem pequeno.
}

\section{Busca probabilistica}{
Os tipos de resultado com busca probabilitisca usam como base o algoritmo de
semelhança de Jaro para comparar as strings de 'logradouro' dos dados de
input e da base de endereços do geocodebr. O pacote considera como match o
logradouro da base de endereços que apresenta a maior semelhança de Jaro
condicionado a uma semelhança mínima, e desde que também haja match
determinístico em ao menos um dos campos "cep" e "localidade".
O geocodebr utiliza uma semelhança mínima de \code{0.85} nos casos de match
probabilistico, e de \code{0.90} nos demais casos.
}

\section{Código do setor censitário}{
Quando o usuário passa o argumento \code{resultado_completo = TRUE}, a função
\code{geocode()} também retorna a coluna \code{cod_setor} com o código do setor
censitário do endereço encontrado. Atualmente, a função somente retorna o
código do setor dos casos em que todos os pontos do CNEFE correspondentes estão
100\% dentro de um único setor censitário. Quando os dados do CNEFE
correspondentes ao endereço buscado estão em mais de um setor, o resultado da
coluna \code{cod_setor} é \code{NA}.
}

\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(geocodebr)

# ler amostra de dados
data_path <- system.file("extdata/small_sample.csv", package = "geocodebr")
input_df <- read.csv(data_path)[1:2,]

fields <- geocodebr::definir_campos(
  logradouro = "nm_logradouro",
  numero = "Numero",
  cep = "Cep",
  localidade = "Bairro",
  municipio = "nm_municipio",
  estado = "nm_uf"
)

df <- geocodebr::geocode(
  enderecos = input_df,
  campos_endereco = fields,
  resolver_empates = TRUE
  )

head(df)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{definir_pasta_cache}
\alias{definir_pasta_cache}
\title{Define um diretório de cache para o geocodebr}
\usage{
definir_pasta_cache(path, verboso = TRUE)
}
\arguments{
\item{path}{Uma string. O caminho para o diretório usado para armazenar os
dados em cache.  Se \code{NULL}, o pacote usará um diretório versionado salvo
dentro do diretório retornado por \code{\link[tools:userdir]{tools::R_user_dir()}}.}

\item{verboso}{Um valor lógico. Indica se barras de progresso e mensagens
devem ser exibidas durante o download dos dados do CNEFE e a geocodificação
dos endereços. O padrão é \code{TRUE}.}
}
\value{
Retorna de forma invisível o caminho do diretório de cache.
}
\description{
Define um diretório de cache para os dados do geocodebr. Essa configuração
é persistente entre sessões do R.
}
\examples{
definir_pasta_cache(tempdir())

# retoma pasta padrão do pacote
definir_pasta_cache( path = NULL)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/busca_por_cep.R
\name{busca_por_cep}
\alias{busca_por_cep}
\title{Busca por CEP}
\usage{
busca_por_cep(
  cep,
  h3_res = NULL,
  resultado_sf = FALSE,
  verboso = TRUE,
  cache = TRUE
)
}
\arguments{
\item{cep}{Vetor. Um CEP ou um vetor de CEPs com 8 dígitos.}

\item{h3_res}{Um número que indica a resolução espacial da célula hexagonal
H3 da localização dos pontos retornados. Também aceita um vetor de números,
e.g. \code{c(8, 9)} Por padrão, é \code{NULL}. Detalhes sobre as resoluções disponíveis
em \url{https://h3geo.org/docs/core-library/restable/}}

\item{resultado_sf}{Lógico. Indica se o resultado deve ser um objeto espacial
da classe \code{sf}. Por padrão, é \code{FALSE}, e o resultado é um \code{data.frame} com
as colunas \code{lat} e \code{lon}.}

\item{verboso}{Um valor lógico. Indica se barras de progresso e mensagens
devem ser exibidas durante o download dos dados do CNEFE e a geocodificação
dos endereços. O padrão é \code{TRUE}.}

\item{cache}{Um valor lógico. Indica se os dados do CNEFE devem ser salvos ou
lidos do cache, reduzindo o tempo de processamento em chamadas futuras. O
padrão é \code{TRUE}. Quando \code{FALSE}, os dados do CNEFE são baixados para um
diretório temporário.}
}
\value{
Retorna um \code{data.frame} com os CEPs de input e os endereços presentes
naquele CEP com suas coordenadas geográficas de latitude (\code{lat}) e
longitude (\code{lon}). Alternativamente, o resultado pode ser um objeto \code{sf}.
}
\description{
Busca endereços e suas coordenadas geográficas a partir de um CEP. As
coordenadas de output utilizam o sistema de coordenadas geográficas SIRGAS
2000, EPSG 4674.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(geocodebr)

# amostra de CEPs
ceps <- c("70390-025", "20071-001", "99999-999")

df <- geocodebr::busca_por_cep(
  cep = ceps,
  h3_res = 10,
  verboso = TRUE
  )

head(df)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{apaga_data_release_antigo}
\alias{apaga_data_release_antigo}
\title{Atualiza dados no release local}
\usage{
apaga_data_release_antigo()
}
\value{
Retorna de forma invisível o caminho do diretório de cache.
}
\description{
Detecta automaticamente se o release local esta desatualizado. Se sim, apaga
a pasta de cache do release local e atualiza a versao do release no caminho
da pasta
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{add_precision_col}
\alias{add_precision_col}
\title{Add a column with info of geocode match_type}
\usage{
add_precision_col(con, update_tb = NULL)
}
\arguments{
\item{con}{A db connection}

\item{update_tb}{String. Name of a table to be updated in con}
}
\value{
Adds a new column to a table in con
}
\description{
Add a column with info of geocode match_type
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geocodebr-package.R
\docType{package}
\name{geocodebr}
\alias{geocodebr}
\alias{geocodebr-package}
\title{Package: geocodebr: Geolocalização De Endereços Brasileiros (Geocoding
Brazilian Addresses)}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Método simples e eficiente de geolocalizar dados no Brasil. O pacote é baseado em conjuntos de dados espaciais abertos de endereços brasileiros, utilizando como fonte principal o Cadastro Nacional de Endereços para Fins Estatísticos (CNEFE). O CNEFE é publicado pelo Instituto Brasileiro de Geografia e Estatística (IBGE), órgão oficial de estatísticas e geografia do Brasil. (A simple and efficient method for geolocating data in Brazil. The package is based on open spatial datasets of Brazilian addresses, primarily using the Cadastro Nacional de Endereços para Fins Estatísticos (CNEFE), published by the Instituto Brasileiro de Geografia e Estatística (IBGE), Brazil's official statistics and geography agency.)
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/ipeaGIT/geocodebr}
  \item \url{https://ipeagit.github.io/geocodebr/}
  \item Report bugs at \url{https://github.com/ipeaGIT/geocodebr/issues}
}

}
\author{
\strong{Maintainer}: Rafael H. M. Pereira \email{rafa.pereira.br@gmail.com} (\href{https://orcid.org/0000-0003-2125-7465}{ORCID})

Authors:
\itemize{
  \item Daniel Herszenhut \email{dhersz@gmail.com} (\href{https://orcid.org/0000-0001-8066-1105}{ORCID})
  \item Gabriel Garcia de Almeida (\href{https://orcid.org/0009-0003-3557-7328}{ORCID})
}

Other contributors:
\itemize{
  \item Arthur Bazolli \email{baz.arthur@gmail.com} [contributor]
  \item Pedro Milreu Cunha \email{pedro.cunha@ipea.gov.br} [contributor]
  \item ITpS - Instituto Todos pela Saúde [funder]
  \item Ipea - Instituto de Pesquisa Econômica Aplicada [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_cnefe.R
\name{download_cnefe}
\alias{download_cnefe}
\title{Faz download dos dados do CNEFE}
\usage{
download_cnefe(tabela = "todas", verboso = TRUE, cache = TRUE)
}
\arguments{
\item{tabela}{Nome da tabela para ser baixada. Por padrão, baixa \code{"todas"}.}

\item{verboso}{Um valor lógico. Indica se barras de progresso e mensagens
devem ser exibidas durante o download dos dados do CNEFE e a geocodificação
dos endereços. O padrão é \code{TRUE}.}

\item{cache}{Um valor lógico. Indica se os dados do CNEFE devem ser salvos ou
lidos do cache, reduzindo o tempo de processamento em chamadas futuras. O
padrão é \code{TRUE}. Quando \code{FALSE}, os dados do CNEFE são baixados para um
diretório temporário.}
}
\value{
Retorna o caminho para o diretório onde os dados foram salvos.
}
\description{
Faz o download de uma versão pre-processada e enriquecida do CNEFE (Cadastro
Nacional de Endereços para Fins Estatísticos) que foi criada para o uso deste
pacote.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
download_cnefe(verboso = FALSE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{deletar_pasta_cache}
\alias{deletar_pasta_cache}
\title{Deletar pasta de cache do geocodebr}
\usage{
deletar_pasta_cache()
}
\value{
Retorna de forma invisível o caminho do diretório de cache.
}
\description{
Deleta todos arquivos da pasta do cache.
}
\examples{
\dontshow{if (identical(TRUE, FALSE)) withAutoprint(\{ # examplesIf}
deletar_pasta_cache()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{cache_message}
\alias{cache_message}
\title{Message when caching file}
\usage{
cache_message(
  local_file = parent.frame()$local_file,
  cache = parent.frame()$cache
)
}
\arguments{
\item{local_file}{The address of a file passed from the download_file function.}

\item{cache}{Logical. Whether the cached data should be used.}
}
\value{
A message
}
\description{
Message when caching file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{create_index}
\alias{create_index}
\title{create index}
\usage{
create_index(con, tb, cols, operation, overwrite = TRUE)
}
\description{
create index
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/definir_campos.R
\name{definir_campos}
\alias{definir_campos}
\title{Especifica as colunas que descrevem os campos dos endereços}
\usage{
definir_campos(
  estado,
  municipio,
  logradouro = NULL,
  numero = NULL,
  cep = NULL,
  localidade = NULL
)
}
\arguments{
\item{estado}{Uma string. O nome da coluna que representa o estado do
endereço. Campo obrigatório. Na tabela de endereços, essa coluna pode
conter os nomes dos estados por extenso, ou a abrevição oficial dos estados
com duas letras, e.g. "AM", "SP", "DF", "RJ".}

\item{municipio}{Uma string. O nome da coluna que representa o município do
endereço. Campo obrigatório. Na tabela de endereços, essa coluna pode
conter o nome dos municípios, ou o seu código IBGE de 7 dígitos.}

\item{logradouro}{Uma string. O nome da coluna que representa o \emph{logradouro}
(endereço da rua) do endereço. Pode ser \code{NULL} se o campo não estiver
especificado na tabela de endereços. Na tabela de endereços, essa coluna
deve incluir o \emph{tipo} do lograoduro, indicando se trata-se de uma "Rua" ou
"Avenida" etc, por exemplo "Avenida Presidente Getúlio Vargas". Além disso,
essa coluna \emph{não} deve incluir o \code{numero} do endereço, pois o número deve
ser indicado numa coluna separada.}

\item{numero}{Uma string. O nome da coluna que representa o número do endereço.
Pode ser \code{NULL} se o campo não estiver especificado na tabela de endereços.
Na tabela de endereços, valores como \code{0} ou caracteres não numerciso como
\code{"S/N"} ou \code{"10a"} são considerados como \code{NA}.}

\item{cep}{Uma string. O nome da coluna que representa o \emph{CEP} (Código de
Endereçamento Postal) do endereço. Pode ser \code{NULL} se o campo não estiver
especificado na tabela de endereços.}

\item{localidade}{Uma string. O nome da coluna que representa a localidade
(equivalente ao 'bairro' em áreas urbanas) do endereço. Pode ser \code{NULL} se
esse campo não estiver presente na tabela de endereços.}
}
\value{
Um vetor de caracteres no qual os nomes são os campos do endereço e os
valores são as colunas que os representam na tabela de endereços.
}
\description{
Cria um vetor de caracteres especificando as colunas que representam cada
campo do endereço na tabela de endereços. Os campos \code{estado} e \code{municipio}
são obrigatórios.
}
\examples{
definir_campos(
  logradouro = "Nome_logradouro",
  numero = "Numero",
  cep = "CEP",
  localidade = "Bairro",
  municipio = "Cidade",
  estado = "UF"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{listar_dados_cache}
\alias{listar_dados_cache}
\title{Listar dados em cache}
\usage{
listar_dados_cache(print_tree = FALSE)
}
\arguments{
\item{print_tree}{Um valor lógico. Indica se o conteúdo da pasta de cache
deve ser exibido em um formato de árvore. O padrão é \code{FALSE}.}
}
\value{
O caminho para os arquivos em cache
}
\description{
Lista os dados salvos localmente na pasta de cache
}
\examples{
listar_dados_cache()

listar_dados_cache(print_tree = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{arrow_open_dataset}
\alias{arrow_open_dataset}
\title{Safely use arrow to open a Parquet file}
\usage{
arrow_open_dataset(filename)
}
\arguments{
\item{filename}{A local Parquet file}
}
\value{
An \code{arrow::Dataset}
}
\description{
This function handles some failure modes, including if the Parquet file is
corrupted.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{listar_pasta_cache}
\alias{listar_pasta_cache}
\title{Obtém a pasta de cache usado no geocodebr}
\usage{
listar_pasta_cache()
}
\value{
O caminho da pasta de cache.
}
\description{
Obtém o caminho da pasta utilizada para armazenar em cache os dados do
geocodebr. Útil para inspecionar a pasta configurada com \code{\link[=definir_pasta_cache]{definir_pasta_cache()}}
em uma sessão anterior do R. Retorna a pasta de cache padrão caso nenhuma
pasta personalizado tenha sido configurada anteriormente.
}
\examples{
listar_pasta_cache()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geocode_reverso.R
\name{geocode_reverso}
\alias{geocode_reverso}
\title{Geocode reverso de coordenadas espaciais no Brasil}
\usage{
geocode_reverso(
  pontos,
  dist_max = 1000,
  verboso = TRUE,
  cache = TRUE,
  n_cores = NULL
)
}
\arguments{
\item{pontos}{Uma tabela de dados com classe espacial \verb{sf data frame} no
sistema de coordenadas geográficas SIRGAS 2000, EPSG 4674.}

\item{dist_max}{Integer. Distancia máxima aceitável (em metros) entre os
pontos de input e o endereço Por padrão, a distância é de 1000 metros.}

\item{verboso}{Um valor lógico. Indica se barras de progresso e mensagens
devem ser exibidas durante o download dos dados do CNEFE e a geocodificação
dos endereços. O padrão é \code{TRUE}.}

\item{cache}{Um valor lógico. Indica se os dados do CNEFE devem ser salvos ou
lidos do cache, reduzindo o tempo de processamento em chamadas futuras. O
padrão é \code{TRUE}. Quando \code{FALSE}, os dados do CNEFE são baixados para um
diretório temporário.}

\item{n_cores}{Um número. O número de núcleos de CPU a serem utilizados no
processamento dos dados. Por padrão, \code{n_cores = NULL} e o pacote utiliza o
número máximo de cores físicos disponíveis.}
}
\value{
Retorna o \verb{sf data.frame} de input adicionado das colunas do endereço
encontrado. O output inclui uma coluna "distancia_metros" que indica
a distância entre o ponto de input e o endereço mais próximo
encontrado.
}
\description{
Geocode reverso de coordenadas geográficas para endereços. A função recebe um
\verb{sf data frame} com pontos e retorna o endereço mais próximo dando uma
distância máxima de busca.
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
library(geocodebr)
library(sf)

# ler amostra de dados
pontos <- readRDS(
    system.file("extdata/pontos.rds", package = "geocodebr")
    )

pontos <- pontos[1:3,]

# geocode reverso
df_enderecos <- geocodebr::geocode_reverso(
  pontos = pontos,
  dist_max = 800,
  verboso = TRUE
  )

head(df_enderecos)
\dontshow{\}) # examplesIf}
}
